#include "host.h"


#include "immap_85xx.h"

#define FPGA_HOST_MSG_DAT_OFFSET 0xc
#define FPGA_DEVICE_MSG_DAT_OFFSET 0x10
#define FPGA_HOST_SIGNALS_OFFSET 0x14

#define PCIEXP(d) ((ccsr_pcix_t *)(d->pciexp + CONFIG_SYS_MPC85xx_PCIE1_OFFSET))

static u32 fix_reg_bits(struct devdata *devdata)
{
    u32 reg_bits = ioread32((void*)((u8*)devdata->fpgamsg + FPGA_DEVICE_MSG_DAT_OFFSET));

    while(reg_bits){
        devdata->reg_bits |= reg_bits;
        reg_bits = ioread32((void*)((u8*)devdata->fpgamsg + FPGA_DEVICE_MSG_DAT_OFFSET));
    }

    PRINTK_DEBUG("Fix Reg Bits 0x%08x\n", devdata->reg_bits);

    return devdata->reg_bits;
}


static void sgt_setup(struct devdata *devdata)
{
    u32 data;

    devdata->host2dev = &devdata->iomem->new.host2dev.fifo;
    devdata->host_read_wait = &devdata->iomem->new.dev2host.read_wait;
    devdata->host_write_wait = &devdata->iomem->new.host2dev.write_wait;
    devdata->dev2host = &devdata->iomem->new.dev2host.fifo;
    devdata->dev_read_wait = &devdata->iomem->new.host2dev.read_wait;
    devdata->dev_write_wait = &devdata->iomem->new.dev2host.write_wait;

    fix_reg_bits(devdata);

    devdata->reg_bits = (devdata->reg_bits & ~DEASSERT_IRQ_A);

    if (devdata->reg_bits)
        wake_up_all(&devdata->io_wq);

    if (readl(devdata->host_read_wait)) {
        if (rfifo_len(devdata->dev2host) >= readl(devdata->host_read_wait)) {
            devdata->read_irqs++;
            writel(0, devdata->host_read_wait);
            wake_up_all(&devdata->read_wq);
        }
    }

    if (readl(devdata->host_write_wait)) {
        if (rfifo_unused(devdata->host2dev) >= readl(devdata->host_write_wait)) {
            devdata->write_irqs++;
            writel(0, devdata->host_write_wait);
            wake_up_all(&devdata->write_wq);
        }
    }

    data = ioread32((void*)((u8*)devdata->fpgamsg + FPGA_HOST_SIGNALS_OFFSET));
    iowrite32((data | DEVICE_ENABLE_IRQ),(void*)((u8*)devdata->fpgamsg + FPGA_HOST_SIGNALS_OFFSET));
}

static void sgt_reset(struct devdata *devdata)
{
    return;
}

static void sgt_link(struct devdata *devdata, struct devdata *linkdev)
{
    return;
}

static void sgt_unlink(struct devdata *devdata)
{
    return;
}

static void sgt_gen_inbound_mbox_irq(struct devdata * devdata)
{
    iowrite32(0x1,(void*)((u8*)devdata->fpgamsg + FPGA_HOST_MSG_DAT_OFFSET));
}

static irqreturn_t _sgt_isr(struct devdata * devdata)
{

    u32 data;

    disable_irq_nosync(devdata->dev->irq);

    data = ioread32((void*)((u8*)devdata->fpgamsg + FPGA_HOST_MSG_DAT_OFFSET));
    iowrite32((data | DEASSERT_IRQ_A),(void*)((u8*)devdata->fpgamsg + FPGA_HOST_MSG_DAT_OFFSET));

    while((fix_reg_bits(devdata) & DEASSERT_IRQ_A) == 0) // wait till interrupt is dasserted
        ;

    devdata->reg_bits = (devdata->reg_bits & ~DEASSERT_IRQ_A);

    if (devdata->reg_bits)
        wake_up_all(&devdata->io_wq);

    if (readl(devdata->host_read_wait)) {
        if (rfifo_len(devdata->dev2host) >= readl(devdata->host_read_wait)) {
            devdata->read_irqs++;
            writel(0, devdata->host_read_wait);
            wake_up_all(&devdata->read_wq);
        }
    }

    if (readl(devdata->host_write_wait)) {
        if (rfifo_unused(devdata->host2dev) >= readl(devdata->host_write_wait)) {
            devdata->write_irqs++;
            writel(0, devdata->host_write_wait);
            wake_up_all(&devdata->write_wq);
        }
    }

    enable_irq(devdata->dev->irq);

    return IRQ_HANDLED;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static irqreturn_t sgt_isr(int irq, void *dev_id, struct pt_regs *regs)
#else
static irqreturn_t sgt_isr(int irq, void *dev_id)
#endif
{
	struct devdata * devdata = (struct devdata *)dev_id;
	irqreturn_t ret;

	spin_lock(&devdata->lock);
	ret = _sgt_isr(devdata);
	spin_unlock(&devdata->lock);

	return ret;
}

struct sg_dev sgt_device = {
	.setup = sgt_setup,
	.reset = sgt_reset,
	.link = sgt_link,
	.unlink = sgt_unlink,
	.gen_inbound_mbox_irq = sgt_gen_inbound_mbox_irq,
	.isr = sgt_isr,
};
