#ifndef _SGPCIE_H_
#define _SGPCIE_H_

#include "linux/ioctl.h"

#ifndef __user
#define	__user
#endif

#define	SGS_REG_SIZE		4096
#define	SGS_DEVDATA_SIZE	256

#define _SGS_IOC_MAGIC		0xa9
#define	_SGS_IOCTL_REGMAP	0x00
#define	_SGS_IOCTL_REGS		0x01
#define	_SGS_IOCTL_REG		0x02

#define SGS_WAIT_IRQ		_IOW(_SGS_IOC_MAGIC, 0x0, unsigned int)
#define SGS_WAIT_IRQ_MASK	_IOW(_SGS_IOC_MAGIC, 0x1, struct wait_mask_t)
#define SGS_LINK		_IOW(_SGS_IOC_MAGIC, 0x2, unsigned int)
#define SGS_UNLINK		_IO(_SGS_IOC_MAGIC, 0x3)
#define SGS_DET_LINE		_IO(_SGS_IOC_MAGIC, 0x4)
#define SGS_SET_DEVDATA		_IOW(_SGS_IOC_MAGIC, 0x5, dev_data_t)
#define SGS_GET_DEVDATA		_IOR(_SGS_IOC_MAGIC, 0x6, dev_data_t)
#define SGS_DET_LINE_OFF	_IO(_SGS_IOC_MAGIC, 0x7)
#define SGS_DET_LINE_ON		_IO(_SGS_IOC_MAGIC, 0x8)
#define SGS_WAIT_UP		_IO(_SGS_IOC_MAGIC, 0x9)
#define SGS_UNPLUG_FPGA   _IO(_SGS_IOC_MAGIC, 0xa)
#define SGS_PLUG_FPGA   _IO(_SGS_IOC_MAGIC, 0xb)

struct wait_mask_t {
	unsigned int timeout;
	unsigned int mask;
};

typedef unsigned char dev_data_t[SGS_DEVDATA_SIZE];

#ifdef __KERNEL__

#include "rfifo.h"

#define	SGS_VENDOR_SUB		0x162f
#define	SGS_PRODUCT_SUB		0x1335
#define SGT_PRODUCT_SUB     0x1336
#define	SGS_PRODUCT_SUB_OLD	0x132e

#define	SGSPCIE_BAR1ORDER	8
#define	SGSPCIE_BAR1SIZE	(1 << (SGSPCIE_BAR1ORDER + 12))	/* 1 MegaByte */

#define	SGSDEV_IO_SIZE		SGSPCIE_BAR1SIZE

struct sgpcie_iomsg {
	__le32 _align(32)	write_wait;
	__le32 _align(32)	read_wait;
	struct rfifo _align(32)	fifo;
} _align(32);

struct sgpcie_iomem {
	u8 _align(4096)			reg_data[SGS_REG_SIZE];
	struct sgpcie_iomsg _align(32)	dev2host;	
	struct sgpcie_iomsg _align(32)	host2dev;	
	struct sgpcie_iomsg _align(32)	ep2ep;
	u8 _align(32)			dev_data[SGS_DEVDATA_SIZE];
	u8 _align(4096)			reg_data_ep[SGS_REG_SIZE];
} _align(4096);

struct sgpcie_iomem_old {
	u8 _align(4096)		reg_data[SGS_REG_SIZE];
	__le32 _align(32)	dev_read_wait;
	__le32 _align(32)	dev_write_wait;
	__le32 _align(32)	host_read_wait;
	__le32 _align(32)	host_write_wait;
	struct rfifo _align(32)	dev2host;
	struct rfifo _align(32)	host2dev;
} _align(4096);

#endif

#endif

