#include <visa.h>

/*= Useful Macros ===========================================================*/

/* - Basic Macros ---------------------------------------------------------- */

/* Error checking macro */
#define CHECKERR(fCal) if (rssgs_status = (fCal), rssgs_status < VI_SUCCESS) \
    return rssgs_status

/* Error checking macro with jump to the ERROR: label */
#define GOCHECKERR(fCal) if (rssgs_status = (fCal), rssgs_status < VI_SUCCESS) \
                            goto ERROR_LABEL
/**#define GOCHECKERR(fCal) if (rssgs_status = (fCal), rssgs_status < VI_SUCCESS) \
                            goto ERROR_LABEL; else
*/
/* Parameter checking macro */
#define CHECKPARAM(fCal,param) if (fCal != VI_SUCCESS) return VI_ERROR_PARAMETER##param

/* Checking path parameter */
#define CHECKPATH if (rssgs_invalidViInt32Range (path, RSSGS_PATH_A, RSSGS_PATH_B)) return VI_ERROR_PARAMETER2;

/* Initialize default function parameters */
#define INIT\
    ViStatus            rssgs_status = VI_SUCCESS; \
    int                 handle = -1;\
    rssgs_instrRange    instrPtr = NULL; \
    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &handle));\
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer(handle));\
    if (instrPtr == NULL)\
        return VI_ERROR_PARAMETER1

/* Close function and return */
#define RETURN\
    return rssgs_status;

/* - Additional Macros ----------------------------------------------------- */

#define CHECKOPT(option) if (instrPtr->optionChecking == VI_TRUE)\
                         {\
                            if (instrPtr->instrOptions.option == 0)\
                                return RSSGS_ERROR_INSTRUMENT_OPTION;\
                         }

#define CHECKOPTCONT(option) if (instrPtr->optionChecking == VI_TRUE)\
                         {\
                            if (instrPtr->instrOptions.option == 0)\
                                rssgs_status = RSSGS_ERROR_INSTRUMENT_OPTION;\
                         }
/* Remap parameter error code */
#define REMAP(par_in, par_out) if (rssgs_status == VI_ERROR_PARAMETER##par_in)\
                                    rssgs_status = VI_ERROR_PARAMETER##par_out;\
                                if (rssgs_status < VI_SUCCESS)\
                                    return rssgs_status

#define REMAPINSTR(par_in, par_out) if (rssgs_status == VI_ERROR_PARAMETER##par_in)\
                                    rssgs_status = VI_ERROR_INSTR_PARAMETER##par_out;\
                                if (rssgs_status < VI_SUCCESS)\
                                    return rssgs_status

/* Check NULL pointer */
#define CHECKNULL(pointer, param) if (pointer == NULL) return VI_ERROR_PARAMETER##param


#if defined(LONG_MAX) && (LONG_MAX > 0x7FFFFFFFL)
  #define _PERCD "%d"
  #define _PERCU "%u"
  #define _PERCUD "%ud"
  #define _PERC1U "%1u"
  #define _PERCX "%x"
#else
  #define _PERCD "%ld"
  #define _PERCU "%lu"
  #define _PERCUD "%lud"
  #define _PERC1U "%1lu"
  #define _PERCX "%lx"
#endif



/* Supported instruments */

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT STATUS/RANGE STRUCTURE  ============================*/
/*****************************************************************************/
/* rssgs_stringValPair is used in the rssgs_error_message function           */
/* rssgs_statusDataRanges is used to track session dependent status & ranges */
/*===========================================================================*/
typedef struct  rssgs_stringValPair
{
   ViStatus stringVal;
   ViString stringName;
}  rssgs_tStringValPair;

typedef struct rssgs_instrOptions
{
  int B1:1;
  int B26:1;
  int B106:1;
  int B106V:1;
  int B112:1;
  int B112V:1;
  int K22:1;
  int K90:1;
} rssgs_tinstrOptions;

struct rssgs_statusDataRanges
{
    ViChar              instrDriverRevision[256];
    ViBoolean           errorChecking;
    ViBoolean           optionChecking;
    ViBoolean           rangeChecking;
    ViInt32             instrVer;
    rssgs_tinstrOptions instrOptions;
};

typedef struct rssgs_statusDataRanges *rssgs_instrRange;
ViBoolean rssgs_invalidViBooleanRange (ViSession instrSession, ViBoolean val);
ViBoolean rssgs_invalidViInt32Range (ViSession instrSession, ViInt32 val, ViInt32 min, ViInt32 max);
ViBoolean rssgs_invalidViReal64Range (ViSession instrSession, ViReal64 val, ViReal64 min,
                                       ViReal64 max);
ViStatus rssgs_checkStatus (ViSession instrSession);
ViStatus rssgs_waitOPC (ViSession instrSession, ViPString command,
                        ViPString instrResp, ViInt32 timeout);
ViStatus rssgs_sendCmd (ViSession instrSession, ViString cmd);
ViStatus rssgs_setViInt32 (ViSession instrSession,
                           ViString cmd,
                           ViInt32 value,
                           ViInt32 min,
                           ViInt32 max,
                           ViStatus parameterErrorCode);
ViStatus rssgs_getViInt32 (ViSession instrSession,
                           ViString cmd,
                           ViInt32 *value,
                           ViStatus parameterErrorCode);
ViStatus rssgs_setViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean value,
                           ViStatus parameterErrorCode);
ViStatus rssgs_getViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean *value,
                           ViStatus parameterErrorCode);
ViStatus rssgs_setViReal64 (ViSession vi,
                            ViString cmd,
                            ViReal64 value,
                            ViReal64 min,
                            ViReal64 max,
                            ViStatus parameterErrorCode);
ViStatus rssgs_getViReal64 (ViSession instrSession,
                            ViString cmd,
                            ViReal64 *value,
                            ViStatus parameterErrorCode);
ViStatus rssgs_setViReal64Units (ViSession instrSession,
                                 ViString cmd,
                                 ViReal64 value,
                                 ViString units,
                                 ViReal64 min,
                                 ViReal64 max,
                                 ViStatus parameterErrorCode);
ViStatus rssgs_setViString (ViSession instrSession,
                            ViString cmd,
                            ViString value,
                            ViStatus parameterErrorValue);
ViStatus rssgs_getViString (ViSession instrSession,
                            ViString cmd,
                            ViString value,
                            ViUInt32 buffer_size,
                            ViStatus parameterErrorCode);
ViStatus rssgs_setViStringQuoted (ViSession instrSession,
                            ViString cmd,
                            ViString quotes,
                            ViString value,
                            ViStatus parameterErrorValue);
ViStatus rssgs_setEnum (ViSession instrSession,
                        ViString cmd,
                        ViInt32 value,
                        ViString list[],
                        ViStatus parameterErrorCode);
ViStatus rssgs_getEnum (ViSession instrSession,
                        ViString cmd,
                        ViInt32 *value,
                        ViString list[],
                        ViStatus parameterErrorCode);
ViStatus rssgs_checkIndexInList (ViInt32 index,
                                 ViString list[],
                                 ViStatus statusIfError);
ViStatus rssgs_getIndexFromString (ViString string,
                                   ViString *list,
                                   ViInt32 *index);
ViStatus rssgs_getCatalog (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 arraySize,
                           ViChar* array,
                           ViInt32 parameterErrorCode);
ViStatus rssgs_getViReal64Array (ViSession instrSession,
                            ViString cmd,
                            ViReal64 *value,
                            ViInt32  count,
                            ViStatus parameterErrorCode);
ViStatus rssgs_getViInt32Array (ViSession instrSession,
                            ViString cmd,
                            ViInt32 *value,
                            ViInt32  count,
                            ViStatus parameterErrorCode);
ViStatus rssgs_readDataUnknownLength(ViSession instrSession,
                                     /*@null@*/ViChar **outBuffer,
                                     ViUInt32 *count);
ViStatus rssgs_getViUInt32 (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 *value,
                           ViStatus parameterErrorCode);
ViStatus rssgs_getViStringQuoted (ViSession instrSession,
                                  ViString cmd,
                                  ViString quotes,
                                  ViInt32  arraySize,
                                  ViChar   value[],
                                  ViStatus parameterErrorValue);
void rssgs_SWAP_32(unsigned char *data, long points);
ViStatus _VI_FUNC rssgs_ReadTraceData(
    ViSession   instrSession,
    ViString    cmd,
    ViInt32*    noofPoints,
    ViReal64    _VI_FAR traceData[]
);
ViStatus rssgs_checkBeforeRead (ViSession instrSession);
ViStatus rssgs_readDataToFile (ViSession instrSession, ViString cmd,
                           ViString destination);
ViStatus rssgs_checkInstrument (ViSession instrSession, ViInt32 instrument, ViBoolean generateErrorIfPresent);

unsigned int rssgs_AddPointer(void* pPtr);
void rssgs_RemovePointer(unsigned int handle);
void* rssgs_GetPointer(unsigned int handle);

ViInt32 rssgs_StringIndex(ViString array[], ViChar buff[]);

ViStatus rssgs_getViBooleanArray (ViSession instrSession,
                            ViString cmd,
                            ViBoolean *value,
                            ViInt32  count,
							ViInt32  *returnedValues,
                            ViStatus parameterErrorCode);

