
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "rssgs.h"
#include "SgmaCommandIds.h"

#ifdef __linux__
  #include <termios.h>
  #include <unistd.h>
  #include <time.h>

  #define _stricmp strcasecmp
  #define Sleep(msec) usleep((msec)*1000)
  #ifndef BOOL
    typedef int BOOL;
  #endif

  #ifndef TRUE
    #define TRUE 1    
    #define FALSE 0    
  #endif

  typedef long long __int64;
  
  #define __fastcall
  typedef int HANDLE;
  #define INVALID_HANDLE_VALUE -1    
#else
  #pragma comment(linker, "/SUBSYSTEM:CONSOLE")
  #include "Windows.h"
  #include <conio.h>
  #include <intrin.h>
  #pragma intrinsic(__rdtsc)
#endif


#ifdef WITHPOWERSENSOR
  #include "rsnrpz.h"
#endif

// Controlling the serial interface for test purposes
typedef HANDLE(__fastcall *FuncSerialOpen) (const char* i_port);
typedef void(__fastcall *FuncSerialClose)(HANDLE fd_);
typedef void(__fastcall *FuncSerialSetRTS) (HANDLE fd_, int level);
typedef void(__fastcall *FuncSerialSetDTR) (HANDLE fd_, int level);
FuncSerialOpen   SerialOpen = NULL;
FuncSerialClose  SerialClose = NULL;
FuncSerialSetRTS SerialSetRTS = NULL;
FuncSerialSetDTR SerialSetDTR = NULL;
HANDLE SerialHandle = INVALID_HANDLE_VALUE;

void InitControlSerial();
void SET_RTS();
void CLR_RTS();
void SET_DTR();
void CLR_DTR();
void INIT_SERIAL();
void CLOSE_SERIAL();

double g_dProcessorClockrate = -1.0;


#define SIGGENTYPEUNKNOWN 0
#define SIGGENTYPESGS     1
#define MAXCMDSIZE 6

#define MaxNumSGS 4
#define MaxNumSGU 4

typedef char COMMAND [MAXCMDSIZE][100];  
enum HWCHAN { LAN, USB, PCIE, SOCK, HISLIP, UNKNOWN };
int g_hwChannel = UNKNOWN;


void CommandLoop();
const char* DecodeVisaError (ViSession instrumentHandle, ViStatus i_stat);
BOOL determineProcessorClockrate();
double getClockInMicroSeconds(BOOL bIsToCalculate);
void TestExecutionTime(int instrumentHandle);
void TestExecutionTime2(int instrumentHandle);

ViBoolean g_useFastSettings = VI_TRUE;
ViBoolean g_enableAsynchronousFastSettings = VI_FALSE;

ViSession g_instrumentHandle[MaxNumSGS];
int g_activeDevice = 0;
ViInt32 g_SgsSerialNumbers[10] = {0};
ViInt32 g_SguSerialNumbers[10] = {0};
ViInt32 g_nSgsCount = 1;
ViInt32 g_nSguCount = 0;


int g_ExtensionIndexOfSguSerialIndex[MaxNumSGS][MaxNumSGU];


#ifdef __linux__
int _getch(void)
{
  struct termios oldattr, newattr;
  int ch;
  tcgetattr( STDIN_FILENO, &oldattr );
  newattr = oldattr;
  newattr.c_lflag &= ~( ICANON | ECHO );
  tcsetattr( STDIN_FILENO, TCSANOW, &newattr );
  ch = getchar();
  tcsetattr( STDIN_FILENO, TCSANOW, &oldattr );
  return ch;
}
#endif

// ---------------------------------------------------------------------
//  OpenDevices
// 
// ---------------------------------------------------------------------

void OpenDevices (char* szResource)
{
  ViStatus status = VI_SUCCESS;
  char buffer[1000] = { 0 };
  ViInt32 errorCode = 0;
  int iSgsDevice;

  for (iSgsDevice = 0; iSgsDevice < g_nSgsCount; iSgsDevice++)
  {
    ViBoolean idQuery = VI_FALSE;
    ViBoolean resetDevice = VI_FALSE;


    if (szResource)
    {
      strcpy(buffer, szResource);
    }
    else 
    {
      if (g_hwChannel == PCIE)
        sprintf(buffer, "PCIe::0x162f::0x132e::%d::INSTR", (int)g_SgsSerialNumbers[iSgsDevice]);

      if (g_hwChannel == SOCK)
        sprintf(buffer, "TCPIP::RSSGS100A%d::5025::SOCKET", (int)g_SgsSerialNumbers[iSgsDevice]);
    }
    status = rssgs_init(buffer, idQuery, resetDevice, &g_instrumentHandle[iSgsDevice]);
//    printf("Open <%s> %d\n", buffer, g_instrumentHandle[iSgsDevice]);
    if (status)
    {
      g_instrumentHandle[iSgsDevice] = 0;
      rssgs_error_message(g_instrumentHandle[iSgsDevice], status, buffer);
      printf("open FAILED status = 0x%x (%s)\n", (unsigned int)status, buffer);
    }
    else
    {
      status = rssgs_error_query(g_instrumentHandle[iSgsDevice], &errorCode, buffer);
      if (errorCode)
        printf("%d %s\n", (int)errorCode, buffer);

      rssgs_UseFastSettings(g_instrumentHandle[iSgsDevice], g_useFastSettings, g_enableAsynchronousFastSettings);
    }
  }
}

void CloseDevices()
{
  int iSgsDevice = 0;
  for (iSgsDevice = 0; iSgsDevice < g_nSgsCount; iSgsDevice++)
  {
    if (g_instrumentHandle[iSgsDevice])
      rssgs_close(g_instrumentHandle[iSgsDevice]);
  }
}


// ---------------------------------------------------------------------
//  TestRFSignalConnected
// 
//  returns true if the  RF signal is connected to the SGU
//  else returns false
// ---------------------------------------------------------------------


ViBoolean TestRFSignalConnected(int instrumentHandle)
{
  ViInt32  error = 1, lasttimeout;
  ViStatus status = VI_SUCCESS;

  rsGetAttribute(instrumentHandle, VI_ATTR_TMO_VALUE, &lasttimeout);
  rsSetAttribute(instrumentHandle, VI_ATTR_TMO_VALUE, 30000);

  status = rssgs_GetTestRFSignalConnection(instrumentHandle, &error);

  rsSetAttribute(instrumentHandle, VI_ATTR_TMO_VALUE, lasttimeout);

  return status == VI_SUCCESS && error == 0;
}


void ScanExtensions(int instrumentHandle, int* ExtensionIndices)
{
  ViInt32 Int32 = 0;
  ViInt32 numExtensions = 0;
  int i,j;
  ViInt32 hwChannels[10];
  ViInt32 serialNumbers[10];
  ViInt32 status;
 
  do
    {	
    status = rssgs_SetExtensionInstScanState (instrumentHandle, 1);
    if (status < 0)
      break;

    status = rssgs_OperationCompleteQuery (instrumentHandle, &Int32);
    if (status < 0)
      break;

    status = rssgs_ExtensionInstHardwareChannels (instrumentHandle, 10, hwChannels, &numExtensions);
    if (status < 0)
      break;

    status = rssgs_ExtensionInstSerialNumbers (instrumentHandle, 10, serialNumbers, &numExtensions);

  }while (0);


  for (j = 0; j < MaxNumSGU; j++)
  {
    ExtensionIndices[j] = -1;
    for (i = 0; i < numExtensions; i++)
    {
      if (serialNumbers[i] == g_SguSerialNumbers[j] && hwChannels[i] == g_hwChannel)
      {
        ExtensionIndices[j] = i+1;
        //printf ("SGU-%d at extension list index %d\n", serialNumbers[i], i+1);
      }
    }
    // prefer PCIe over socket
    for (i = 0; i < numExtensions; i++)
    {
      if (serialNumbers[i] == g_SguSerialNumbers[j] && hwChannels[i] == PCIE)
      {
        ExtensionIndices[j] = i + 1;
        //printf("SGU-%d at extension list index %d via PCIe\n", serialNumbers[i], i + 1);
      }
    }
  }
}

ViBoolean HasActiveExtension(ViSession hdlSGS, ViBoolean poll)
{
  ViBoolean remotestate = 0;
  ViBoolean busystate = 0;
  int i = 0;

  for (i = 0; i < (poll ? 30 : 1) && !remotestate; i++)
  {
    rssgs_GetExtensionRemoteState(hdlSGS, &remotestate);
    Sleep(100);
  }
  if (remotestate)
    rssgs_GetExtensionBusyState(hdlSGS, &busystate);

  return remotestate && !busystate;
}


ViBoolean ConnectExtension(int iSgsDevice, int iSguDevice)
{
  ViBoolean remotestate = 0;

  printf("Connecting SGS-%d and SGU-%d\n", (int)g_SgsSerialNumbers[iSgsDevice], (int)g_SguSerialNumbers[iSguDevice]);
  rssgs_SetExtensionSelect(g_instrumentHandle[iSgsDevice], g_ExtensionIndexOfSguSerialIndex[iSgsDevice][iSguDevice]);

  remotestate = HasActiveExtension (g_instrumentHandle[iSgsDevice], VI_TRUE);

  if (!remotestate)
    printf("Cannot establish remote connection between SGS-%d and SGU-%d\n", (int)g_SgsSerialNumbers[iSgsDevice], (int)g_SguSerialNumbers[iSguDevice]);

  return remotestate;
}

ViBoolean DisconnectExtension(int iSgsDevice)
{
  ViBoolean remotestate = 1;
  int i = 0;

  rssgs_SetExtensionSelect(g_instrumentHandle[iSgsDevice], 0);

  for (i = 0; i < 100 && remotestate; i++)
  {
    rssgs_GetExtensionRemoteState(g_instrumentHandle[iSgsDevice], &remotestate);
    Sleep(100);
  }

  if (remotestate)
    printf("Cannot disconnect remote connection between SGS-%d and its SGU\n", (int)g_SgsSerialNumbers[iSgsDevice] );

  return !remotestate;
}


void PairExtensions()
{
  int iSgsDevice = 0;
  int iSguDevice = 0;
  ViBoolean SguPaired[MaxNumSGU];

  for (iSguDevice = 0; iSguDevice < g_nSguCount; iSguDevice++)
    SguPaired[iSguDevice] = VI_FALSE;


  for (iSgsDevice = 0; iSgsDevice < g_nSgsCount; iSgsDevice++)
  {
    for (iSguDevice = 0; iSguDevice < g_nSguCount; iSguDevice++)
    {
      if (!SguPaired[iSguDevice])
      {
        ViBoolean remotestate = 0;
        ViBoolean rfconnected = 0;

        remotestate = ConnectExtension(iSgsDevice, iSguDevice);
        if (remotestate)
        {
          rfconnected = TestRFSignalConnected(g_instrumentHandle[iSgsDevice]);

          if (rfconnected)
          {
            printf("RF connected\n");
            SguPaired[iSguDevice] = VI_TRUE;
            break;
          }
          else
          {
            printf("RF not connected\n");
            DisconnectExtension(iSgsDevice);
          }
        }
      }
    } // for SGU
  } // for SGS
}


void PrintCommandList()
{
  printf ("cmd:\n" \
    "\t1..\tToggle active SGS\n" \
    "\tap\tAutopair SGSs with SGUs\n" \
    "\te[?]\tError status\n" \
    "\tf[?]\tFrequency\n" \
    "\tl[?]\tLevel\n" \
    "\tfl\tFrequency and Level\n" \
    "\tr[?]\tRF State\n" \
    "\tm[?]\tIQ Modulator State\n" \
    "\tw[?]\tIQ Wideband State\n" \
    "\ti?\tIdn Query\n" \
    "\taf?\tAdjust Frequency\n" \
    "\tfm[?]\tFast Mode\n" \
    "\tscpi[?]\tWrite / Query SCPI\n" \
    "\ttsig\tTest SGU signals\n" \
    "\tfm[?]\tFast Mode\n" \
    "\tq\tQuit\n");
}


// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------
#ifndef __linux__

// Start SGMANDISPROT protocol driver for fast socket channel
BOOL StartStopNdis(BOOL bStart)
{
  LPCWSTR szServiceName = L"SGMANDISPROT";
  SC_HANDLE serviceDbHandle = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
  SC_HANDLE serviceHandle = serviceDbHandle ? OpenService(serviceDbHandle, szServiceName, SC_MANAGER_ALL_ACCESS) : NULL;

  SERVICE_STATUS_PROCESS status;
  DWORD bytesNeeded;
  LPWSTR szOperation = bStart ? L"start" : L"stop";
  BOOL bSuccess = TRUE;

  if (serviceHandle)
  {
    QueryServiceStatusEx(serviceHandle, SC_STATUS_PROCESS_INFO, (LPBYTE)&status, sizeof(SERVICE_STATUS_PROCESS), &bytesNeeded);

    if (status.dwCurrentState == SERVICE_RUNNING && !bStart)
      bSuccess = ControlService(serviceHandle, SERVICE_CONTROL_STOP, (LPSERVICE_STATUS)&status);
    else if (status.dwCurrentState != SERVICE_RUNNING && bStart)
      bSuccess = StartService(serviceHandle, 0, NULL);
  }
  else
  {
    bSuccess = FALSE;
  }

  if (!bSuccess)
  {
    DWORD dwLastError = GetLastError();
    WCHAR   lpBuffer[256];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      dwLastError,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      lpBuffer,
      255,
      NULL);
    wprintf(L"%s failed to %s: %s\n", szServiceName, szOperation, lpBuffer);
  }

  CloseServiceHandle(serviceHandle);
  CloseServiceHandle(serviceDbHandle);
  return bSuccess;
}

BOOL StartNdis()
{
  return StartStopNdis(TRUE);
}
#endif
// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

ViStatus InterpretCommand (ViSession instrumentHandle, COMMAND szCommand)
{
  ViStatus status = VI_SUCCESS;
  double dValue = 0;
  ViInt32 nValue = 0;
  ViBoolean bState = 0;
  char buffer [1000] = {0};
  ViInt32 errorCode = 0; 

  if (!strcmp(szCommand[0], "?"))
  {
    PrintCommandList();
  }
  // Select active device
  else if (strlen(szCommand[0]) == 1 && strchr("123456789", szCommand[0][0]))
  {
    int n = szCommand[0][0] - '1';
    if (n < g_nSgsCount)
    {
      int activeExtension;
      g_activeDevice = n;
      activeExtension = HasActiveExtension(g_instrumentHandle[g_activeDevice], VI_FALSE);
      printf("Active device is SGS-%d (%s)\n", (int)g_SgsSerialNumbers[g_activeDevice], activeExtension ? "paired" : "unpaired");
    }
    else
    {
      printf("Value out of range\n");
    }
  }

  // Set / get frequency ------------------------- 
  else if (!strcmp(szCommand[0], "f"))
  {
    size_t len = strlen(szCommand[1]);
    double value = atof (szCommand[1]);
    if (len > 3 || *szCommand[2])
    {
      if (!_stricmp(szCommand[1]+len-3, "mhz") || !_stricmp(szCommand[2], "mhz"))
        value *= 1e6;
      if (!_stricmp(szCommand[1]+len-3, "ghz") || !_stricmp(szCommand[2], "ghz"))
        value *= 1e9;
    }
    status = rssgs_SetFrequency(instrumentHandle, value);

  }

  else if (!strcmp(szCommand[0], "f?"))
  {
    status = rssgs_GetFrequency(instrumentHandle, &dValue);

    if (status == VI_SUCCESS)
    {
      if (dValue > 1e9)
        printf ("Freq = %f GHz\n", dValue/1e9 );
      else if (dValue > 1e6)
        printf ("Freq = %f MHz\n", dValue/1e6 );
      else 
        printf ("Freq = %f Hz\n", dValue);
    }
  }
  // frequency offset
  else if (!strcmp(szCommand[0], "fo"))
  {
    size_t len = strlen(szCommand[1]);
    double value = atof(szCommand[1]);
    if (len > 3 || *szCommand[2])
    {
      if (!_stricmp(szCommand[1] + len - 3, "mhz") || !_stricmp(szCommand[2], "mhz"))
        value *= 1e6;
      if (!_stricmp(szCommand[1] + len - 3, "ghz") || !_stricmp(szCommand[2], "ghz"))
        value *= 1e9;
    }
    status = rssgs_SetFrequencyOffset(instrumentHandle, value);

  }

  else if (!strcmp(szCommand[0], "fo?"))
  {
    status = rssgs_GetFrequencyOffset(instrumentHandle, &dValue);

    if (status == VI_SUCCESS)
    {
      if (dValue > 1e9)
        printf("Freq = %f GHz\n", dValue / 1e9);
      else if (dValue > 1e6)
        printf("Freq = %f MHz\n", dValue / 1e6);
      else
        printf("Freq = %f Hz\n", dValue);
    }
  }

  // Test signal connections ------------------------- 
  else if (!strcmp(szCommand[0], "tsig"))
  {
    ViBoolean connected = TestRFSignalConnected(instrumentHandle);
   	if (!connected)	
     	rssgs_SYSTemErrorQueueAll (instrumentHandle,1000, buffer);
    printf(connected ? "SGU connected\n" : "SGU not connected\n");
  }


  // Test ------------------------- 
  else if (!strcmp(szCommand[0], "test"))
  {
    int i = 0;  
    for (i = 0; i < 100; i++)
    {
      double freq = 2e9;
      if (i & 1)
        freq = 3e9;      

      status = rssgs_SetFrequency(instrumentHandle, freq);

      Sleep (100);  
    }
  }

  else if (!strcmp(szCommand[0], "tt"))
  {
    TestExecutionTime(instrumentHandle);
  }
  else if (!strcmp(szCommand[0], "tt2"))
  {
    TestExecutionTime2(instrumentHandle);
  }



  // Autopairing ------------------------- 
  else if (!strcmp(szCommand[0], "ap"))
  {
    int iSgsDevice = 0;
    for (iSgsDevice = 0; iSgsDevice < g_nSgsCount; iSgsDevice++)
    {
      DisconnectExtension(iSgsDevice);

      if (g_ExtensionIndexOfSguSerialIndex[iSgsDevice][0] <= 0)
        ScanExtensions(g_instrumentHandle[iSgsDevice], g_ExtensionIndexOfSguSerialIndex[iSgsDevice]);
    }

    PairExtensions();
  }

  // Set / get level ------------------------- 
  else if (!strcmp(szCommand[0], "l"))
  {
    status = rssgs_SetLevel(instrumentHandle, atof (szCommand[1]));
  }
  else if (!strcmp(szCommand[0], "l?"))
  {
    status = rssgs_GetLevel(instrumentHandle, &dValue);

    if (status == VI_SUCCESS)
      printf ("Level = %f dBm\n", dValue);
  }

  // Set / get level offset------------------------- 
  else if (!strcmp(szCommand[0], "lo"))
  {
    status = rssgs_SetLevelOffset(instrumentHandle, atof(szCommand[1]));
  }
  else if (!strcmp(szCommand[0], "lo?"))
  {
    status = rssgs_GetLevelOffset(instrumentHandle, &dValue);

    if (status == VI_SUCCESS)
      printf("Offset = %f dB\n", dValue);
  }

  // Set level and frequency ------------------------ 
  else if (!strcmp(szCommand[0], "fl"))
  {
    size_t len = 0;
    double value = 0;

    status = rssgs_SetTrigConnnectorOpMode(instrumentHandle, RSSGS_TRG_CON_OP_MODE_MEASLATENCY );
    rssgs_delay (instrumentHandle, 0.1);

    len = strlen(szCommand[1]);
    value = atof (szCommand[1]);
    if (len > 3)
    {
      if (!_stricmp(szCommand[1]+len-3, "mhz"))
        value *= 1e6;
      if (!_stricmp(szCommand[1]+len-3, "ghz"))
        value *= 1e9;
    }

    status |= rssgs_SetFrequencyAndLevel(instrumentHandle, value, atof(szCommand[2]) );
  }

  // Set / get RF state ------------------------- 
  else if (!strcmp(szCommand[0], "r"))
  {
    bState = !_stricmp (szCommand[1], "ON") || !_stricmp (szCommand[1],"1"); 

     status = rssgs_SetRFState(instrumentHandle, bState);
  }
  else if (!strcmp(szCommand[0], "r?"))
  {
    status = rssgs_GetRFState(instrumentHandle, &bState);

    if (status == VI_SUCCESS)
      printf ("RF State = %s\n", bState ? "ON" : "OFF");
  }

  // Get error staus ----------------------- 
  else if (!strcmp(szCommand[0], "e?"))
  {
    ViChar errorQueue[1000];
    rssgs_error_query (instrumentHandle, &errorCode, buffer);
    rssgs_SYSTemStaticErrorQueue (instrumentHandle, 1000, errorQueue);
    printf ("Error:\t %d, %s\n", (int) errorCode, buffer);
    printf ("Static:\t %s\n", errorQueue);
   }
  
  // Set / get IQ modulator state ------------------------- 
  else if (!strcmp(szCommand[0], "m"))
  {
    bState = !_stricmp (szCommand[1], "ON") || !_stricmp (szCommand[1],"1"); 
    status = rssgs_SetIQModulatorState(instrumentHandle, bState);
  }
  else if (!strcmp(szCommand[0], "m?"))
  {
    status = rssgs_GetIQModulatorState(instrumentHandle, &bState);

    if (status == VI_SUCCESS)
      printf ("IQ Modulator State = %s\n", bState ? "ON" : "OFF");
  }
  // Set / get IQ wideband state ------------------------- 
  else if (!strcmp(szCommand[0], "w"))
  {
    bState = !_stricmp (szCommand[1], "ON") || !_stricmp (szCommand[1],"1"); 
    status = rssgs_SetIQWideband(instrumentHandle, bState);
  }
  else if (!strcmp(szCommand[0], "w?"))
  {
    status = rssgs_GetIQWideband(instrumentHandle, &bState);
    if (status == VI_SUCCESS)
      printf ("IQ Wideband State = %s\n", bState ? "ON" : "OFF");
  }

  else if (!strcmp(szCommand[0], "i?"))
  {
    status = rssgs_IdentificationQuery (instrumentHandle, buffer);

    if (status == VI_SUCCESS)
      printf ("%s\n",buffer);
  }

  else if (!strcmp(szCommand[0], "af?"))
  {
    status = rssgs_CALibrationSynthesis (instrumentHandle, 60000, &nValue);
    if (status == VI_SUCCESS && nValue == 0)
      printf ("OK\n");
    else
      printf ("FAIL\n");
  }
  // Set / get fast mode ------------------------- 
  else if (!strcmp(szCommand[0], "fm"))
  {
    g_useFastSettings = !_stricmp (szCommand[1], "ON") || !_stricmp (szCommand[1],"1"); 
    status = rssgs_UseFastSettings (instrumentHandle, g_useFastSettings, g_enableAsynchronousFastSettings);

  }
  else if (!strcmp(szCommand[0], "fm?"))
  {
    ViBoolean useFastSettings = VI_FALSE, enableAsynchronousFastSettings = VI_FALSE;
    status = rssgs_GetFastSettings (instrumentHandle, &useFastSettings, &enableAsynchronousFastSettings);
    if (status)
      printf ("status = %d\n", status);
    printf ("Fast Mode = %s\n", useFastSettings ? "ON" : "OFF");
    if (useFastSettings)
      printf ("Async Settings = %s\n", enableAsynchronousFastSettings ? "ON" : "OFF");

  }

  else if (!strcmp(szCommand[0], "scpi"))
  {
    char szArg[1000] = {0};
    int i = 1;
    while (*szCommand[i])
    {
      strcat (szArg, szCommand[i]);
      strcat (szArg, " ");
      i++;
    }
    
    status = rssgs_writeInstrData (instrumentHandle, szArg);
  }
  else if (!strcmp(szCommand[0], "scpi?"))
  {
    char szArg[1000] = {0};
    int i = 1;
    ViInt32 nBytesRead = 0;
    while (*szCommand[i])
    {
      strcat (szArg, szCommand[i]);
      strcat (szArg, " ");
      i++;
    }
   
    status = rssgs_writeInstrData (instrumentHandle, szArg);
    status = rssgs_readInstrData (instrumentHandle, 1000, szArg, &nBytesRead);
    szArg[nBytesRead] = 0;

    printf ("%s", szArg);    
  }

  else if (!strcmp(szCommand[0], "ac1?") ||
	         !strcmp(szCommand[0], "ac2?") ||
	         !strcmp(szCommand[0], "ac3?") ||
	         !strcmp(szCommand[0], "ac4?") ||
	         !strcmp(szCommand[0], "ac5?") )
  {
    char szArg[1000] = {0};
    ViInt32 nBytesRead = 0;
    sprintf (szArg, "ext:send? 'diag:info:ecount%c?'",szCommand[0][2]);
    status = rssgs_writeInstrData (instrumentHandle, szArg);
    status = rssgs_readInstrData (instrumentHandle, 1000, szArg, &nBytesRead);
    szArg[nBytesRead] = 0;
    printf ("%s", szArg);    
  }



  else if (!strcmp(szCommand[0], "clr"))
  {
    status = rssgs_Clear(instrumentHandle);
  }  


  else
  {
    printf ("Unknown command\n");
  }

  rssgs_error_query (instrumentHandle, &errorCode, buffer);
  if (errorCode)
    printf ("%s\n", buffer);

  return status;
}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------


void WaitForEnterPressed()
{
  printf ("Press Enter ...");
  getchar();
}



// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

void TestDropout (char* szSerialSgs[], char* szSerialSgu[], int nNumPairs)
{
  #define MAXPAIRS 3
  char buffer[1000];
  unsigned int hdlSgs[MAXPAIRS];
  unsigned int hdlSgu[MAXPAIRS];
  ViBoolean idQuery = VI_FALSE;
  ViBoolean resetDevice = VI_FALSE;
  ViBoolean fail = VI_FALSE;
  int i, j, p, pair, errorCode;
  double f, fSGU, pSGU;


  for (pair = 0; pair < nNumPairs; pair++)
  {
    int status;
    printf("open SGS %s\n", szSerialSgs[pair]);
    sprintf (buffer, "PCIE::0x0162f::0x132e::%s::INSTR",szSerialSgs[pair]);
    status = rssgs_init(buffer, idQuery, resetDevice, hdlSgs + pair);
    if (status)
    {
      rssgs_error_message(hdlSgs[pair], status, buffer);
      printf("open FAILED status = 0x%x (%s)\n", (unsigned int)status, buffer);
      return;
    }

    rssgs_error_query(hdlSgs[pair], &errorCode, buffer);
    if (errorCode)
      printf("%d %s\n", (int)errorCode, buffer);

    rssgs_UseFastSettings(hdlSgs[pair], 1, 0);

    if (!HasActiveExtension (hdlSgs[pair], VI_FALSE))
    {
      printf("SGS %s is unpaired\n", szSerialSgs[pair]);
      return;
    }



    printf("open SGU %s\n", szSerialSgu[pair]);
    sprintf (buffer, "PCIE::0x0162f::0x132f::%s::INSTR",szSerialSgu[pair]);
    status = rssgs_init(buffer, idQuery, resetDevice, hdlSgu + pair);
    if (status)
    {
      rssgs_error_message(hdlSgu[pair], status, buffer);
      printf("open FAILED status = 0x%x (%s)\n", (unsigned int)status, buffer);
      return;
    }
  }
//  status = rssgs_SetRFState(hdlSgs, VI_TRUE);



  f = 10e9;
  j = 1;
  #define PSTART -8
  #define PSTOP  18
  for (i=0;  ; i++)
  {
    for (p=PSTART; p <= PSTOP && !fail ; p++)
    {
      for (pair = 0; pair < nNumPairs && !fail; pair++)
      {
        rssgs_SetFrequencyAndLevel(hdlSgs[pair], f, p);
//        rssgs_FastSync(hdlSgs[pair]);
      }

      for (pair = 0; pair < nNumPairs && !fail; pair++)
      {
        Sleep (1);
        rssgs_GetFrequency(hdlSgu[pair], &fSGU);
        rssgs_GetLevel(hdlSgu[pair], &pSGU);
        printf ("%5d %4.0f %4.0f %d %d\n",j, f/1e6, fSGU/1e6, p, (int)pSGU);

        if (f != fSGU || p != (int)pSGU )
        {
          printf ("Error\n");
          fail = VI_TRUE;
          break;
        }
      if (fail)
        break;
      }
      j++;
    }
    if (fail)
      break;

    f += 1e6;
    if (f > 15e9)
      f = 10e9;
  }


  for (pair = 0; pair < nNumPairs; pair++)
  {
    rssgs_close(hdlSgs[pair]);
    rssgs_close(hdlSgu[pair]);
  }

}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

#ifdef WITHPOWERSENSOR

void TestDropoutSensor (char* szSerialSgs, char* szSerialSgu, char* szRsrcSensor)
{
  char buffer[1000];
  unsigned int hdlSgs = 0;
  unsigned int hdlSgu = 0;
  ViBoolean idQuery = VI_FALSE, readbackError;
  ViBoolean resetDevice = VI_FALSE;
  ViBoolean fail = VI_FALSE;
  int i, j,  r ;
  double f, p, fSGU, pSGU;
  time_t StartTime, EndTime;

  // Power sensor
  ViSession iHandle;
  ViStatus st;

  printf("open Sensor %s\n", szRsrcSensor);

  // Initialization
  // e.g. "USB::0x0aad::0x0023::100034"
  st = rsnrpz_init (szRsrcSensor, 1, 1, &iHandle);
  if (st != VI_SUCCESS)
  {
      printf ("Init status: %x\n", st);
      return;
  }

  // Setup - continuous average of 5 values, 1GHz carrier, immediate trigger
  rsnrpz_chan_mode (iHandle, 1, RSNRPZ_SENSOR_MODE_CONTAV);
  rsnrpz_chan_setContAvAperture(iHandle, 1, 10e-6);
  rsnrpz_trigger_setSource (iHandle, 1, RSNRPZ_TRIGGER_SOURCE_IMMEDIATE);
  rsnrpz_avg_configureAvgManual (iHandle, 1, 1);

  printf("open SGS %s\n", szSerialSgs);
  sprintf (buffer, "PCIE::0x0162f::0x132e::%s::INSTR",szSerialSgs);

  int status = rssgs_init(buffer, idQuery, resetDevice, &hdlSgs);
  if (status)
  {
    rssgs_error_message(hdlSgs, status, buffer);
    printf("open FAILED status = 0x%x (%s)\n", (unsigned int)status, buffer);
    return;
  }

  if (!HasActiveExtension (hdlSgs), VI_FALSE)
  {
    printf("SGS is unpaired\n");
    return;
  }


  printf("open SGU %s\n", szSerialSgu);
  sprintf (buffer, "PCIE::0x0162f::0x132f::%s::INSTR",szSerialSgu);
  status = rssgs_init(buffer, idQuery, resetDevice, &hdlSgu);
  if (status)
  {
    rssgs_error_message(hdlSgu, status, buffer);
    printf("open FAILED status = 0x%x (%s)\n", (unsigned int)status, buffer);
    return;
  }

  rssgs_UseFastSettings(hdlSgs, VI_TRUE, VI_TRUE);
  rssgs_SetRFState(hdlSgs, VI_TRUE);

//  status = rssgs_SetRFState(hdlSgs, VI_TRUE);

  time(&StartTime);


  float totalTime;
  float loopTime;
  #define FSTART 10.7e9
  #define FEND  14.7e9
  #define FSTEP 50e6
  #define ATTCABLE 7 // splitter
  #define RETRIES 20
  #define ERRORMARGIN 5 // dB

  f = FSTART;
  j = 1; // number of settings
  for (i=0;  !fail ; i++)
  {
    rsnrpz_chan_setCorrectionFrequency (iHandle, 1, f);

    for (p=-28; p <= 10 && !fail ; p += 0.1)
    {
      rssgs_SetFrequencyAndLevel(hdlSgs, f, p);
//      rssgs_FastSync(hdlSgs);

//      usleep (20000);

      // Start measurement
      double result = -110;
      for (r = 0; r < RETRIES; r++)
      {
        rsnrpz_chans_initiate (iHandle);

        // Wait for completion
        ViBoolean bMeasCompleted;
        do
        {
            st = rsnrpz_chan_isMeasurementComplete(iHandle, 1, &bMeasCompleted);
            if (bMeasCompleted || st != VI_SUCCESS)
              break;
            usleep(2000);
        } while (VI_TRUE);

        // Read result
        rsnrpz_meass_fetchMeasurement(iHandle, 1, &result);

        result = (result < 1e-14) ? -110.0 : 10.0*log10(result)+30.0 + ATTCABLE;
        if (fabs(result-p)  < ERRORMARGIN)
          break;
      }
      rssgs_GetFrequency(hdlSgu, &fSGU);
      rssgs_GetLevel(hdlSgu, &pSGU);
      readbackError = fabs(f - fSGU) > 1e-3 || fabs(p - pSGU) > 1e-2;

      printf ("%5d %5d %4.0f %4.0f %5.1f %5.1f\t %5.2f \t%s \n",j, r,f/1e6, fSGU/1e6, p, pSGU, result, readbackError ? "SGU No Match" : "" );


      if (r >= RETRIES || readbackError)
      {
        //totalTime = stop_timer ();
        time(&EndTime);
        totalTime = difftime(EndTime, StartTime);
        loopTime = totalTime / j;
        printf ("Error: measured power incorrect or readback error\n");
        printf ("Total Time : %f sec\n", totalTime);
        printf ("Average Time between settings : %f sec\n", loopTime);
        fail = VI_TRUE;
        break;
      }
      j++; // increment # of settings
    }

    if (fail)
    {
      rsnrpz_close (iHandle);
      break;
    }

    f += FSTEP;
    if (f > FEND)
      f = FSTART;
  } // for (i=0; i < 100000; i++)

  //totalTime = stop_timer ();
  time(&EndTime);
  totalTime = difftime(EndTime, StartTime);
  loopTime = totalTime / j;
  printf ("End of test loop. Number of Settings : %d\n", j);
  printf ("Total Time : %f sec\n", totalTime);
  printf ("Average Time between settings : %f sec\n", loopTime);

  rsnrpz_close (iHandle);

} // void TestDropoutSensor


#endif


int startTestDropout(int argc, char* argv[])
{
  if (argc==4 || argc==6 || argc==8)
  {
    int i;
    char* szRsrcSgs[3];
    char* szRsrcSgu[3];
    int numPairs = (argc-1) / 2;

    for (i = 0; i < numPairs; i++)
    {
      szRsrcSgs[i] = argv[2*i+2];
      szRsrcSgu[i] = argv[2*i+3];
    }
    TestDropout(szRsrcSgs, szRsrcSgu, numPairs);
  }
  else
  {
    printf ("Usage: -td SerialSGS1 SerialSGU1 [SerialSGS2 SerialSGU2 [SerialSGS3 SerialSGU3]]");
  }
  return 0;
}


int startTestDropoutSensor(int argc, char* argv[])
{
  if (argc == 5)
  {
#ifdef WITHPOWERSENSOR
    TestDropoutSensor(argv[2], argv[3], argv[4]);
#else
    printf("Compile with power sensor support to use this function!\n");
#endif

  }
  else
  {
    printf ("Usage: -tds SerialSGS SerialSGU ResourceSensor");
  }
  return 0;
}

int startShowHelp(int argc, char* argv[])
{
  fprintf(stderr,"Usage: SgsDriverDemo resourcestring [cmd arg]\n\n" \
    "resourcestring:\n" \
    "\tTCPIP::ipaddress::[hislip0::]INSTR\n" \
    "\tTCPIP::ipaddress::5025::SOCKET\n" \
    "\tUSB::0x0aad::0x0088::serialno::INSTR\n" \
    "\tPCIE::0x0162f::0x132e::serialno::INSTR\n" \
    "\t-td SerialSGS1 SerialSGU1 [SerialSGS2 SerialSGU2 [SerialSGS3 SerialSGU3]]\n" \
    "\t-tds SerialSGS SerialSGU ResourceSensor\n" \
    "\ttcpip\n" \
    "\tpcie\n" \
  );

  WaitForEnterPressed();
  return 1;
}



// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------
int startCommandLoop(int argc, char* argv[])
{
  char buffer[1000] = { 0 };
  ViStatus status = 0;
  int i;

  InitControlSerial();
  INIT_SERIAL();
  strcpy(buffer, argv[1]);

  for (i = 0; i < MaxNumSGS; i++)
  {
    g_instrumentHandle[i] = 0;
    g_ExtensionIndexOfSguSerialIndex[i][0] = 0;
  }

  for (i = 0; buffer[i]; i++)
    buffer[i] = (char)tolower(buffer[i]);

#ifndef __linux__

  if (strstr(buffer, "::socket") || !strcmp(buffer, "tcpip"))
  {
    if (!StartNdis())
      g_useFastSettings = FALSE;
  }
#endif


  if (argc == 2 && !strcmp(buffer, "pcie"))
  {
    g_hwChannel = PCIE;
    g_nSgsCount = g_nSguCount = 10;

    status = rssgs_enumerate_pcie_devices(0x162f, 0x132e, g_SgsSerialNumbers, &g_nSgsCount);
    if (status >= VI_SUCCESS)
      status = rssgs_enumerate_pcie_devices(0x162f, 0x132f, g_SguSerialNumbers, &g_nSguCount);
    OpenDevices(NULL);
  }
  else if (argc == 2 && !strcmp(buffer, "tcpip"))
  {
    int i = 0;
    ViInt32 infoCount = 0;
    typedef char INFOTEXT[64]; // reserve 64 chars for IP address, type , serial no

    g_hwChannel = SOCK;
    g_nSgsCount = g_nSguCount = 0;

                               // get number of devices first
    status = rssgs_enumerate_socket_devices(NULL, &infoCount);
    if (infoCount > 0)
    {
      // allocate enough space for device info texts
      INFOTEXT** infoList = calloc(infoCount, sizeof(INFOTEXT*));
      for (i = 0; i < infoCount; i++)
        infoList[i] = calloc(1, sizeof(INFOTEXT));

      // get device info 
      status = rssgs_enumerate_socket_devices((ViPString *)infoList, &infoCount);
      for (i = 0; i < infoCount; i++)
      {
        sprintf(buffer, "%s", (char*)infoList[i]);
        buffer[14] = 0;

        if (!strncmp(buffer, "SGS100A", 7))
          g_SgsSerialNumbers[g_nSgsCount++] = atoi(buffer + 8);
        else if (!strncmp(buffer, "SGU100A", 7))
          g_SguSerialNumbers[g_nSguCount++] = atoi(buffer + 8);
//        printf("%s\n", buffer);
      }

      for (i = 0; i < infoCount; i++)
        free(infoList[i]);
      free(infoList);

    }
    OpenDevices(NULL);
  }
  else 
  {
    OpenDevices(buffer);
  }

  if (status >= VI_SUCCESS && g_nSgsCount > 1)
  {
    int nSelection = -1;
    int i = 0;

    if (g_nSguCount)
    {
      printf("Found %d SGU device(s):\n", (int)g_nSguCount);


      for (i = 0; i < g_nSguCount; i++)
      {
        printf("SGU-%d\n", (int)g_SguSerialNumbers[i]);
      }
      printf("\n\n");
    }

    printf("Found %d SGS device(s):\n", (int)g_nSgsCount);

    for (i = 0; i < g_nSgsCount; i++)
    {
      printf("%d\tSGS-%d\n", i+1, (int)g_SgsSerialNumbers[i]);
    }
    printf("\n");

    if (g_nSgsCount >1)
    {
      int c = 0;
      printf("Select SGS (1...%d or hit ESC) ", (int)g_nSgsCount);
      do
      {
        c = _getch();
        if (c - '0' >= 1 && c - '0' <= (int)g_nSgsCount)
        {
          nSelection = c - '0' - 1;
          printf("%c\n", c);
          break;
        }
      } while (c != 27); // ESC
      if (nSelection < 0)
        return 0;
    }
    else
    {
      nSelection = 0;
    }
    g_activeDevice = nSelection;

    if (g_nSguCount)
    {
      int activeExtension = HasActiveExtension (g_instrumentHandle[g_activeDevice], VI_FALSE);
      printf("Active device is SGS-%d (%s)\n", (int)g_SgsSerialNumbers[g_activeDevice], activeExtension ? "paired" : "unpaired");
    }
    else
    {
      printf("Active device is SGS-%d\n", (int)g_SgsSerialNumbers[g_activeDevice]);
    }

  }

  
  CommandLoop();

  CloseDevices();

  printf("\n");

#ifndef __linux__
  if (argc == 2)
    WaitForEnterPressed();
#endif

  CLOSE_SERIAL();

  return 0;
}


// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

int main(int argc, char* argv[])
{
  if (argc >= 2 && !strcmp(argv[1],"-td"))
    return startTestDropout (argc, argv);

  if (argc >= 2 && !strcmp(argv[1],"-tds"))
    return startTestDropoutSensor (argc, argv);


  if (argc < 2 || (argc == 2 && !strcmp(argv[1], "?")))
    return startShowHelp (argc, argv);

  return startCommandLoop(argc, argv);
}



// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

// Enter commands on the command line
void CommandLoop()
{
  char buffer[1000];
  char lastbuffer[1000];
  COMMAND szCommand;
  int i = 0;
  int c = 0;
  ViStatus status = VI_SUCCESS;
    



  do 
  {
    status = VI_SUCCESS;

    c = 0;
    for (i = 0; i < MAXCMDSIZE; i++)
      szCommand[i][0] = 0;

    printf ("Enter command >");
    for (i = 0; i < 1000-1 && c != '\n' && (i>0 || c != 'p'); ) 
    {
      c = getchar();
      if (c != '\n')
      {
        buffer[i++] = (char)c;
      }
      buffer[i] = 0;
    }
    if (!strcmp(buffer,"p"))
      strcpy (buffer, lastbuffer);
    else
      strcpy (lastbuffer, buffer);

    sscanf (buffer,"%s%s%s",(char*)(szCommand+0), (char*)(szCommand+1), (char*)(szCommand+2));

    if (!strcmp(szCommand[0], "q"))
    {
      break;
    }
    else
    {
      status = InterpretCommand(g_instrumentHandle[g_activeDevice], szCommand);
      if (status)
        printf ("%s (0x%x)\n", DecodeVisaError (g_instrumentHandle[g_activeDevice],status), (unsigned int)status);
    }



  } while (1);
}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

const char* DecodeVisaError (ViSession instrumentHandle, ViStatus i_stat)
{
  static char buffer[1000];
  ViStatus status = VI_SUCCESS;
  *buffer = 0;

  if (i_stat == VI_SUCCESS)
    return "VI_SUCCESS";

  status = rssgs_error_message (instrumentHandle, i_stat, buffer);
  if (status != VI_SUCCESS)
  {
    sprintf (buffer, "Unknown Visa Error 0x%x", (unsigned int)i_stat);
  }
  return buffer;
}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

void TestExecutionTime(int instrumentHandle)
{
  const int nloops = 1000;
  double dTInstructionAvg = 0;
  double dTInstructionMax = 0;
  double dTInstruction = 0;
  int i = 0;
  double dFreq;

  int* timearray = calloc(nloops, sizeof(int));
  if (timearray == 0)
  {
    printf("Cannot allocate array\n");
    exit(1);
  }

  determineProcessorClockrate();

  for (i = 0; i < nloops; i++)
  {
    dFreq = 100e6 + i * 1e3;

    getClockInMicroSeconds(0);

    //status = rssgs_SetRFState(instrumentHandle, i%1);
    rssgs_SetFrequencyAndLevel(instrumentHandle, dFreq, -10);
    //    if (useFastSettings == VI_FALSE)


    rssgs_FastSync(instrumentHandle);
    dTInstruction = getClockInMicroSeconds(1);

    timearray[i] = (int)dTInstruction;
    dTInstructionAvg += dTInstruction;
    if (dTInstruction > dTInstructionMax)
      dTInstructionMax = dTInstruction;
  }

  dTInstructionAvg /= nloops;
  printf("Avg. Roundtrip Time (Frequency) = %5.0f us, Max = %5.0f us\n", dTInstructionAvg, dTInstructionMax);

  {
    const char* szFileName = "executiontimes.csv";
    FILE* f = fopen(szFileName, "w");
    if (f)
    {
      for (i = 0; i < nloops; i++)
        fprintf(f, "%d\n", timearray[i]);
      fclose(f);
      printf("Saved execution time array to %s\n", szFileName);
    }

  }
  free(timearray);

}


// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------

void TestExecutionTime2(int instrumentHandle)
{
  const int nloops = 2000;
  double dTInstructionAvg = 0;
  double dTInstructionMax = 0;
  double dTInstruction = 0;
  int i = 0;
  const double dFmin = 1e9;
  const double dFmax = 40e9;
  double dFreq = dFmin;


  determineProcessorClockrate();

  rssgs_SetRFState(instrumentHandle, VI_TRUE);

  for (i = 0; i < nloops; i++)
  {

    getClockInMicroSeconds(0);

    rssgs_ConfigureIQ(instrumentHandle, VI_TRUE, VI_TRUE);
    rssgs_FastSync(instrumentHandle);

    rssgs_SetFrequencyAndLevel(instrumentHandle, dFreq, -16.735022);
    rssgs_FastSync(instrumentHandle);

    // other commands

    rssgs_SetFrequencyAndLevel(instrumentHandle, dFreq, -16.735022);
    rssgs_FastSync(instrumentHandle);

    rssgs_ConfigureIQ(instrumentHandle, VI_FALSE, VI_FALSE);
    rssgs_FastSync(instrumentHandle);


    dTInstruction = getClockInMicroSeconds(1);
    //    if (dTInstruction > 100000)
    printf("%5.2f GHz %5.0f ms\n", dFreq / 1e9, dTInstruction / 1000);

    dTInstructionAvg += dTInstruction;
    if (dTInstruction > dTInstructionMax)
      dTInstructionMax = dTInstruction;

    dFreq += 100e6;
    if (dFreq > dFmax)
      dFreq = dFmin;


  }
  dTInstructionAvg /= nloops;
  printf("Avg. Roundtrip Time = %5.0f us, Max = %5.0f us\n", dTInstructionAvg, dTInstructionMax);

}



// #################################################################


// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------
BOOL determineProcessorClockrate()
{
  if (g_dProcessorClockrate < 0)
  {

#ifdef __linux__
    // Read clockrate from /proc
# ifdef __powerpc__
    const char* tag[] = { "clock" };
    std::ifstream f;
    f.open("/proc/cpuinfo");
    if (f.is_open()) for (std::string str; !f.eof();)
    {
      f >> str;
      if (str == tag[0])
      {
        f >> str;
        f >> str;
        g_dProcessorClockrate = 1E6 * atof(str.c_str());
        break;
      }
    }
#else
    const char* tag[] = { "cpu", "MHz" };
    FILE* file = fopen("/proc/cpuinfo", "r");
    if (file)
    {
      char str[1000];
      int rc = 0;
      do
      {
        rc = fscanf(file, "%s", str);
        if (rc != 1)
          break;
        if (!strcmp(str, tag[0]))
        {
/*          __int64*/
          rc = fscanf(file, "%s", str);
          if (!strcmp(str, tag[1]))
          {
            rc = fscanf(file, "%s", str);
            rc = fscanf(file, "%s", str);
            g_dProcessorClockrate = 1E6 * atof(str);

            break;
          }
        }
      } while (rc == 1);
    } // file
#endif // !power PC
#else
    // Read clockrate from registry
    HKEY hKey;
    if (ERROR_SUCCESS == RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0", 0, KEY_READ, &hKey)) //lint !e1924
    {
      DWORD valType;
      DWORD valSize = sizeof(DWORD);
      DWORD val;
      const BOOL ok = (ERROR_SUCCESS == RegQueryValueEx(hKey, L"~MHz", 0, &valType, (BYTE*)&val, &valSize)); //lint !e1924
      RegCloseKey(hKey);
      if (ok && REG_DWORD == valType)
      {
        g_dProcessorClockrate = 1E6 * val;
      }
    }
#endif // Windows
  }
  // Plausibility test
  return g_dProcessorClockrate > 1E7;
}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------
double getProcessorClockrate()
{
  return g_dProcessorClockrate;
}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------
#ifndef __linux__
#  pragma warning(disable:4035)
#endif
__int64 readProcessorClock()
{
#ifdef __linux__
# ifdef __powerpc__
  unsigned long int upper, lower, tmp;
  __asm__ __volatile__(
    "1:\n\t"
    "mftbu %0\n\t"
    "mftb  %1\n\t"
    "mftbu %2\n\t"
    "cmpw  %2,%0\n\t"
    "bne   1b\n"
    : "=r"(upper), "=r"(lower), "=r"(tmp)
  );
  unsigned long long result(upper);
  return ((result << 32) | lower) * 6;
# else
  __int64 val; __asm__ __volatile__("rdtsc" : "=&A" (val)); return val;
# endif
#else
  return __rdtsc();
  //   __asm { RDTSC }
#endif
}//lint !e533 : method returns a value, but the this is made in assembler


 // ---------------------------------------------------------------------
 //
 // ---------------------------------------------------------------------
double readSeconds()
{
  return readProcessorClock() / g_dProcessorClockrate;
}

// ---------------------------------------------------------------------
//
// ---------------------------------------------------------------------
double getClockInMicroSeconds(BOOL bIsToCalculate)
{
  static long long sliClockTicksActual = 0;
  if (bIsToCalculate == 1)
  {
    long long liClockTicksPrev = sliClockTicksActual;
    sliClockTicksActual = readProcessorClock();
    if (sliClockTicksActual < liClockTicksPrev)
    {
      //case with type long long overflow
      return ((double)((LLONG_MAX - liClockTicksPrev + 1 + sliClockTicksActual)) /
        (getProcessorClockrate() /*[Hz]*/ * 1.e-6 /*[micro-seconds]*/));
    }
    else
    {
      //normal case
      return ((double)((sliClockTicksActual - liClockTicksPrev)) /
        (getProcessorClockrate() /*[Hz]*/ * 1.e-6) /*[micro-seconds]*/);
    }
  }
  else
  {
    //initial case
    sliClockTicksActual = readProcessorClock();
    return (0. /*[micro-seconds]*/);
  }
}

// ---------------------------------------------------------------------
// Functions for controlling the pins of a serial interface
// using an external DLL
// ---------------------------------------------------------------------

void InitControlSerial()
{
#ifndef __linux__

  void* handle = LoadLibraryA("ControlSerial.dll");
  if (handle)
  {
    SerialOpen = (FuncSerialOpen)GetProcAddress(handle, "SerialOpen");
    SerialClose = (FuncSerialClose)GetProcAddress(handle, "SerialClose");
    SerialSetRTS = (FuncSerialSetRTS)GetProcAddress(handle, "SerialSetRTS");
    SerialSetDTR = (FuncSerialSetDTR)GetProcAddress(handle, "SerialSetDTR");
  }
#endif
}

void SET_RTS()
{
  if (SerialSetRTS)
    SerialSetRTS(SerialHandle, 1);
}

void CLR_RTS()
{
  if (SerialSetRTS)
    SerialSetRTS(SerialHandle, 0);
}

void SET_DTR()
{
  if (SerialSetDTR)
    SerialSetDTR(SerialHandle, 1);
}

void CLR_DTR()
{
  if (SerialSetDTR)
    SerialSetDTR(SerialHandle, 0);
}

void INIT_SERIAL()
{
  if (SerialOpen)
    SerialHandle = SerialOpen("COM1");
}

void CLOSE_SERIAL()
{
  if (SerialClose)
    SerialClose(SerialHandle);
}




