/*= R&S SGS Signal Generator ================================================
 * LabWindows/CVI 9.1.0 VXIpnp Instrument Driver
 * Original Release: September 2010
 * By: Petr Kominek, Jiri Kominek
 *
 * Should you have any technical questions please contact the hotline of
 * Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 * e-mail: CustomerSupport@rohde-schwarz.com
 *
 *   see ChangeLog
 *
*===========================================================================*/
#ifdef _WIN32
        #include <windows.h>
#include "rssgs.h"
#include "rssgs.h"
#endif
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include "rssgs.h"
#include "rssgs_utility.h"
#include <limits.h>
#include <locale.h>  
#include "visaabstraction.h"
#include "SgmaCommandIds.h"

/*= Macros ==================================================================*/

#define CHECKERR_CU(fCal) if (rssgs_status = (fCal), rssgs_status < VI_SUCCESS) \
    return rssgs_initCleanUp (rmSession, instrSession, rssgs_status); else

/*= DEFINES =================================================================*/
#define RSSGS_REVISION          "Rev 1.6.3, 04/2018, CVI 2017" /* Instrument driver revision */
#define BUFFER_SIZE             2048L    /* File I/O buffer size */

#define RSSGS_DEFAULT_TIMEOUT   10000   /* Default Timeout value in ms (10 sec) */

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT COMMAND ARRAYS =====================================*/
/*****************************************************************************/
static ViString instrVersionList[] = {"SGS", VI_NULL};
static ViString operationModeArr[] = {"NORM","BBBY", VI_NULL};
static ViString FrequencyLocalOscillatorSourceArr[] = {"INT", "EXT", VI_NULL};
static ViString FrequencyLocalOscillatorOutputArr[] = {"REF", "LO", "OFF", VI_NULL};
static ViString ROSCSourceArr[] = {"INT", "EXT" , VI_NULL};
static ViString ROSCExternalReferenceFrequencyArr[] = {"10MHZ", "100MHZ", "1000MHZ", VI_NULL};
static ViString ROSCExternalSyncBandwidthArr[] = {"WIDE", "NARR", VI_NULL};
static ViString LevelModeArr[] = {"NORM", "LOWN", "LOWD", VI_NULL};
static ViString LevelSettingCharacteristicArr[] = {"AUTO", "UNIN", "MON", "CVSW", "USER", VI_NULL};
static ViString LevelPowerOnStateArr[] = {"OFF", "UNCH", VI_NULL};
static ViString LevelALCStateArr[] = {"ON", "OFFT", "ONT", VI_NULL};
static ViString LevelALCDSensitivityArr[] = {"OFF", "LOW", "MED", "HIGH", VI_NULL};
//static ViString RemoteOperationModeArr[] = {"EXT", "STD", VI_NULL};
static ViString AngleUnitsArr[] = {"DEG", "RAD", VI_NULL};
static ViString PowerUnitsArr[] = {"V", "DBM", VI_NULL};
static ViString ConfigureNetworkIPArr[] = {"AUTO", "STAT", VI_NULL};
static ViString ROSCreferenceFrequencyArr[] = {"10MHZ","100MHZ","1000MHZ", VI_NULL};
static ViString attenuatorAfterArr[] = {"FIX","MAX", VI_NULL};
static ViString attenuatorModeArr[] = {"AUTO","FIX", "APAS", VI_NULL};
static ViString terminatorModeArr[] = {"EOI","STAN", VI_NULL};
static ViString formatDataArr[] = {"ASC","PACK", VI_NULL};
static ViString regGroupArr[] = {"OPER","QUES", VI_NULL};
static ViString regTypeArr[] = {"COND","EVEN","ENAB","NTR","PTR", VI_NULL};
static ViString testResult[] = {"0", "1","RUN","STOP", VI_NULL};
static ViString trigConOpModeArr[] = {"IN", "LOW", "MLAT", "MARR", "HIGH", "FBF", "SVAL", "EO1", "EO2", "BLAN", "SNV", "PEMS", "PVO", "PETR", VI_NULL};
static ViString mmacBarArr[] = {"SIGMACOM", "RF", VI_NULL};
static ViString ecoModeArr[] = {"OFF", "EM1", VI_NULL};
static ViString ExtensionHardwareChannelArr[] = {"LAN", "USB", "PCIE", "SOCK", VI_NULL};
static ViString pulmSourceArr[] = {"INT","EXT", VI_NULL};
static ViString pulmTriggerModeArr[] = {"AUTO","EXT","EGAT", VI_NULL};
static ViString polarityArr[] = {"NORM","INV", VI_NULL};
static ViString dmemExternalInputImpedance[] = {"G50","G10K", VI_NULL};
static ViString slopeArr[] = {"POS","NEG", VI_NULL};
static ViString PULMModeArr[] = {"DOUB","SING", VI_NULL};
static ViString SerialBaudRateArr[] = {"2400", "4800", "9600", "19200", "38400", "57600", "115200", VI_NULL};
static ViString SerialParityArr[] = {"NONE", "ODD", "EVEN", VI_NULL};
static ViString SerialStopBitsArr[] = {"1", "2", VI_NULL};
static ViString CalRoscDataModeArr[] = {"FACT", "CUST", VI_NULL};
static ViString dexcListModeArr[] = {"IMP","EXP", VI_NULL};
static ViString dexcExtensionArr[] = {"TXT","CSV", VI_NULL};
static ViString dexcSeparatorArr[] = {"TAB","SEM","COMM","SPAC", VI_NULL};
static ViString dexcDecimalSeparatorArr[] = {"DOT","COMM", VI_NULL};
static ViString statusRegisterDataArr[] = {"ASC","BIN", "HEX","OCT", VI_NULL};
static ViString byteOrderArr[] = {"NORM","SWAP", VI_NULL};
static ViString powerSourceArr[] = {"A","USER","B", VI_NULL};
static ViString powerPriorityArr[] = {"AVER","PEAK", VI_NULL};
static ViString sensePowerUnitArr[] = {"DBM","DBUV","WATT", VI_NULL};
static ViString powerFilterModeArr[] = {"AUTO","USER","NSR", VI_NULL};
/*****************************************************************************/
/*= UTILITY ROUTINE DECLARATIONS (Non-Exportable Functions) =================*/
/*****************************************************************************/
ViStatus rssgs_initCleanUp (ViSession openRMSession, ViPSession openInstrSession,
                             ViStatus currentStatus);
ViStatus rssgs_defaultInstrSetup (ViSession openInstrSession);




/*****************************************************************************
 * Function: Initialize
 * Purpose:  This function performs the following initialization actions:
 *           
 *           - Opens a session to the Default Resource Manager resource and a 
 *           session to the specified device using the interface and address 
 *           specified in the Resource_Name control.
 *           
 *           - Performs an identification query on the Instrument.
 *           
 *           - Resets the instrument to a known state.
 *           
 *           - Sends initialization commands to the instrument that set any 
 *           necessary programmatic variables such as Headers Off, Short 
 *           Command form, and Data Transf er Binary to the state necessary for 
 *           the operation of the instrument driver.
 *           
 *           - Returns an Instrument Handle which is used to differentiate 
 *           between different sessions of this instrument driver.
 *           
 *           - Each time this function is invoked a Unique Session is opened.  
 *           It is possible to have more than one session open for the same 
 *           resource.
 *           
 *****************************************************************************/

ViStatus _VI_FUNC rssgs_init(
    ViRsrc  resourceName,
    ViBoolean   idQuery,
    ViBoolean   resetDevice,
    ViSession*  instrSession
)
{
  ViStatus            rssgs_status = VI_SUCCESS;
  ViSession           rmSession = 0;
  ViChar              buffer[BUFFER_SIZE] = "";
  rssgs_instrRange    instrPtr = NULL;
  unsigned int        ptrHandle;

  *instrSession = 0;
  /* --- Set standard "C" locale --- */
  setlocale(LC_ALL, "C");

  /*- Open instrument session ---------------------------------------------*/
  CHECKERR( rsOpenDefaultRM (&rmSession));


  if ((rssgs_status = rsOpen (rmSession, resourceName, VI_LOAD_CONFIG,
    VI_NULL, instrSession)) < 0)
  {
    (void)rsClose (rmSession);
    return rssgs_status;
  }

  /* Determine if the structure has been initialized for the current VISA  */
  /* Session and malloc if it has not.                                     */

  instrPtr = (rssgs_instrRange)(calloc (1, sizeof (struct rssgs_statusDataRanges)));
  ptrHandle = rssgs_AddPointer(instrPtr);       


  CHECKERR( rsSetAttribute (*instrSession, VI_ATTR_USER_DATA, ptrHandle));

  //With remote control via Ethernet, remote control is not automatically
  //set by means of a command. The instrument must be explicitly set to
  //the REMOTE state, e.g. by sending the intf command &GTR
  CHECKERR (rsPrintf (*instrSession, "&GTR\n"));

  /*- Configure VISA Formatted I/O ----------------------------------------*/
  CHECKERR( rsSetAttribute (*instrSession, VI_ATTR_TMO_VALUE,
    RSSGS_DEFAULT_TIMEOUT));
  CHECKERR( rsSetBuf (*instrSession, (ViUInt16) (VI_READ_BUF|VI_WRITE_BUF), 4000));
  CHECKERR( rsSetAttribute (*instrSession, VI_ATTR_WR_BUF_OPER_MODE,
    VI_FLUSH_ON_ACCESS));
  CHECKERR( rsSetAttribute (*instrSession, VI_ATTR_RD_BUF_OPER_MODE,
    VI_FLUSH_ON_ACCESS));

  /* -------------------------------------------------------------------- */
  /* Drawback of Agilent VISA is patched by adding extra termination      */
  /* character "\n" to each I/O operation                                 */
  /*                                                                      */
  /* VI_ATTR_SEND_END_EN will stay with this attribute set to VI_TRUE     */
  /* -------------------------------------------------------------------- */

  CHECKERR( rsSetAttribute (*instrSession, VI_ATTR_SEND_END_EN, VI_TRUE));

  // read/reset ESR which may still indicate device dependent error from 
  // previous handling!
  (void) rssgs_checkStatus(*instrSession);

  /*- Identification Query ------------------------------------------------*/
  if (idQuery == VI_TRUE)
  {
    int i = 0;
    CHECKERR( rsQueryf (*instrSession, "*IDN?\n", "%*[^,],%[^,],%*[\r\n]",
      buffer));


    /* Check if instrument is supported or not */
    while (instrVersionList[i] != NULL)
    {
      if (strncmp (buffer, instrVersionList[i], strlen (instrVersionList[i])) == 0)
        break;
      else i++;
    }
    if (instrVersionList[i] == NULL)
      return rssgs_initCleanUp (rmSession, instrSession, VI_ERROR_FAIL_ID_QUERY);
    else
      instrPtr->instrVer = i;
  }

  /* Check for instruments options */
  if ((rssgs_status = rssgs_OptionIdentificationQuery (*instrSession, buffer)) < VI_SUCCESS)
    return rssgs_status;
  else
    rssgs_status = VI_SUCCESS;

  instrPtr->instrOptions.B1 =     (strstr (buffer, "SGS-B1") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.B26 =    (strstr (buffer, "SGS-B26") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.B106 =   (strstr (buffer, "SGS-B106") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.B106V =  (strstr (buffer, "SGS-B106V") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.B112 =   (strstr (buffer, "SGS-B112") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.B112V =  (strstr (buffer, "SGS-B112V") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.K22 =    (strstr (buffer, "SGS-K22") != VI_NULL) ? 1 : 0;
  instrPtr->instrOptions.K90 =    (strstr (buffer, "SGS-K90") != VI_NULL) ? 1 : 0;

  /*- Reset instrument ----------------------------------------------------*/
  if (resetDevice == VI_TRUE)
  {
    CHECKERR( rssgs_reset (*instrSession));
  }
  else  /*- Send Default Instrument Setup ---------------------------------*/
  {
    CHECKERR( rssgs_defaultInstrSetup (*instrSession));
  }
  return rssgs_status;
}

/// HIFN This function enables and disables special PCI express features.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR enableFast/Activates/deactivates the fast mode.
/// HIPAR enableAsynchronous/Activates/deactivates the asynchronous mode.
ViStatus _VI_FUNC rssgs_UseFastSettings (ViSession instrSession,ViBoolean enableFast, ViBoolean enableAsynchronous)
{
  ViStatus rssgs_status = rsSetAttribute  (instrSession, SGS_ATTR_FAST, enableFast);
  if (rssgs_status == VI_SUCCESS && enableFast)
    rssgs_status = rsSetAttribute  (instrSession, SGS_ATTR_FASTASYNCHRONOUS, enableAsynchronous);

  if (rssgs_status == VI_SUCCESS)
    CHECKERR( rssgs_checkStatus (instrSession));

  return rssgs_status;
}

/// HIFN This function returns the states of parameters PCIe enable and PCIe asynchronous mode enable.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR bEnableFast/Returns the state of the fast mode.
/// HIPAR bEnableAsynchronous/Returns the state of the asynchronous mode.
ViStatus _VI_FUNC rssgs_GetFastSettings(ViSession instrSession, ViPBoolean bEnableFast, ViPBoolean bEnableAsynchronous)
{
  ViInt32 enableFast = 0, enableAsynchronous = 0;
  ViStatus rssgs_status = rsGetAttribute(instrSession, SGS_ATTR_FAST, &enableFast);

  if ((rssgs_status == VI_SUCCESS) && (enableFast))
  {
    rssgs_status = rsGetAttribute(instrSession, SGS_ATTR_FASTASYNCHRONOUS, &enableAsynchronous);
  }
  *bEnableFast = enableFast != 0;
  *bEnableAsynchronous = enableAsynchronous != 0;
  return rssgs_status;
}



/// HIFN In fast mode this function sends an empty command to instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_FastSync (ViSession instrSession)
{
  ViStatus rssgs_status = rsFastSync (instrSession);
  if (rssgs_status == VI_ERROR_NSUP_OPER)
  {
    ViInt32 OPCState = 0;
    rssgs_status = rssgs_OperationCompleteQuery (instrSession, &OPCState);
  }
  return rssgs_status;
}


//=====================================================================================
/// HIFN This function sets the state of the signal output.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control sets the state of the signal output.
ViStatus _VI_FUNC rssgs_SetRFState (ViSession instrSession, ViBoolean state)
{
    ViUInt32 supported = 0;
    viFastSettingSupported (instrSession, IdSgsRfState, &supported);
    if (supported)
        return viFastSetInt32(instrSession, IdSgsRfState, state);

    return rssgs_setViBoolean ( instrSession,
                                ":OUTP",
                                state,
                                VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the state of the signal output.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the state of the signal output.
ViStatus _VI_FUNC rssgs_GetRFState (ViSession instrSession, ViBoolean *state)
{

    ViUInt32 supported = 0;
    viFastQuerySupported(instrSession, IdSgsRfState, &supported);
    if (supported)
    {
      ViInt32 sel = 0;
      int status = 0;
      status = viFastQueryInt32(instrSession, IdSgsRfState, &sel);
      *state = (sel != 0);
      return status;
    }
    return rssgs_getViBoolean ( instrSession,
                                ":OUTP?",
                                state,
                                VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the number of RF paths.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR paths/Displays the number of RF paths.
ViStatus _VI_FUNC rssgs_GetRFNumberofPaths (ViSession instrSession,
                                            ViInt32 *paths)
{
  return rssgs_getViInt32 (instrSession,
    ":SOUR:PATH:COUN?",
    paths,
    VI_ERROR_PARAMETER2);
}



/// HIFN This function sets the mode of the attenuator at the RF output (Attenuator MODe).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR attenuatorMode/This control sets the mode of the attenuator at the RF output (Attenuator MODe).
ViStatus _VI_FUNC rssgs_SetRFAttenuatorMode (ViSession instrSession,
                                             ViInt32 attenuatorMode)
{
    return rssgs_setEnum (instrSession,
                          "OUTP:AMOD",
                          attenuatorMode,
                          attenuatorModeArr,
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the mode of the attenuator at the RF output (Attenuator MODe).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR attenuatorMode/Displays the mode of the attenuator at the RF output (Attenuator MODe).
ViStatus _VI_FUNC rssgs_GetRFAttenuatorMode (ViSession instrSession,
                                             ViInt32 *attenuatorMode)
{
    return rssgs_getEnum (instrSession,
                          "OUTP:AMOD?",
                          attenuatorMode,
                          attenuatorModeArr,
                          VI_ERROR_PARAMETER2);
}


//=====================================================================================
/// HIFN This function sets the operation mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR operationMode/This control sets the operation mode.
ViStatus _VI_FUNC rssgs_SetOperationMode (ViSession instrSession,
                                          ViInt32 operationMode)
{
    return rssgs_setEnum (instrSession,
                          ":SOUR:OPM",
                          operationMode,
                          operationModeArr,
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the operation mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR operationMode/Displays the operation mode.
ViStatus _VI_FUNC rssgs_GetOperationMode (ViSession instrSession,
                                          ViInt32 *operationMode)
{
    return rssgs_getEnum (instrSession,
                          ":SOUR:OPM?",
                          operationMode,
                          operationModeArr,
                          VI_ERROR_PARAMETER2);
}




//=====================================================================================
/// HIFN This function configures the local oscillator signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR source/This control sets the source of the local oscillator signal.
/// HIPAR output/This control sets the signal to be connected to the REF output connector.

ViStatus _VI_FUNC rssgs_ConfigureFrequencyLocalOscillator (ViSession instrSession,
                                                           ViInt32 source,
                                                           ViInt32 output)
{
    ViStatus    rssgs_status;

    CHECKERR (rssgs_SetFrequencyLocalOscillatorSource (instrSession, source));

    rssgs_status = rssgs_SetFrequencyLocalOscillatorOutput (instrSession, output);
    REMAP (2, 3);

    return rssgs_status;
}

//=====================================================================================
/// HIFN This function sets the RF frequency of the RF output connector
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/Sets the RF frequency of the RF output connector.
ViStatus _VI_FUNC rssgs_SetFrequency (ViSession instrSession,
                                      ViReal64 frequency)
{

    ViUInt32 supported = 0;
    viFastSettingSupported (instrSession, IdSgsFreqFreq, &supported);
    if (supported)
        return viFastSetReal64(instrSession, IdSgsFreqFreq, frequency);

    return rssgs_setViReal64 (instrSession,
                              ":FREQ",
                              frequency,
                              0, 0,
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the RF frequency of the RF output connector.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/Displays the RF frequency of the RF output connector.
ViStatus _VI_FUNC rssgs_GetFrequency (ViSession instrSession,
                                      ViReal64 *frequency)
{

    ViUInt32 supported = 0;
    viFastQuerySupported (instrSession, IdSgsFreqFreq, &supported);
    if (supported)
      return viFastQueryReal64(instrSession, IdSgsFreqFreq, frequency);

    return rssgs_getViReal64 (instrSession,
                              ":FREQ?",
                              frequency,
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the RF frequency offset 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/Sets the RF frequency offset .
ViStatus _VI_FUNC rssgs_SetFrequencyOffset(ViSession instrSession,
  ViReal64 frequency)
{
  return rssgs_setViReal64(instrSession,
    ":FREQ:OFFS",
    frequency,
    0, 0,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the RF frequency offset
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/Displays the RF frequency offset
ViStatus _VI_FUNC rssgs_GetFrequencyOffset(ViSession instrSession,
  ViReal64 *frequency)
{
  return rssgs_getViReal64(instrSession,
    ":FREQ:OFFS?",
    frequency,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the RF signal's relative phase.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR deltaPhase/Sets the RF signal's relative phase.
ViStatus _VI_FUNC rssgs_SetFrequencyDeltaPhase (ViSession instrSession,
                                                ViReal64 deltaPhase)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:PHAS", 
                              deltaPhase, -360.0, 360.0, 
                              VI_ERROR_PARAMETER2);    
}

//=====================================================================================
/// HIFN This function queries the RF signal's relative phase.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR deltaPhase/Displays the RF signal's relative phase.
ViStatus _VI_FUNC rssgs_GetFrequencyDeltaPhase (ViSession instrSession,
                                                ViReal64 *deltaPhase)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:PHAS?", 
                              deltaPhase,  
                              VI_ERROR_PARAMETER2);    
}

/// HIFN This function adopt the phase set with SOURce:PHASe as the current phase.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_SetFrequencyDeltaPhaseAsCurrent
             (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, ":SOUR:PHAS:REF");
}

//=====================================================================================
/// HIFN This function sets the source of the local oscillator signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR source/This control sets the source of the local oscillator signal.
ViStatus _VI_FUNC rssgs_SetFrequencyLocalOscillatorSource (ViSession instrSession,
                                                           ViInt32 source)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:LOSC:SOUR", 
                          source, 
                          FrequencyLocalOscillatorSourceArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the source of the local oscillator signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR source/Displays the source of the local oscillator signal.
ViStatus _VI_FUNC rssgs_GetFrequencyLocalOscillatorSource (ViSession instrSession,
                                                           ViInt32 *source)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:LOSC:SOUR?", 
                          source, 
                          FrequencyLocalOscillatorSourceArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the signal to be connected to the REF output connector.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR output/This control sets the signal to be connected to the REF output connector.
ViStatus _VI_FUNC rssgs_SetFrequencyLocalOscillatorOutput (ViSession instrSession,
                                                           ViInt32 output)
{
    return rssgs_setEnum (instrSession, 
                          ":CONN:REFL:OUTP", 
                          output,   
                          FrequencyLocalOscillatorOutputArr,    
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the signal to be connected to the REF output connector.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR output/Displays the signal to be connected to the REF output connector.
ViStatus _VI_FUNC rssgs_GetFrequencyLocalOscillatorOutput (ViSession instrSession,
                                                           ViInt32 *output)
{
    return rssgs_getEnum (instrSession, 
                          ":CONN:REFL:OUTP?", 
                          output,   
                          FrequencyLocalOscillatorOutputArr,    
                          VI_ERROR_PARAMETER2);
}



//=====================================================================================
/// HIFN This function sets the operation mode of the trigger connector.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/This control sets the operation mode of the trigger connector.
ViStatus _VI_FUNC rssgs_SetTrigConnnectorOpMode(ViSession instrSession, 
                                                ViInt32 mode)
{
  return rssgs_setEnum (instrSession, 
    ":CONNector:TRIGger:OMODe", 
    mode,   
    trigConOpModeArr,    
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the operation mode of the trigger connector.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/Displays the operation mode of the trigger connector.
ViStatus _VI_FUNC rssgs_GetTrigConnnectorOpMode (ViSession instrSession, 
                                                 ViPInt32 mode)
{
  return rssgs_getEnum (instrSession, 
    ":CONNector:TRIGger:OMODe?", 
    mode,   
    trigConOpModeArr,    
    VI_ERROR_PARAMETER2);

}

/// HIFN This function configures reference oscillator parameters.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR source/This control sets the reference oscillator signal source.
/// HIPAR referenceFrequency/This control sets the reference output frequency.
/// HIPAR externalReferenceFrequency/This control sets the external reference input frequency.
/// HIPAR localOscillatorSource/This control sets the source of the local oscillator signal.
/// HIPAR localOscillatorOutput/This control sets the signal to be connected to the REF output connector.
ViStatus _VI_FUNC rssgs_ConfigureReferenceOscillator (ViSession instrSession,
                                                      ViInt32 source,
                                                      ViInt32 referenceFrequency,
                                                      ViInt32 externalReferenceFrequency,
                                                      ViInt32 localOscillatorSource,
                                                      ViInt32 localOscillatorOutput)
{
    ViStatus rssgs_status = VI_SUCCESS;
    
    CHECKERR (rssgs_SetROSCSource (instrSession, source));
    rssgs_status = rssgs_SetFrequencyLocalOscillatorSource (instrSession, localOscillatorSource);
    REMAP (2,5);
    rssgs_status = rssgs_SetFrequencyLocalOscillatorOutput (instrSession, localOscillatorOutput);
    REMAP (2,6);
    rssgs_status = rssgs_SetROSCReferenceOutputFrequency (instrSession, referenceFrequency);
    REMAP (2,3);
    rssgs_status = rssgs_SetROSCExternalReferenceFrequency (instrSession, externalReferenceFrequency);
    REMAP (2,4);
    
    return rssgs_status;
}

//=====================================================================================
/// HIFN This function sets the reference oscillator signal source.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR source/This control sets the reference oscillator signal source.
ViStatus _VI_FUNC rssgs_SetROSCSource (ViSession instrSession,
                                       ViInt32 source)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:ROSC:SOUR", 
                          source, 
                          ROSCSourceArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the reference oscillator signal source.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR source/Displays the reference oscillator signal source.
ViStatus _VI_FUNC rssgs_GetROSCSource (ViSession instrSession,
                                       ViInt32 *source)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:ROSC:SOUR?", 
                          source, 
                          ROSCSourceArr, 
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function sets the reference output frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR referenceFrequency/This control sets the reference output frequency.
ViStatus _VI_FUNC rssgs_SetROSCReferenceOutputFrequency (ViSession instrSession,
                                                         ViInt32 referenceFrequency)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:ROSC:OUTP:FREQ", 
                          referenceFrequency, 
                          ROSCreferenceFrequencyArr, 
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the reference output frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR referenceFrequency/Displays the reference output frequency.
ViStatus _VI_FUNC rssgs_GetROSCReferenceOutputFrequency (ViSession instrSession,
                                                         ViInt32 *referenceFrequency)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:ROSC:OUTP:FREQ?", 
                          referenceFrequency, 
                          ROSCreferenceFrequencyArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the external reference input frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR externalReferenceFrequency/This control sets the external reference input frequency.
ViStatus _VI_FUNC rssgs_SetROSCExternalReferenceFrequency (ViSession instrSession,
                                                           ViInt32 externalReferenceFrequency)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:ROSC:EXT:FREQ", 
                          externalReferenceFrequency, 
                          ROSCExternalReferenceFrequencyArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the external reference input frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR externalReferenceFrequency/Displays the external reference input frequency.
ViStatus _VI_FUNC rssgs_GetROSCExternalReferenceFrequency (ViSession instrSession,
                                                           ViInt32 *externalReferenceFrequency)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:ROSC:EXT:FREQ?", 
                          externalReferenceFrequency, 
                          ROSCExternalReferenceFrequencyArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the external synchronization bandwidth.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR externalSyncBandwidth/This control sets the synchronization bandwidth.
ViStatus _VI_FUNC rssgs_SetROSCExternalSyncBandwidth (ViSession instrSession, 
                                                      ViInt32 externalSyncBandwidth)
{
  return rssgs_setEnum (instrSession, 
    ":SOUR:ROSC:EXT:SBAN", 
    externalSyncBandwidth, 
    ROSCExternalSyncBandwidthArr, 
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the external synchronization bandwidth.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR externalSyncBandwidth/Displays the synchronization bandwidth.
ViStatus _VI_FUNC rssgs_GetROSCExternalSyncBandwidth (ViSession instrSession, 
                                                      ViPInt32 externalSyncBandwidth)
{
  return rssgs_getEnum (instrSession, 
    ":SOUR:ROSC:EXT:SBAN?", 
    externalSyncBandwidth, 
    ROSCExternalSyncBandwidthArr, 
    VI_ERROR_PARAMETER2);
}



/// HIFN This function sets the oscillator calibration value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR data/This control sets the oscillator calibration value.
ViStatus _VI_FUNC rssgs_SetROSCCalibrationData (ViSession instrSession,
                                                ViReal64 data)
{
    return rssgs_setViReal64 (instrSession,
                              ":CAL:ROSC",
                              data,
                              0.0, 0.0,
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the oscillator calibration value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR data/Displays the oscillator calibration value.
ViStatus _VI_FUNC rssgs_GetROSCCalibrationData (ViSession instrSession,
                                                ViReal64 *data)
{
    return rssgs_getViReal64 (instrSession,
                              ":CAL:ROSC?",
                              data,
                              VI_ERROR_PARAMETER2);
}


//=====================================================================================
/// HIFN This function defines whether the factory provided or a custom defined calibration value is used to
//// adjust the reference oscillator
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/This control sets the data mode.
ViStatus _VI_FUNC rssgs_SetROSCCalibrationDataMode (ViSession instrSession, 
                                      ViInt32 mode)
{
  return rssgs_setEnum (instrSession, 
    ":CAL:ROSC:DATA:MODE", 
    mode, 
    CalRoscDataModeArr, 
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries whether the factory provided or a custom defined calibration value is used to
////adjust the reference oscillator
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/Displays the data mode.
ViStatus _VI_FUNC rssgs_GetROSCCalibrationDataMode (ViSession instrSession, 
                                      ViInt32 *mode)
{
  return rssgs_getEnum (instrSession, 
    ":CAL:ROSC:DATA:MODE?", 
    mode, 
    CalRoscDataModeArr, 
    VI_ERROR_PARAMETER2);
}



/// HIFN This function determines whether the calibrated (OFF) or a user-defined (ON) adjustment value is used for fine adjustment of the frequency. With STATe ON, the instrument is no longer in the calibrated state.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR adjustmentActive/Activates/deactivates the pulse modulation.
ViStatus _VI_FUNC rssgs_SetROSCAdjustmentActive (ViSession instrSession,
                                                 ViBoolean adjustmentActive)
{
    return rssgs_setViBoolean (instrSession,
                               "SOUR:ROSC:ADJ",
                               adjustmentActive,
                               VI_ERROR_PARAMETER2);
}

/// HIFN This function determines whether the calibrated (OFF) or a user-defined (ON) adjustment value is used for fine adjustment of the frequency. With STATe ON, the instrument is no longer in the calibrated state.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR adjustmentActive/Determines whether the calibrated (OFF) or a user-defined (ON) adjustment value is used for fine adjustment of the frequency.
ViStatus _VI_FUNC rssgs_GetROSCAdjustmentActive (ViSession instrSession,
                                                 ViBoolean *adjustmentActive)
{
    return rssgs_getViBoolean (instrSession,
                               "SOUR:ROSC:ADJ?",
                               adjustmentActive,
                               VI_ERROR_PARAMETER2);
}


/// HIFN This function specifies the frequency correction value (adjustment value). A detailed definition is given in  section RF Reference Frequency - Reference Oscillator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR adjustmentFreq/Specifies the frequency correction value (adjustment value).
ViStatus _VI_FUNC rssgs_SetROSCAdjustmentValue (ViSession instrSession,
                                                    ViInt32   adjustmentFreq)
{
    return rssgs_setViInt32 (instrSession,
                             "SOUR:ROSC:ADJ:VAL",
                             adjustmentFreq,
                             0,4095,
                             VI_ERROR_PARAMETER2);
}

/// HIFN This function returns the frequency correction value (adjustment value). A detailed definition is given in  section  RF Reference Frequency - Reference Oscillator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR adjustmentValue/Returns the frequency correction value (adjustment value).
ViStatus _VI_FUNC rssgs_GetROSCAdjustmentValue (ViSession  instrSession,
                                                    ViInt32*   adjustmentValue)
{
    return rssgs_getViInt32 (instrSession,
                             "SOUR:ROSC:ADJ:VAL?",
                             adjustmentValue,
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function Configures the power measurement display.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR trace/selects the trace.
/// HIPAR permanent/Activates the permanent display of the measured power level results in the winbar.
/// HIPAR priority/Selects average or peak power for permanent display.
ViStatus _VI_FUNC rssgs_ConfigurePowerSensorDisplay (ViSession instrSession,
                                                     ViInt32 sensor,
													 ViInt32 trace,
                                                     ViBoolean permanentDisplay,
                                                     ViInt32 priority)
{

	ViStatus rssgs_status;
    
	CHECKERR (rssgs_SetPowerDisplayPermanentState (instrSession, sensor, permanentDisplay));
    
    rssgs_status = rssgs_SetPowerDisplayPermanentPriority (instrSession, trace, priority);
    REMAP (3,5);
    
    return rssgs_status;
}

//=====================================================================================
/// HIFN This function Performs zeroing of the sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
ViStatus _VI_FUNC rssgs_ConfigurePowerSensorZero (ViSession instrSession,
                                                  ViInt32 sensor)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:ZERO", sensor);

    return rssgs_sendCmd (instrSession, buffer);
}

//=====================================================================================
/// HIFN This function determines the signal to be measured. You can select the RF
/// HIFN signal of path A, path B or a user defined source.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR level/This control sets the RF signal's power level.
ViStatus _VI_FUNC rssgs_ConfigurePowerSensorSource (ViSession instrSession,
                                                    ViInt32 sensor,
                                                    ViInt32 source)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:SOUR", sensor);

    return rssgs_setEnum (instrSession, 
                          buffer,
                          source,
                          powerSourceArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function configures the power measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR levelOffsetState/Activates the addition of the level offset to the measured value.
/// HIPAR levelOffset/Sets a level offset which is added to the measured level value
/// HIPAR filter/Selects the filter mode.
/// HIPAR SParameter/Activates or deactivates the use of the s-parameters correction data of the connected power sensor.
/// HIPAR logging/Activates the recording of the power values
ViStatus _VI_FUNC rssgs_ConfigurePowerSensor (ViSession instrSession,
                                              ViInt32 sensor,
                                              ViBoolean levelOffsetState,
                                              ViReal64 levelOffset,
                                              ViInt32 filter,
                                              ViBoolean SParameter,
                                              ViBoolean logging)
{
	ViStatus    rssgs_status;

    CHECKERR (rssgs_SetPowerOffsetState (instrSession, sensor, levelOffsetState));

    rssgs_status = rssgs_SetPowerOffset (instrSession, sensor, levelOffset);
    REMAP (4, 3);
	
	rssgs_status = rssgs_SetPowerFilterMode (instrSession, sensor, filter);
    REMAP (5, 3);
	
	rssgs_status = rssgs_SetPowerUseSParameters (instrSession, sensor, SParameter);
    REMAP (6, 3);
	
	rssgs_status = rssgs_SetPowerLoggingState (instrSession, sensor, logging);
    REMAP (7, 3);

    return rssgs_status;
}

//=====================================================================================
/// HIFN Configures the power measurement aperture time.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR defaultTime/Deactivates the default aperture time of the respective sensor.
/// HIPAR apertureTime/Defines the aperture time (size of the acquisition interval) for the corresponding sensor.
ViStatus _VI_FUNC rssgs_ConfigurePowerSensorApertureTime (ViSession instrSession,
                                                          ViInt32 sensor,
                                                          ViBoolean defaultTime,
                                                          ViReal64 apertureTime)
{
    ViStatus    rssgs_status;

    CHECKERR (rssgs_SetPowerDefaultTime (instrSession, sensor, defaultTime));

    rssgs_status = rssgs_SetPowerApertureTime (instrSession, sensor, apertureTime);
    REMAP (4, 3);
	
    return rssgs_status;
}
//=====================================================================================
/// HIFN  This function configure the power sense (NRP-Z Power View)
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR initPowerMeas/Activates/deactivates the continuous power measurement by the R&S NRP-Z power sensors.
/// HIPAR unit/Sets the unit used for result query with command READ.
/// HIPAR frequency/Sets the RF frequency of the source if the user source is selected
ViStatus _VI_FUNC rssgs_ConfigurePowerSensorSense (ViSession instrSession,
                                                  ViInt32 sensor,
                                                  ViBoolean initPowerMeas,
                                                  ViInt32 unit,
                                                  ViReal64 frequency)
{
     ViStatus    rssgs_status;

    CHECKERR (rssgs_SetInitiatePowerMeasurement (instrSession, sensor, initPowerMeas));

    rssgs_status = rssgs_SetPowerSensorUnit (instrSession, sensor, unit);
    REMAP (4, 3);
	
	rssgs_status = rssgs_SetPowerFrequency (instrSession, sensor, frequency);
    REMAP (5, 3);
	
    return rssgs_status;
}

//=====================================================================================
/// HIFN  This function activates/deactivates the continuous power measurement by the R&S NRP-Z power sensors.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR initPowerMeas/Activates/deactivates the continuous power measurement by the R&S NRP-Z power sensors.
ViStatus _VI_FUNC rssgs_SetInitiatePowerMeasurement (ViSession instrSession,
                                                     ViInt32 sensor,
                                                     ViBoolean initPowerMeas)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "INIT" _PERCD ":POW:CONT", sensor);

    return rssgs_setViBoolean (instrSession, 
                               buffer,
                               initPowerMeas,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function returns activation the continuous power measurement by the R&S NRP-Z power sensors.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR initPowerMeas/Activates/deactivates the continuous power measurement by the R&S NRP-Z power sensors.
ViStatus _VI_FUNC rssgs_GetInitiatePowerMeasurement (ViSession instrSession,
                                                     ViInt32 sensor,
                                                     ViBoolean *initPowerMeas)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "INIT" _PERCD ":POW:CONT?", sensor);

    return rssgs_getViBoolean (instrSession, 
                               buffer,
                               initPowerMeas,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function activates/deactivates the use of the s-parameters correction data of the connected power sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR useSParameters/Activates/deactivates the use of the s-parameters correction data of the connected power sensor.
ViStatus _VI_FUNC rssgs_SetPowerUseSParameters (ViSession instrSession,
                                                ViInt32 sensor,
                                                ViBoolean useSParameters)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:CORR:SPD:STAT", sensor);

    return rssgs_setViBoolean (instrSession, 
                               buffer,
                               useSParameters,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function activates/deactivates the use of the s-parameters correction data of the connected power sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR useSParameters/Activates/deactivates the use of the s-parameters correction data of the connected power sensor.
ViStatus _VI_FUNC rssgs_GetPowerUseSParameters (ViSession instrSession,
                                                ViInt32 sensor,
                                                ViBoolean *useSParameters)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:CORR:SPD:STAT?", sensor);

    return rssgs_getViBoolean (instrSession, 
                               buffer,
                               useSParameters,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function configures the state of recording the power values, measured by a connected R&S NRP-Z power sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR logging/Activates the recording of the power values.
ViStatus _VI_FUNC rssgs_SetPowerLoggingState (ViSession instrSession,
                                              ViInt32 sensor, ViBoolean logging)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:LOGG:STAT", sensor);

    return rssgs_setViBoolean (instrSession, 
                              buffer,
                              logging,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function Returns the state of recording the power values, measured by a connected R&S NRP-Z power sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR logging/Returns the state of the recording of the power values.
ViStatus _VI_FUNC rssgs_GetPowerLoggingState (ViSession instrSession,
                                              ViInt32 sensor, ViBoolean *logging)
{
   ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:LOGG:STAT?", sensor);

    return rssgs_getViBoolean (instrSession, 
                              buffer,
                              logging,
                              VI_ERROR_PARAMETER3); 
}

//=====================================================================================
/// HIFN  This function Deactivate or activates the default aperture time of the respective sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR defaultTime/Deactivates or activates the default aperture time of the respective sensor.
ViStatus _VI_FUNC rssgs_SetPowerDefaultTime (ViSession instrSession,
                                             ViInt32 sensor,
                                             ViBoolean defaultTime)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:APER:DEF:STAT", sensor);

    return rssgs_setViBoolean (instrSession, 
                              buffer,
                              defaultTime,
                              VI_ERROR_PARAMETER3); 
}

//=====================================================================================
/// HIFN  This function returns the state of the default aperture time of the respective sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR defaultTime/returns the state of the default aperture time of the respective sensor.
ViStatus _VI_FUNC rssgs_GetPowerDefaultTime (ViSession instrSession,
                                             ViInt32 sensor,
                                             ViBoolean *defaultTime)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:APER:DEF:STAT?", sensor);

    return rssgs_getViBoolean (instrSession, 
                              buffer,
                              defaultTime,
                              VI_ERROR_PARAMETER3); 
}

//=====================================================================================
/// HIFN  This function Defines the aperture time (size of the acquisition interval) for the corresponding sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR apertureTime/Defines the aperture time (size of the acquisition interval) for the corresponding sensor.
ViStatus _VI_FUNC rssgs_SetPowerApertureTime (ViSession instrSession,
                                              ViInt32 sensor,
                                              ViReal64 apertureTime)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:APER:TIM", sensor);

    return rssgs_setViReal64 (instrSession, 
                              buffer,
                              apertureTime,
                              0.0,0.0,
                              VI_ERROR_PARAMETER3);
}


//=====================================================================================
/// HIFN  This function returns the aperture time (size of the acquisition interval) for the corresponding sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR apertureTime/Returns the aperture time (size of the acquisition interval) for the corresponding sensor.
ViStatus _VI_FUNC rssgs_GetPowerApertureTime (ViSession instrSession,
                                              ViInt32 sensor,
                                              ViReal64 *apertureTime)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:APER:TIM?", sensor);

    return rssgs_getViReal64 (instrSession, 
                              buffer,
                              apertureTime,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function sets the RF frequency of the source if the user source is selected (SENSe[:POWer]:SOURce USER.)
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR frequency/Sets the RF frequency of the source
ViStatus _VI_FUNC rssgs_SetPowerFrequency (ViSession instrSession,
                                           ViInt32 sensor, ViReal64 frequency)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FREQ", sensor);

    return rssgs_setViReal64 (instrSession, 
                              buffer,
                              frequency,
                              0.0,0.0,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function returns the RF frequency of the source if the user source is selected (SENSe[:POWer]:SOURce USER.)
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR frequency/Returns the RF frequency of the source
ViStatus _VI_FUNC rssgs_GetPowerFrequency (ViSession instrSession,
                                           ViInt32 sensor, ViReal64 *frequency)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FREQ?", sensor);

    return rssgs_getViReal64 (instrSession, 
                              buffer,
                              frequency,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function sets a level offset which is added to the measured level value
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR offset/Sets a level offset which.
ViStatus _VI_FUNC rssgs_SetPowerOffset (ViSession instrSession,
                                        ViInt32 sensor, ViReal64 offset)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    CHECKPARAM (rssgs_invalidViReal64Range (instrSession, offset, -100.0, 100.0), 3);
    
    sprintf (buffer, "SENS" _PERCD ":POW:OFFS %f", sensor, offset);

    return rssgs_sendCmd (instrSession, buffer);
}

//=====================================================================================
/// HIFN  This function returns a level offset which is added to the measured level value
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR offset/displays a level offset.
ViStatus _VI_FUNC rssgs_GetPowerOffset (ViSession instrSession,
                                        ViInt32 sensor, ViReal64 *offset)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:OFFS?", sensor);

    return rssgs_getViReal64 (instrSession, 
                              buffer,
                              offset,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function activates/deactivates the addition of the level offset to the measured value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR offsetState/Activates or deactivates the addition of the level offset to the measured value.
ViStatus _VI_FUNC rssgs_SetPowerOffsetState (ViSession instrSession,
                                             ViInt32 sensor,
                                             ViBoolean offsetState)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:OFFS:STAT", sensor);

    return rssgs_setViBoolean (instrSession, 
                               buffer,
                               offsetState,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN  This function returns the state of activation the addition of the level offset to the measured value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session. 
/// HIPAR sensor/Sets the sensor number.
/// HIPAR offsetState/Displays activation the addition of the level offset to the measured value.
ViStatus _VI_FUNC rssgs_GetPowerOffsetState (ViSession instrSession,
                                             ViInt32 sensor,
                                             ViBoolean *offsetState)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:OFFS:STAT?", sensor);

    return rssgs_getViBoolean (instrSession, 
                               buffer,
                               offsetState,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function selects the unit used for result query with command READ.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR unit/Sets the unit used for result query with command READ.
ViStatus _VI_FUNC rssgs_SetPowerSensorUnit (ViSession instrSession, ViInt32 sensor,
                                            ViInt32 unit)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":UNIT:POW", sensor);

    return rssgs_setEnum (instrSession, 
                          buffer,
                          unit,
                          sensePowerUnitArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function returns the unit used for result query with command READ.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR unit/Gets the unit used for result query with command READ.
ViStatus _VI_FUNC rssgs_GetPowerSensorUnit (ViSession instrSession, ViInt32 sensor,
                                            ViInt32 *unit)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":UNIT:POW?", sensor);

    return rssgs_getEnum (instrSession, 
                          buffer,
                          unit,
						  sensePowerUnitArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function switches on and off  the permanent indication of 
/// HIFN the power measurement result in the upper right corner of the block diagram.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR permanentDisplay/Switches on and off  the permanent indication of the power measurement result
ViStatus _VI_FUNC rssgs_SetPowerDisplayPermanentState (ViSession instrSession,
                                                       ViInt32 sensor,
                                                       ViBoolean permanentDisplay)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:DISP:PERM:STAT", sensor);

    return rssgs_setViBoolean (instrSession, 
                               buffer,
                               permanentDisplay,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function returns thee state  the permanent indication of 
/// HIFN the power measurement result in the upper right corner of the block diagram.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR permanentDisplay/Returns the permanent indication of the power measurement result.
ViStatus _VI_FUNC rssgs_GetPowerDisplayPermanentState (ViSession instrSession,
                                                       ViInt32 sensor,
                                                       ViBoolean *permanentDisplay)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:DISP:PERM:STAT?", sensor);

    return rssgs_getViBoolean (instrSession, 
                               buffer,
                               permanentDisplay,
                               VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function selects which power measurement result (average or
/// HIFN peak power) is indicated when permanent display is active.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR trace/Selects the trace.
/// HIPAR priority/Selects which power measurement result (average or
/// HIPAR peak power) is indicated when permanent display is active.
ViStatus _VI_FUNC rssgs_SetPowerDisplayPermanentPriority (ViSession instrSession,
                                                          ViInt32 trace,
                                                          ViInt32 priority)
{
    
    ViChar buffer[BUFFER_SIZE];
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, trace, 1, 3), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:DISP:PERM:PRI", trace);
    
    return rssgs_setEnum (instrSession,
                          buffer,
                          priority,
                          powerPriorityArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function returns which power measurement result (average or
/// HIFN peak power) is indicated when permanent display is active.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR trace/Selects the trace.
/// HIPAR priority/Returns which power measurement result (average or
/// HIPAR peak power) is indicated when permanent display is active.
ViStatus _VI_FUNC rssgs_GetPowerDisplayPermanentPriority (ViSession instrSession,
                                                          ViInt32 trace,
                                                          ViInt32 *priority)
{
    ViChar buffer[BUFFER_SIZE];
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, trace, 1, 3), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:DISP:PERM:PRI?", trace);
    
    return rssgs_getEnum (instrSession,
                          buffer,
                          priority,
                          powerPriorityArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function selects the filter mode. For mode NSR selects
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR filterMode/Sets the filter mode.
/// HIPAR noiseContent/Sets the noise content for fixed noise filter mode.
/// HIPAR timeout/Sets the timeout for fixed noise filter mode.
ViStatus _VI_FUNC rssgs_ConfigurePowerFilter (ViSession instrSession,
                                              ViInt32 sensor, ViInt32 filterMode,
                                              ViReal64 noiseContent,
                                              ViReal64 timeout)
{
    ViStatus rssgs_status;
    
    CHECKERR (rssgs_SetPowerFilterMode (instrSession, sensor, filterMode));
    
    rssgs_status = rssgs_SetPowerFilterFixedNoiseContent (instrSession, sensor, noiseContent);
    REMAP (3,4);
    
    rssgs_status = rssgs_SetPowerFilterFixedNoiseTimeout (instrSession, sensor, timeout);
    REMAP (3,5);

    return rssgs_status; 
}

//=====================================================================================
/// HIFN This function selects the filter length for user filter mode
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR filterLength/Sets the filter length for user filter mode.
ViStatus _VI_FUNC rssgs_SetPowerFilterLength (ViSession instrSession,
                                              ViInt32 sensor,
                                              ViInt32 filterLength)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:LENG:USER", sensor);

    return rssgs_setViInt32 (instrSession, 
                             buffer,
                             filterLength,
                             1, 65536,
                             VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function selects the filter length for user filter mode.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR filterLength/Displays the filter length for user filter mode.
ViStatus _VI_FUNC rssgs_GetPowerFilterLength (ViSession instrSession,
                                              ViInt32 sensor,
                                              ViInt32 *filterLength)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:LENG:USER?", sensor);

    return rssgs_getViInt32 (instrSession, 
                             buffer,
                             filterLength,
                             VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function selects the filter mode.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR filterMode/Sets the filter mode.
ViStatus _VI_FUNC rssgs_SetPowerFilterMode (ViSession instrSession,
                                            ViInt32 sensor, ViInt32 filterMode)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:TYPE", sensor);

    return rssgs_setEnum (instrSession, 
                          buffer,
                          filterMode,
                          powerFilterModeArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the filter mode for user.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR filterMode/Displays the filter mode.
ViStatus _VI_FUNC rssgs_GetPowerFilterMode (ViSession instrSession,
                                            ViInt32 sensor, ViInt32 *filterMode)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:TYPE?", sensor);

    return rssgs_getEnum (instrSession, 
                          buffer,
                          filterMode,
                          powerFilterModeArr,
                          VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function sets the noise content for fixed noise filter mode
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR noiseContent/This control sets the noise content for fixed noise filter mode.
ViStatus _VI_FUNC rssgs_SetPowerFilterFixedNoiseContent (ViSession instrSession,
                                                         ViInt32 sensor,
                                                         ViReal64 noiseContent)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:NSR", sensor);

    return rssgs_setViReal64 (instrSession, 
                              buffer,
                              noiseContent,
                              0.0001, 1.0,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function sets the noise content for fixed noise filter mode
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR noiseContent/This control sets the noise content for fixed noise filter mode.
ViStatus _VI_FUNC rssgs_GetPowerFilterFixedNoiseContent (ViSession instrSession,
                                                         ViInt32 sensor,
                                                         ViReal64 *noiseContent)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:NSR?", sensor);

    return rssgs_getViReal64 (instrSession, 
                              buffer,
                              noiseContent,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function sets the timeout for fixed noise filter mode
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR timeout/This control sets the timeout for fixed noise filter mode.
ViStatus _VI_FUNC rssgs_SetPowerFilterFixedNoiseTimeout (ViSession instrSession,
                                                         ViInt32 sensor,
                                                         ViReal64 timeout)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:NSR:MTIM", sensor);

    return rssgs_setViReal64 (instrSession, 
                              buffer,
                              timeout,
                              1.0, 999.99,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function returns the timeout for fixed noise filter mode.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR timeout/Gets the timeout for fixed noise filter mode.
ViStatus _VI_FUNC rssgs_GetPowerFilterFixedNoiseTimeout (ViSession instrSession,
                                                         ViInt32 sensor,
                                                         ViReal64 *timeout)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:NSR:MTIM?", sensor);

    return rssgs_getViReal64 (instrSession, 
                              buffer,
                              timeout,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function returns the filter length automatically set value.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR filterLengthAuto/Gets the filter length automatically set value.
ViStatus _VI_FUNC rssgs_GetPowerFilterLengthAuto (ViSession instrSession,
                                                  ViInt32 sensor,
                                                  ViInt32 *filterLengthAuto)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:LENG:AUTO?", sensor);

    return rssgs_getViInt32 (instrSession, 
                             buffer,
                             filterLengthAuto,
                             VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function the search for the optimum filter length for the current measurement conditions.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
ViStatus _VI_FUNC rssgs_PowerFilterSearchOnce (ViSession instrSession,
                                               ViInt32 sensor)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:FILT:SONC", sensor);

    return rssgs_sendCmd (instrSession, buffer);
}

//=====================================================================================
/// HIFN This function the search for the optimum filter length for the current measurement conditions.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR devoce/This control displays the if a sensor is connected at the selected port to the signal generator.
ViStatus _VI_FUNC rssgs_PowerSensorDevice (ViSession instrSession,
                                           ViInt32 sensor, ViInt32 *device)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:STAT:DEV?", sensor);

    return rssgs_getViInt32 (instrSession, 
                             buffer,
                             device,
                             VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the type of sensor connected at the selected port.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/Sets the sensor number.
/// HIPAR type/This control displays the sensor type.
ViStatus _VI_FUNC rssgs_PowerSensorType (ViSession instrSession,
                                         ViInt32 sensor, ViChar type[])
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:TYPE?", sensor);

    return rssgs_getViString (instrSession, 
                              buffer,
                              type,
                              1024,
                              VI_ERROR_PARAMETER3);
}


//=====================================================================================
/// HIFN This function retrieves the power measurement result of the selected power sensor.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR read/This control sets the power sensor connected to the SENSOR port.
/// HIPAR power/This control displays the power measurement result of the selected power sensor.
ViStatus _VI_FUNC rssgs_PowerReadPower (ViSession instrSession, ViInt32 read,
                                        ViReal64 *power)
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, read, 1, 3), 2);
    
    sprintf (buffer, "READ" _PERCD ":POW?", read);

    return rssgs_getViReal64 (instrSession, 
                              buffer,
                              power,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the software version of the connected NRP power sensor.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/This control sets the power sensor connected to the SENSOR port.
/// HIPAR version/This control queries the version of the connected NRP power sensor.
ViStatus _VI_FUNC rssgs_PowerSensorVersion (ViSession instrSession,
                                               ViInt32 sensor, ViChar version[])
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 3), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:SVER?", sensor);

    return rssgs_getViString (instrSession, 
                              buffer,
                              version,
                              1024,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the serial number of the sensor.
/// HIFN noise content and noise timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sensor/This control sets the power sensor connected to the SENSOR port.
/// HIPAR serialNumber/This control displays the power sensor serial number.
ViStatus _VI_FUNC rssgs_PowerSensorSerialNumber (ViSession instrSession,
                                                 ViInt32 sensor,
                                                 ViChar serialNumber[])
{
    ViChar buffer[BUFFER_SIZE];
    INIT;
    
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, sensor, 1, 4), 2);
    
    sprintf (buffer, "SENS" _PERCD ":POW:SNUM?", sensor);

    return rssgs_getViString (instrSession, 
                              buffer,
                              serialNumber,
                              1024,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function sets the RF signal's power level
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR level/This control sets the RF signal's power level.
ViStatus _VI_FUNC rssgs_SetLevel (ViSession instrSession, ViReal64 level)
{

    ViUInt32 supported = 0;
    viFastSettingSupported (instrSession, IdSgsPowPow, &supported);
    if (supported)
        return viFastSetReal64(instrSession, IdSgsPowPow, level);

    return rssgs_setViReal64 (instrSession, 
                              ":POW",
                              level, 0, 0, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the RF signal's power level
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR level/Displays the RF signal's power level.
ViStatus _VI_FUNC rssgs_GetLevel(ViSession instrSession, ViReal64 *level)
{

  ViUInt32 supported = 0;
  viFastQuerySupported(instrSession, IdSgsPowPow, &supported);
  if (supported)
    return viFastQueryReal64(instrSession, IdSgsPowPow, level);

  return rssgs_getViReal64(instrSession,
    ":POW?",
    level,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the RF signal's power level offset
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR level/This control sets the RF signal's power level offset.
ViStatus _VI_FUNC rssgs_SetLevelOffset(ViSession instrSession, ViReal64 level)
{
  return rssgs_setViReal64(instrSession,
    ":POW:OFFS",
    level, -100.0, 100.0,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the RF signal's power level offset
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR level/Displays the RF signal's power level offset.
ViStatus _VI_FUNC rssgs_GetLevelOffset(ViSession instrSession, ViReal64 *level)
{
  return rssgs_getViReal64(instrSession,
    ":POW:OFFS?",
    level,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function configures the level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/This control sets the level mode.
/// HIPAR settingCharacteristic/This control sets the characteristic for the level setting.
/// HIPAR limit/This control sets the upper limit of the RF signal power.
ViStatus _VI_FUNC rssgs_ConfigureLevelSettings (ViSession instrSession,
                                                ViInt32 mode,
                                                ViInt32 settingCharacteristic,
                                                ViReal64 limit)
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rssgs_SetLevelMode (instrSession, mode));

    rssgs_status = rssgs_SetLevelSettingCharacteristic (instrSession, settingCharacteristic);
    REMAP (2, 3);

    rssgs_status = rssgs_SetLevelLimit (instrSession, limit);
    REMAP (2, 4);

    return rssgs_status;
}

//=====================================================================================
/// HIFN This function configures frequency and level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/RF frequency.
/// HIPAR level/RF level.
ViStatus _VI_FUNC rssgs_SetFrequencyAndLevel (ViSession instrSession,
                                              ViReal64  frequency, ViReal64  level )
{
    ViUInt32 supported = 0;
    char buf[100] = {0};

    viFastSettingSupported (instrSession, IdSgsFreqFreq, &supported);

    if (supported)
    {

    	ParaItem ParaList[2] = {
            {IdSgsFreqFreq, PARATYPE_REAL64, 0},
            {IdSgsPowPow,  PARATYPE_REAL64, 0}};
            ParaList[0].pData = (unsigned char *)((void*)&frequency);
            ParaList[1].pData = (unsigned char *)((void*)&level);


            return viFastSetMultiple(instrSession, ParaList, 2);
    }

    sprintf (buf, "FREQ %f;POW %f\n", frequency, level);
    return rssgs_sendCmd(instrSession,buf);
}

//=====================================================================================
/// HIFN This function configures frequency and level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR frequency/RF frequency.
/// HIPAR level/RF level.
ViStatus _VI_FUNC rssgs_TestSpeed (ViSession instrSession, ViReal64  frequency, ViReal64  level )
{
  char buf[100] = {0};
  sprintf (buf, "TEST:SPEED %f,%f\n", frequency, level);
  return rssgs_sendCmd(instrSession,buf);
}


//=====================================================================================
/// HIFN This function sets the level mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/This control sets the level mode.
ViStatus _VI_FUNC rssgs_SetLevelMode (ViSession instrSession, 
                                      ViInt32 mode)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:POW:LMOD", 
                          mode, 
                          LevelModeArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the level mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/Displays the level mode.
ViStatus _VI_FUNC rssgs_GetLevelMode (ViSession instrSession, 
                                      ViInt32 *mode)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:POW:LMOD?", 
                          mode, 
                          LevelModeArr, 
                          VI_ERROR_PARAMETER2);
}
  
//=====================================================================================
/// HIFN This function sets the characteristic for the level setting.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR settingCharacteristic/This control sets the characteristic for the level setting.
ViStatus _VI_FUNC rssgs_SetLevelSettingCharacteristic (ViSession instrSession,
                                                       ViInt32 settingCharacteristic)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:POW:SCH", 
                          settingCharacteristic, 
                          LevelSettingCharacteristicArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the characteristic for the level setting.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR settingCharacteristic/Displays the characteristic for the level setting.
ViStatus _VI_FUNC rssgs_GetLevelSettingCharacteristic (ViSession instrSession,
                                                       ViInt32 *settingCharacteristic)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:POW:SCH?", 
                          settingCharacteristic, 
                          LevelSettingCharacteristicArr, 
                          VI_ERROR_PARAMETER2);
}


/// HIFN This function queries he RF signal's peak envelope power.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR PEP/Returns he RF signal's peak envelope power.
ViStatus _VI_FUNC rssgs_GetPeakEnvelopePower (ViSession instrSession,
                                              ViReal64 *PEP)
{
	return rssgs_getViReal64 (instrSession, 
                              ":SOUR:POW:PEP?", 
                              PEP, 
                              VI_ERROR_PARAMETER2);    
}

/// HIFN This function queries the minimum level range in the current level mode
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR LowerLimit/Returns the minimum level range in the current level mode
ViStatus _VI_FUNC rssgs_GetLevelLowerRange(ViSession instrSession,
                                              ViReal64 *LowerLimit)
{
  return rssgs_getViReal64 (instrSession, 
    ":SOUR:POW:RANG:LOW?", 
    LowerLimit, 
    VI_ERROR_PARAMETER2);    
}

/// HIFN This function queries the maximum level range in the current level mode
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR UpperLimit/Returns the maximum level range in the current level mode
ViStatus _VI_FUNC rssgs_GetLevelUpperRange(ViSession instrSession,
                                           ViReal64 *UpperLimit)
{
  return rssgs_getViReal64 (instrSession, 
    ":SOUR:POW:RANG:UPP?", 
    UpperLimit, 
    VI_ERROR_PARAMETER2);    
}

//=====================================================================================
/// HIFN This function sets the upper limit of the RF signal power.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR limit/This control sets the upper limit of the RF signal power.
ViStatus _VI_FUNC rssgs_SetLevelLimit (ViSession instrSession,
                                       ViReal64 limit)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:POW:LIM:AMPL", 
                              limit, -300.0, 30.0, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the upper limit of the RF signal power.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR limit/Displays the upper limit of the RF signal power.
ViStatus _VI_FUNC rssgs_GetLevelLimit (ViSession instrSession,
                                       ViReal64 *limit)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:POW:LIM:AMPL?", 
                              limit, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the state which the RF output assumes when the instrument is switched on.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR powerOnState/This control sets the state which the RF output assumes when the instrument is switched on.
ViStatus _VI_FUNC rssgs_SetLevelPowerOnState (ViSession instrSession,
                                              ViInt32 powerOnState)
{
    return rssgs_setEnum (instrSession, 
                          ":OUTP:STAT:PON", 
                          powerOnState, 
                          LevelPowerOnStateArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the state which the RF output assumes when the instrument is switched on.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR powerOnState/Displays the state which the RF output assumes when the instrument is switched on.
ViStatus _VI_FUNC rssgs_GetLevelPowerOnState (ViSession instrSession,
                                              ViInt32 *powerOnState)
{
    return rssgs_getEnum (instrSession, 
                          ":OUTP:STAT:PON?", 
                          powerOnState, 
                          LevelPowerOnStateArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the automatic level control.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control sets automatic level control.
ViStatus _VI_FUNC rssgs_SetLevelALCState (ViSession instrSession,
                                          ViInt32 state)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:POW:ALC:STAT", 
                          state, 
                          LevelALCStateArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the automatic level control.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the automatic level control.
ViStatus _VI_FUNC rssgs_GetLevelALCState (ViSession instrSession,
                                          ViInt32 *state)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:POW:ALC:STAT?", 
                          state, 
                          LevelALCStateArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the detector sensitivity of the automatic level control.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control sets the detector sensitivity of the automatic level control.
ViStatus _VI_FUNC rssgs_SetLevelALCDensitivity (ViSession instrSession,
                                                ViInt32 state)
{
  return rssgs_setEnum (instrSession, 
    ":SOURce:POWer:ALC:DSENsitivity", 
    state, 
    LevelALCDSensitivityArr, 
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the detector sensitivity of the automatic level control.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the detector sensitivity of the automatic level control.
ViStatus _VI_FUNC rssgs_GetLevelALCDensitivity (ViSession instrSession,
                                          ViInt32 *state)
{
  return rssgs_getEnum (instrSession, 
    ":SOURce:POWer:ALC:DSENsitivity?", 
    state, 
    LevelALCDSensitivityArr, 
    VI_ERROR_PARAMETER2);
}







//=====================================================================================
/// HIFN This function activate level control for correction purposes.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_ReadjustLevelALC (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, ":SOUR:POW:ALC:SONC");
}

/// HIFN This function sets the state which the attenuator is to assume after the instrument is switched on.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/This control sets the state which the attenuator is to assume after the instrument is switched on.
ViStatus _VI_FUNC rssgs_SetAttenuatorModeAfterOn (ViSession instrSession,
                                                  ViInt32 mode)
{
    return rssgs_setEnum (instrSession, 
                          ":SOUR:POW:ATT:RFOF:MODE", 
                          mode, 
                          attenuatorAfterArr, 
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the state which the attenuator is to assume after the instrument is switched on.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/Displays the state which the attenuator is to assume after the instrument is switched on.
ViStatus _VI_FUNC rssgs_GetAttenuatorModeAfterOn (ViSession instrSession,
                                                  ViInt32 *mode)
{
    return rssgs_getEnum (instrSession, 
                          ":SOUR:POW:ATT:RFOF:MODE?", 
                          mode, 
                          attenuatorAfterArr, 
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function sets the switch-over offset value of the attenuator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR offset/This control sets the switch-over offset value of the attenuator.
ViStatus _VI_FUNC rssgs_SetAttenuatorSwitchOverOffset (ViSession instrSession,
                                                       ViReal64 offset)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:POW:ATT:SOV", 
                              offset, 
                              -10.0, 10.0, 
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the switch-over offset value of the attenuator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR offset/Displays the switch-over offset value of the attenuator.
ViStatus _VI_FUNC rssgs_GetAttenuatorSwitchOverOffset (ViSession instrSession,
                                                       ViReal64 *offset)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:POW:ATT:SOV?", 
                              offset, 
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the minimum level which can be set without the attenuator being adjusted (Attenuator FIXed).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR attenuatorLowerRange/Returns the minimum level which can be set without the attenuator being adjusted (Attenuator FIXed).
ViStatus _VI_FUNC rssgs_GetAttenuatorLowerRange (ViSession instrSession,
                                                 ViReal64 *attenuatorLowerRange)
{
    return rssgs_getViReal64 (instrSession, 
                              "OUTP:AFIX:RANG:LOW?", 
                              attenuatorLowerRange, 
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the maximum level which can be set without the attenuator being adjusted (Attenuator FIXed).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR attenuatorUpperRange/Returns the maximum level which can be set without the attenuator being adjusted (Attenuator FIXed).
ViStatus _VI_FUNC rssgs_GetAttenuatorUpperRange (ViSession instrSession,
                                                 ViReal64 *attenuatorUpperRange)
{
    return rssgs_getViReal64 (instrSession, 
                              "OUTP:AFIX:RANG:UPP?", 
                              attenuatorUpperRange, 
                              VI_ERROR_PARAMETER2);
}

/// HIFN  This function activates/deactivates level correction. Level correction 
/// HIFN  is performed using the table which has been selected with the command 
/// HIFN  CORRection:CSET:SELect.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR correctionState/Activates/deactivates level correction.
ViStatus _VI_FUNC rssgs_SetUserCorrectionDataState(
	ViSession	instrSession,
	ViBoolean	correctionState
)
{
	return rssgs_setViBoolean (instrSession,
                               ":CORR",
                               correctionState,
                               VI_ERROR_PARAMETER2);    
}


/// HIFN  This function queries state of the level correction. Level correction 
/// HIFN  is performed using the table which has been selected with the command 
/// HIFN  CORRection:CSET:SELect.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR correctionState/Returns state of the level correction.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataState(
	ViSession	instrSession,
	ViBoolean*	correctionState
)
{
	return rssgs_getViBoolean (instrSession,
                               ":CORR?",
                               correctionState,
                               VI_ERROR_PARAMETER2);      
}


/// HIFN  This function requests a list of user correction tables. The 
/// HIFN  individual lists are separated by commas. 
/// HIFN     
/// HIFN     The lists are stored with the fixed file extensions *.uco in a 
/// HIFN  directory of the user's choice. The directory applicable to the 
/// HIFN  commands is defined with the command MMEMory:CDIR. A path can also be 
/// HIFN  specified in command :SOUR:CORR:CSET:CAT?, in which case the files in 
/// HIFN  the specified folder are read.      
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR arraySize/Array size of the parameter labeled 'Catalog List'.
/// HIPAR catalogList/Reads back catalog list. Size of the list is limited by the 'Array 
/// HIPAR catalogList/Size' parameter.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataCatalog(
	ViSession	instrSession,
	ViString	directoryPath,
	ViInt32	arraySize,
	ViChar	catalogList[]
)
{
	ViChar buffer[BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, ":CSET:CAT?;*OPC?");
    else
        sprintf (buffer, ":CSET:CAT? '%s';*OPC?", directoryPath);

    return rssgs_getCatalog (instrSession,
                                buffer,
                                arraySize,
                                catalogList,
                                VI_ERROR_PARAMETER4);    
}


/// HIFN  This function selects the table for user correction. Level correction 
/// HIFN  must also be activated with the command 
/// HIFN  SOURce<1|2>:CORRection:CSET:STATe ON.
/// HIFN     
/// HIFN     The lists are stored with the fixed file extensions *.uco in a 
/// HIFN  directory of the user's choice. The directory applicable to the 
/// HIFN  commands is defined with the command MMEMory:CDIR. A path can also be 
/// HIFN  specified in command :SOUR:CORR:CSET:SEL, in which case the files in 
/// HIFN  the specified folder are selected.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR correctionTable/Selects the table name for user correction.
ViStatus _VI_FUNC rssgs_SetUserCorrectionDataTable(
	ViSession	instrSession,
	ViString	correctionTable
)
{
	return rssgs_setViStringQuoted (instrSession, 
								    ":CORR:CSET", 
								    "'",
								    correctionTable, 
								    VI_ERROR_PARAMETER2);   
}


/// HIFN  This function returns the table name for user correction. Level 
/// HIFN  correction must also be activated with the command 
/// HIFN  SOURce<1|2>:CORRection:CSET:STATe ON.
/// HIFN     
/// HIFN     The lists are stored with the fixed file extensions *.uco in a 
/// HIFN  directory of the user's choice. The directory applicable to the 
/// HIFN  commands is defined with the command MMEMory:CDIR. A path can also be 
/// HIFN  specified in command :SOUR:CORR:CSET:SEL, in which case the files in 
/// HIFN  the specified folder are selected.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR correctionTable/Returns the table name for user correction.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataTable (ViSession instrSession,
                                                    ViInt32 arraySize,
                                                    ViChar correctionTable[])
{
	return rssgs_getViStringQuoted (instrSession, 
								    ":CORR:CSET?", 
								    "'",
								    arraySize,
								    correctionTable, 
								    VI_ERROR_PARAMETER3);    
}


/// HIFN  This function transfers the frequency data to the table selected with 
/// HIFN  :CORRection:CSET:SELect. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Defines the number of elements in the 'Frequency Values' array.
/// HIPAR frequencyValues/Frequency correction values. Number of values is defined by 'Array 
/// HIPAR frequencyValues/Size' parameter.
ViStatus _VI_FUNC rssgs_SetUserCorrectionDataFrequencyValues(
	ViSession	instrSession,
	ViInt32	arraySize,
	ViReal64	frequencyValues[]
)
{
	ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (frequencyValues, 3);
    p2buf = pbuffer = (ViChar*) malloc (arraySize * 20 + 20);

    p2buf += sprintf (pbuffer, ":CORR:CSET:DATA:FREQ ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%f,", frequencyValues[i]);

    *--p2buf = '\0';

    return rssgs_sendCmd (instrSession, pbuffer);    
}


/// HIFN  This function queries the frequency data from the table selected with 
/// HIFN  :CORRection:CSET:SELect.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Defines the number of elements in the 'Frequency Values' array.
/// HIPAR frequencyValues/Returns frequency correction values. Number of values is limited by 
/// HIPAR frequencyValues/'Array Size' parameter.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataFrequencyValues(
	ViSession	instrSession,
	ViInt32	arraySize,
	ViReal64	frequencyValues[]
)
{
    return rssgs_getViReal64Array (instrSession,
                                   ":CORR:CSET:DATA:FREQ?",
                                   frequencyValues,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);    
}


/// HIFN  This function queries the number of frequency values in the table 
/// HIFN  selected with :CORRection:CSET:SELect. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR numberofFrequencyValues/Returns the number of frequency values in the selected table.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataNumberofFrequencyValues(
	ViSession	instrSession,
	ViInt32*	numberofFrequencyValues
)
{
    return rssgs_getViInt32 (instrSession,
                             ":CORR:CSET:DATA:FREQ:POIN?",
                             numberofFrequencyValues,
                             VI_ERROR_PARAMETER2);    
}


/// HIFN  This function transfers the level correction data to the table 
/// HIFN  selected with :CORRection:CSET:SELect.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Defines the number of elements in the 'Frequency Values' array.
/// HIPAR levelValues/Level correction values. Number of values is defined by 'Array Size' 
/// HIPAR levelValues/parameter.
ViStatus _VI_FUNC rssgs_SetUserCorrectionDataLevelValues(
	ViSession	instrSession,
	ViInt32	arraySize,
	ViReal64	levelValues[]
)
{
	ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;
    
    CHECKNULL (levelValues, 3);
    p2buf = pbuffer = (ViChar*) malloc (arraySize * 20 + 20);

    p2buf += sprintf (pbuffer, ":CORR:CSET:DATA:POW ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%f,", levelValues[i]);

    *--p2buf = '\0';

    return rssgs_sendCmd (instrSession, pbuffer);    
}


/// HIFN  This function queries the level correction data of the table selected 
/// HIFN  with :CORRection:CSET:SELect.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Defines the number of elements in the 'Level Values' array.
/// HIPAR levelValues/Returns level correction values. Number of values is defined by 'Array 
/// HIPAR levelValues/Size' parameter.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataLevelValues(
	ViSession	instrSession,
	ViInt32	arraySize,
	ViReal64	levelValues[]
)
{
	return rssgs_getViReal64Array (instrSession,
                                   ":CORR:CSET:DATA:POW?",
                                   levelValues,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);    
}


/// HIFN  This function queries the number of level values in the selected 
/// HIFN  table.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR numberofLevelValues/Returns the number of level values in the selected table.
ViStatus _VI_FUNC rssgs_GetUserCorrectionDataNumberofLevelValues(
	ViSession	instrSession,
	ViInt32*	numberofLevelValues
)
{
	return rssgs_getViInt32 (instrSession,
                             ":CORR:CSET:DATA:POW:POIN?",
                             numberofLevelValues,
                             VI_ERROR_PARAMETER2);    
}


/// HIFN  This function deletes the specified table.
/// HIFN     
/// HIFN     The lists are stored with the fixed file extensions *.uco in a 
/// HIFN  directory of the user's choice. The directory applicable to the 
/// HIFN  commands is defined with the command MMEMory:CDIR. A path can also be 
/// HIFN  specified in command :SOUR:CORR:CSET:CAT?, in which case the file in 
/// HIFN  the specified folder is deleted.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR tableName/Specified table name to be deleted.
ViStatus _VI_FUNC rssgs_UserCorrectionDataDeleteTable(
	ViSession	instrSession,
	ViString	tableName
)
{
	return rssgs_setViStringQuoted (instrSession,
                                    ":CORR:CSET:DEL",
                                    "'",
                                    tableName,
                                    VI_ERROR_PARAMETER2);    
}


/// HIFN  This function queries the current value for user correction.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR userCorrectionValue/Returns the current value for user correction.
ViStatus _VI_FUNC rssgs_GetUserCorrectionValue(
	ViSession	instrSession,
	ViReal64*	userCorrectionValue
)
{
    return rssgs_getViReal64 (instrSession,
                              "CORR:VAL?",
                              userCorrectionValue,
                              VI_ERROR_PARAMETER2);    
}


/// HIFN  This function sets the dexchange list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR listMode/This control selects if list mode lists should be imported or 
/// HIPAR listMode/exported. The settings offered below depend on the selected mode.
ViStatus _VI_FUNC rssgs_SetCorrectionImportExportListMode(
	ViSession	instrSession,
	ViInt32	listMode
)
{
    return rssgs_setEnum (instrSession,
                              ":CORR:DEXC:MODE",
                              listMode,
                              dexcListModeArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function returns the dexchange list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR listMode/This control returns if list mode lists should be imported or 
/// HIPAR listMode/exported. The settings offered below depend on the selected mode.
ViStatus _VI_FUNC rssgs_GetCorrectionImportExportListMode(
	ViSession	instrSession,
	ViInt32*	listMode
)
{
	return rssgs_getEnum (instrSession,
                              ":CORR:DEXC:MODE?",
                              listMode,
                              dexcListModeArr,
                              VI_ERROR_PARAMETER3);     
}


/// HIFN  This function calls the File Manager for selecting the list mode list 
/// HIFN  to be exported (source) into an ASCII file or the destination for the 
/// HIFN  ASCII file to be imported (destination) in. 
/// HIFN     
/// HIFN     The list mode files are stored with the fixed file extensions *.lsw in 
/// HIFN  a directory of the user's choice. The directory applicable to the 
/// HIFN  commands is defined with the command MMEMory:CDIR. A path can also be 
/// HIFN  specified in command SOUR:LIST:DEXC:SEL, in which case the files are 
/// HIFN  stored or loaded in the specified directory.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR sourceFile/This control sets the source file.
ViStatus _VI_FUNC rssgs_SetCorrectionImportExportListFile(
	ViSession	instrSession,
	ViString	sourceFile
)
{
	return rssgs_setViStringQuoted (instrSession, 
								    ":CORR:DEXC:SEL", 
								    "'",
								    sourceFile, 
								    VI_ERROR_PARAMETER2);     
}


/// HIFN  This function calls the File Manager for querying the list mode list 
/// HIFN  to be exported (source) into an ASCII file or the destination for the 
/// HIFN  ASCII file to be imported (destination) in.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR sourceFile/This control querying the source file.
ViStatus _VI_FUNC rssgs_GetCorrectionImportExportListFile (ViSession instrSession,
                                                           ViInt32 arraySize,
                                                           ViChar sourceFile[])
{
	return rssgs_getViStringQuoted (instrSession, 
								    ":SOUR:CORR:DEXC:SEL?", 
								    "'",
								    arraySize,
								    sourceFile, 
								    VI_ERROR_PARAMETER3);     
}


/// HIFN  This function starts the export or import of the selected file. 
/// HIFN     When import is selected, the ASCII file is imported as list mode list.
/// HIFN     
/// HIFN     When export is selected, the list mode list is exported into the 
/// HIFN  selected ASCII file.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_CorrectionImportExportExecute(
	ViSession	instrSession
)
{
	return rssgs_sendCmd (instrSession, ":CORR:DEXC:EXEC");     
}


/// HIFN  This function calls the File Manager for selecting the ASCII file to 
/// HIFN  be imported into a list mode list (source) or the ASCII file the list 
/// HIFN  mode list is exported (destination) in.
/// HIFN     
/// HIFN     The ASCII files are stored with the fixed file extensions *.txt or 
/// HIFN  *.csv in a directory of the user's choice. The directory applicable to 
/// HIFN  the commands is defined with the command MMEMory:CDIR.  A path can also 
/// HIFN  be specified in command SOUR:CORR:DEXC:AFIL:SEL, in which case the 
/// HIFN  files are stored or loaded in the specified directory.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR asciiSourceFile/This control sets the ASCII source file.
ViStatus _VI_FUNC rssgs_SetCorrectionImportExportASCIIFile(
	ViSession	instrSession,
	ViString	asciiSourceFile
)
{
	return rssgs_setViStringQuoted (instrSession, 
								    ":CORR:DEXC:AFIL:SEL", 
								    "'",
								    asciiSourceFile, 
								    VI_ERROR_PARAMETER2);    
}


/// HIFN  This function calls the File Manager for querying the ASCII file to be 
/// HIFN  imported into a list mode list (source) or the ASCII file the list mode 
/// HIFN  list is exported (destination) in.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR asciiSourceFile/This control querying the ASCII source file.
ViStatus _VI_FUNC rssgs_GetCorrectionImportExportASCIIFile (ViSession instrSession,
                                                            ViInt32 arraySize,
                                                            ViChar asciiSourceFile[])
{
	return rssgs_getViStringQuoted (instrSession, 
								    ":CORR:DEXC:AFIL:SEL?", 
								    "'",
								    arraySize,
								    asciiSourceFile, 
								    VI_ERROR_PARAMETER3);      
}


/// HIFN  This function selects the file extension of the ASCII file to be 
/// HIFN  imported or exported. Selection TXT (text file) or CSV (Excel file) is 
/// HIFN  available. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR extension/This control selects the file extension of the ASCII file to be 
/// HIPAR extension/imported or exported. Selection TXT (text file) or CSV (Excel file) is 
/// HIPAR extension/available.
ViStatus _VI_FUNC rssgs_SetCorrectionImportExportASCIIExtension(
	ViSession	instrSession,
	ViInt32	extension
)
{
    return rssgs_setEnum (instrSession,
                              ":CORR:DEXC:AFIL:EXT",
                              extension,
                              dexcExtensionArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function queries the selected file extension of the ASCII file to 
/// HIFN  be imported or exported. Selection TXT (text file) or CSV (Excel file) 
/// HIFN  is available. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR extension/This control displays the file extension of the ASCII file to be 
/// HIPAR extension/imported or exported. Selection TXT (text file) or CSV (Excel file) is 
/// HIPAR extension/available.
ViStatus _VI_FUNC rssgs_GetCorrectionImportExportASCIIExtension(
	ViSession	instrSession,
	ViInt32*	extension
)
{
	return rssgs_getEnum (instrSession,
                              ":CORR:DEXC:AFIL:EXT?",
                              extension,
                              dexcExtensionArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function selects the separator between the frequency and level 
/// HIFN  column of the ASCII table.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR separator/This control selects the separator between the frequency and level 
/// HIPAR separator/column of the ASCII table. 
ViStatus _VI_FUNC rssgs_SetCorrectionImportExportASCIIColumnSeparator(
	ViSession	instrSession,
	ViInt32	separator
)
{
	return rssgs_setEnum (instrSession,
                              ":CORR:DEXC:AFIL:SEP:COL",
                              separator,
                              dexcSeparatorArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function queries the selected separator between the frequency and 
/// HIFN  level column of the ASCII table.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR separator/This control displays the separator between the frequency and level 
/// HIPAR separator/column of the ASCII table. 
ViStatus _VI_FUNC rssgs_GetCorrectionImportExportASCIIColumnSeparator(
	ViSession	instrSession,
	ViInt32*	separator
)
{
	return rssgs_getEnum (instrSession,
                              ":CORR:DEXC:AFIL:SEP:COL?",
                              separator,
                              dexcSeparatorArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function selects the decimal separator used in the ASCII data 
/// HIFN  between '.' (decimal point) and ',' (comma) with floating-point 
/// HIFN  numerals.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR decimalSeparator/This control selects the decimal separator used in the ASCII data 
/// HIPAR decimalSeparator/between '.' (decimal point) and ',' (comma) with floating-point 
/// HIPAR decimalSeparator/numerals.  
ViStatus _VI_FUNC rssgs_SetCorrectionImportExportASCIIColumnDecimalSeparator(
	ViSession	instrSession,
	ViInt32	decimalSeparator
)
{
	return rssgs_setEnum (instrSession,
                              "CORR:DEXC:AFIL:SEP:DEC",
                              decimalSeparator,
                              dexcDecimalSeparatorArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function displays the selected decimal separator used in the 
/// HIFN  ASCII data - '.' (decimal point) and ',' (comma) with floating-point 
/// HIFN  numerals.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR decimalSeparator/This control displays the decimal separator used in the ASCII data 
/// HIPAR decimalSeparator/between '.' (decimal point) and ',' (comma) with floating-point 
/// HIPAR decimalSeparator/numerals.  
ViStatus _VI_FUNC rssgs_GetCorrectionImportExportASCIIColumnDecimalSeparator(
	ViSession	instrSession,
	ViInt32*	decimalSeparator
)
{
	return rssgs_getEnum (instrSession,
                              ":CORR:DEXC:AFIL:SEP:DEC?",
                              decimalSeparator,
                              dexcDecimalSeparatorArr,
                              VI_ERROR_PARAMETER3);    
}


/// HIFN  This function requests a list of available ASCII files for 
/// HIFN  export/import of list mode data. The individual files are separated by 
/// HIFN  commas.
/// HIFN      
/// HIFN     The ASCII files are stored with the fixed file extensions *.txt or 
/// HIFN  *.csv in a directory of the user's choice. The directory applicable to 
/// HIFN  the commands is defined with the command MMEMory:CDIR.  A path can also 
/// HIFN  be specified in command SOUR:CORR:DEXC:AFIL:CAT?, in which case the 
/// HIFN  files in the specified directory are read.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR arraySize/Array size of the parameter labeled 'Catalog List'.
/// HIPAR catalogList/Reads back catalog list. Size of the list is limited by the 'Array 
/// HIPAR catalogList/Size' parameter.
ViStatus _VI_FUNC rssgs_CorrectionImportExportCatalog(
	ViSession	instrSession,
	ViString	directoryPath,
	ViInt32	arraySize,
	ViChar	catalogList[]
)
{
	ViChar      buffer[BUFFER_SIZE];
    
    if (!directoryPath)
        sprintf (buffer, ":CORR:DEXC:AFIL:CAT?;*OPC?");
    else
        sprintf (buffer, ":CORR:DEXC:AFIL:CAT? '%s';*OPC?", directoryPath);

    return rssgs_getCatalog (instrSession,
                                     buffer,
                                     arraySize,
                                     catalogList,
                                     VI_ERROR_PARAMETER4); 
        
}

//=====================================================================================
/// HIFN This function configures the IQ modulator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR IQModulatorState/This control sets the state of the IQ modulator.
/// HIPAR IQWideband/This control sets optimized settings for wideband modulation signals.
ViStatus _VI_FUNC rssgs_ConfigureIQ (ViSession instrSession,
                                     ViBoolean IQModulatorState,
                                     ViBoolean IQWideband)
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rssgs_SetIQModulatorState (instrSession, IQModulatorState));

    rssgs_status = rssgs_SetIQWideband (instrSession, IQWideband);
    REMAP (2, 3);

    return rssgs_status;
}

//=====================================================================================
/// HIFN This function configures the IQ Impairments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control sets the impairment or correction values LEAKage, QUADrature and IQRatio for the digital baseband signal.
/// HIPAR iOffset/This control sets the carrier offset for the I-channel of the I/Q modulation.
/// HIPAR qOffset/This control sets the carrier offset for the Q-channel of the I/Q modulation. 
/// HIPAR gainImbalance/This control sets the ratio of I modulation to Q modulation (gain imbalance).
/// HIPAR quadratureOffset/This control sets the quadrature offset for the digital I/Q signal. 
ViStatus _VI_FUNC rssgs_ConfigureIQImpairments (ViSession instrSession,
                                                ViBoolean state,
                                                ViReal64 iOffset,
                                                ViReal64 qOffset,
                                                ViReal64 gainImbalance,
                                                ViReal64 quadratureOffset)
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rssgs_SetIQImpairmentsState (instrSession, state));
    REMAP (2, 2);

    rssgs_status = rssgs_SetIQImpairmentsIOffset (instrSession, iOffset);
    REMAP (2, 3);

    rssgs_status = rssgs_SetIQImpairmentsQOffset (instrSession, qOffset);
    REMAP (2, 4);

    rssgs_status = rssgs_SetIQImpairmentsGainImbalance (instrSession, gainImbalance);
    REMAP (2, 5);

    rssgs_status = rssgs_SetIQImpairmentsQuadratureOffset (instrSession, quadratureOffset);
    REMAP (2, 6);

    return rssgs_status;
}

//=====================================================================================
/// HIFN This function sets the state of the IQ modulator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR IQModulatorState/This control sets the state of the IQ modulator.
ViStatus _VI_FUNC rssgs_SetIQModulatorState (ViSession instrSession,
                                             ViBoolean IQModulatorState)
{

    ViUInt32 supported = 0;
    viFastSettingSupported (instrSession, IdSgsModState, &supported);
    if (supported)
        return viFastSetInt32(instrSession, IdSgsModState, IQModulatorState);

    return rssgs_setViBoolean (instrSession, 
                               ":SOUR:IQ:STAT", 
                               IQModulatorState, 
                               VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the state of the IQ modulator.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR IQModulatorState/Displays the state of the IQ modulator.
ViStatus _VI_FUNC rssgs_GetIQModulatorState (ViSession instrSession,
                                             ViBoolean *IQModulatorState)
{
    ViUInt32 supported = 0;
    viFastQuerySupported(instrSession, IdSgsModState, &supported);
    if (supported)
    {
      ViInt32 sel = 0;
      int status = 0;
      status = viFastQueryInt32(instrSession, IdSgsModState, &sel);
      *IQModulatorState = (sel != 0);
      return status;
    }

    return rssgs_getViBoolean (instrSession, 
                               ":SOUR:IQ:STAT?", 
                               IQModulatorState, 
                               VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the impairment or correction values LEAKage, QUADrature and IQRatio for the digital baseband signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control sets the impairment or correction values LEAKage, QUADrature and IQRatio for the digital baseband signal.
ViStatus _VI_FUNC rssgs_SetIQImpairmentsState (ViSession instrSession,
                                               ViBoolean state)
{
    return rssgs_setViBoolean (instrSession, 
                               ":SOUR:IQ:IMP:STAT", 
                               state, 
                               VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the impairment or correction values LEAKage, QUADrature and IQRatio for the digital baseband signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the impairment or correction values LEAKage, QUADrature and IQRatio for the digital baseband signal.
ViStatus _VI_FUNC rssgs_GetIQImpairmentsState (ViSession instrSession,
                                               ViBoolean *state)
{
    return rssgs_getViBoolean (instrSession, 
                               ":SOUR:IQ:IMP:STAT?", 
                               state, 
                               VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the carrier offset for the I-channel of the I/Q modulation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR iOffset/This control sets the carrier offset for the I-channel of the I/Q modulation.
ViStatus _VI_FUNC rssgs_SetIQImpairmentsIOffset (ViSession instrSession,
                                                 ViReal64 iOffset)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:LEAK:I", 
                              iOffset, -10.0, 10.0, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the carrier offset for the I-channel of the I/Q modulation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR iOffset/Displays the carrier offset for the I-channel of the I/Q modulation.
ViStatus _VI_FUNC rssgs_GetIQImpairmentsIOffset (ViSession instrSession,
                                                 ViReal64 *iOffset)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:LEAK:I?", 
                              iOffset, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the carrier offset for the Q-channel of the I/Q modulation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR qOffset/This control sets the carrier offset for the Q-channel of the I/Q modulation.
ViStatus _VI_FUNC rssgs_SetIQImpairmentsQOffset (ViSession instrSession,
                                                 ViReal64 qOffset)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:LEAK:Q", 
                              qOffset, -10.0, 10.0, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the carrier offset for the Q-channel of the I/Q modulation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR qOffset/Displays the carrier offset for the Q-channel of the I/Q modulation.

ViStatus _VI_FUNC rssgs_GetIQImpairmentsQOffset (ViSession instrSession,
                                                 ViReal64 *qOffset)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:LEAK:Q?", 
                              qOffset,
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the ratio of I modulation to Q modulation (gain imbalance).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR gainImbalance/This control sets the ratio of I modulation to Q modulation (gain imbalance).
ViStatus _VI_FUNC rssgs_SetIQImpairmentsGainImbalance (ViSession instrSession,
                                                       ViReal64 gainImbalance)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:IQR", 
                              gainImbalance, -1.000, 1.000, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the ratio of I modulation to Q modulation (gain imbalance).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR gainImbalance/Displays the ratio of I modulation to Q modulation (gain imbalance).
ViStatus _VI_FUNC rssgs_GetIQImpairmentsGainImbalance (ViSession instrSession,
                                                       ViReal64 *gainImbalance)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:IQR?", 
                              gainImbalance, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the quadrature offset for the digital I/Q signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR quadratureOffset/This control sets the quadrature offset for the digital I/Q signal.
ViStatus _VI_FUNC rssgs_SetIQImpairmentsQuadratureOffset (ViSession instrSession,
                                                          ViReal64 quadratureOffset)
{
    return rssgs_setViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:QUAD:ANGL", 
                              quadratureOffset, -10.0, 10.0, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the quadrature offset for the digital I/Q signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR quadratureOffset/Displays the quadrature offset for the digital I/Q signal.
ViStatus _VI_FUNC rssgs_GetIQImpairmentsQuadratureOffset (ViSession instrSession,
                                                          ViReal64 *quadratureOffset)
{
    return rssgs_getViReal64 (instrSession, 
                              ":SOUR:IQ:IMP:QUAD:ANGL?", 
                              quadratureOffset,
                              VI_ERROR_PARAMETER2);
}


//=====================================================================================
/// HIFN This function sets optimized settings for wideband modulation signals.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR IQWideband/This control sets optimized settings for wideband modulation signals.
ViStatus _VI_FUNC rssgs_SetIQWideband (ViSession instrSession,
                                       ViBoolean IQWideband)
{
    ViUInt32 supported = 0;
    viFastSettingSupported (instrSession, IdSgsIqWidebandState, &supported);
    if (supported)
      return viFastSetInt32(instrSession, IdSgsIqWidebandState, IQWideband);

    return rssgs_setViBoolean (instrSession, 
                               ":SOUR:IQ:WBST", 
                               IQWideband, 
                               VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries optimized settings for wideband modulation signals.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR IQWideband/Displays optimized settings for wideband modulation signals.
ViStatus _VI_FUNC rssgs_GetIQWideband (ViSession instrSession,
                                       ViBoolean *IQWideband)
{
    ViUInt32 supported = 0;
    viFastQuerySupported(instrSession, IdSgsIqWidebandState, &supported);
    if (supported)
    {
      ViInt32 sel = 0;
      int status = 0;
      status = viFastQueryInt32(instrSession, IdSgsIqWidebandState, &sel);
      *IQWideband = (sel != 0);
      return status;
    }


    return rssgs_getViBoolean (instrSession, 
                               ":SOUR:IQ:WBST?", 
                               IQWideband, 
                               VI_ERROR_PARAMETER2);
}

/// HIFN This function sets the crest factor of the IQ modulation signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR crestFactor/Sets the crest factor of the IQ modulation signal.
ViStatus _VI_FUNC rssgs_SetIQCrestFactor (ViSession instrSession,
                                          ViReal64 crestFactor)
{
	return rssgs_setViReal64 (instrSession, 
                              ":SOUR:IQ:CRES", 
                              crestFactor, 0.0, 35.0, 
                              VI_ERROR_PARAMETER2);    
}

/// HIFN This function queries the crest factor of the IQ modulation signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR crestFactor/Returns the crest factor of the IQ modulation signal.
ViStatus _VI_FUNC rssgs_GetIQCrestFactor (ViSession instrSession,
                                          ViReal64 *crestFactor)
{
	return rssgs_getViReal64 (instrSession, 
                              ":SOUR:IQ:CRES?", 
                              crestFactor, 
                              VI_ERROR_PARAMETER2);    
}    



/// HIFN This function pass a lock request id which uniquely identifies the controller to the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR lockID/This control uniquely identifies the controller to the instrument.
/// HIPAR result/Pass a lock request id which uniquely identifies the controller to the instrument.
ViStatus _VI_FUNC rssgs_LockInstrument (ViSession instrSession, ViInt32 lockID,
                                        ViInt32 *result)
{
    ViChar buffer[BUFFER_SIZE];
    
    sprintf (buffer, "LOCK? " _PERCD "", lockID);
    
    return rssgs_getViInt32 (instrSession, 
                             buffer,
                             result,
                             VI_ERROR_PARAMETER3);
}

/// HIFN This function unlock device.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR lockID/This control uniquely identifies the controller to the instrument.
ViStatus _VI_FUNC rssgs_UnlockInstrument (ViSession instrSession,
                                          ViInt32 lockID)
{
    ViChar buffer[BUFFER_SIZE];
    
    sprintf (buffer, "UNL " _PERCD "", lockID);
    
    return rssgs_sendCmd (instrSession, buffer);
}


/// HIFN This function queries the list  of the symbolic names of the available extension devices (the various entries are separated by commas).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR symbolicNames/Returns the  list of the symbolic names of the available extension devices (the various entries are separated by commas).
ViStatus _VI_FUNC rssgs_EXTensionInstNames (ViSession instrSession, 
                                              ViInt32 arraySize, 
                                              ViChar _VI_FAR symbolicNames[])
{
  ViChar  buffer [BUFFER_SIZE];

  sprintf (buffer, ":EXT:INST:NAME?");

  return rssgs_getCatalog (instrSession,
    buffer,
    (ViUInt32) arraySize,
    symbolicNames,
    VI_ERROR_PARAMETER3);
}

/// HIFN This function queries the list  of the lan names of the available extension devices (the various entries are separated by commas).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR lanNames/Returns the  list of the symbolic names of the available extension devices (the various entries are separated by commas).
ViStatus _VI_FUNC rssgs_EXTensionInstLanNames (ViSession instrSession, 
                                            ViInt32 arraySize, 
                                            ViChar _VI_FAR lanNames[])
{
  ViChar  buffer [BUFFER_SIZE];

  sprintf (buffer, ":EXT:INST:REM:LAN:NAME?");

  return rssgs_getCatalog (instrSession,
    buffer,
    (ViUInt32) arraySize,
    lanNames,
    VI_ERROR_PARAMETER3);
}



/// HIFN  Displays the hardware interfaces used by the remote channels.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Array size of the parameter labeled 'Hardware Channel'.
/// HIPAR hardwareChannel/Displays the hardware interfaces used by the remote channels.
ViStatus _VI_FUNC rssgs_ExtensionInstHardwareChannels (ViSession instrSession,
                                                       ViInt32 arraySize,
                                                       ViInt32 hardwareChannel[],
                                                       ViInt32 *returnedValues)
{
  ViStatus  rssgs_status;
  ViChar    buffer[BUFFER_SIZE] = {0};
  ViChar*  	p2buf;
  ViInt32		i = 0;

  rssgs_status = rssgs_getViString (instrSession, 
    ":EXT:INST:REM:CHAN?", 
    buffer, 
    BUFFER_SIZE, 
    VI_ERROR_PARAMETER3);

  p2buf = strtok (buffer, ",");

  while (p2buf && i < arraySize)
  {
    hardwareChannel[i++] = rssgs_StringIndex (ExtensionHardwareChannelArr, p2buf);
    p2buf = strtok (NULL, ",\n");
  }

  *returnedValues = i;

  return rssgs_status;
}

/// HIFN  Displays the list of the serial numbers of the available extension devices
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR arraySize/Array size of the parameter labeled 'Hardware Channel'.
/// HIPAR serialNumbers/Displays the list of the serial numbers of the available extension devices
ViStatus _VI_FUNC rssgs_ExtensionInstSerialNumbers (ViSession instrSession,
                                                       ViInt32 arraySize,
                                                       ViInt32 serialNumbers[],
                                                       ViInt32 *returnedValues)
{
  ViStatus  rssgs_status;
  ViChar    buffer[BUFFER_SIZE] = {0};
  ViChar*  	p2buf;
  ViInt32		i = 0;

  rssgs_status = rssgs_getViString (instrSession, 
    ":EXT:INST:REM:SER?", 
    buffer, 
    BUFFER_SIZE, 
    VI_ERROR_PARAMETER3);

  p2buf = strtok (buffer, ",");

  while (p2buf && i < arraySize)
  {
    serialNumbers[i++] = atoi (p2buf);
    p2buf = strtok (NULL, ",\n");
  }

  *returnedValues = i;

  return rssgs_status;
}


//=====================================================================================
/// HIFN This function starts or stops a scan for extension instruments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This function starts or stops a scan for extension instruments.
ViStatus _VI_FUNC rssgs_SetExtensionInstScanState (ViSession instrSession, ViBoolean state)
{
  return rssgs_setViBoolean (instrSession,
    ":EXT:INST:SCAN",
    state,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the state of the scan for extension instruments
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the state of the scan for extension instruments
ViStatus _VI_FUNC rssgs_GetExtensionInstScanState (ViSession instrSession, ViBoolean *state)
{
  return rssgs_getViBoolean ( instrSession,
    ":EXT:INST:SCAN?",
    state,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the busy state of the selected extension
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the busy state of the selected extension
ViStatus _VI_FUNC rssgs_GetExtensionBusyState (ViSession instrSession, ViBoolean *state)
{
  return rssgs_getViBoolean ( instrSession,
    ":EXT:BUSY:STAT?",
    state,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the remote state of the selected extension
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the remote state of the selected extension
ViStatus _VI_FUNC rssgs_GetExtensionRemoteState (ViSession instrSession, ViBoolean *state)
{
  return rssgs_getViBoolean ( instrSession,
    ":EXT:REM:STAT?",
    state,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the index of the selected extension
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR selIndex/Displays the index of the selected extension
ViStatus _VI_FUNC rssgs_GetExtensionSelect (ViSession instrSession, ViInt32 *selIndex)
{
  return rssgs_getViInt32(instrSession,
    ":EXT:SEL?",
    selIndex,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the index of the selected extension
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR selIndex/This function sets the index of the selected extension
ViStatus _VI_FUNC rssgs_SetExtensionSelect (ViSession instrSession, ViInt32 selIndex)
{
  return rssgs_setViInt32 ( instrSession,
    ":EXT:SEL",
    selIndex,
    0,
    0,
    VI_ERROR_PARAMETER2);
}


//=====================================================================================
/// HIFN This function sends a SCPI command to the extension instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR scpicommand/This control sends a SCPI command to the extension instrument
ViStatus _VI_FUNC rssgs_ExtensionSend (ViSession instrSession,
                                       ViString scpicommand)
{
  return rssgs_setViStringQuoted (instrSession, 
    ":EXT:SEND", 
    "\"",
    scpicommand, 
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the extension instrument by sending a SCPI command.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR scpicommand/This control sets the scpi command to send.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR reply/Displays the reply of the instrument.
ViStatus _VI_FUNC rssgs_ExtensionQuery (ViSession instrSession,
                                        ViString scpicommand,
                                        ViInt32 arraySize,
                                        ViChar reply[])
{
  ViChar buffer[BUFFER_SIZE];
  sprintf (buffer, "EXT:SEND? \"%s\"",scpicommand);

  return rssgs_getViString (instrSession, 
    buffer, 
    reply, 
    arraySize,
    VI_ERROR_PARAMETER3);
}


//=====================================================================================
/// HIFN This function performs all adjustments which affect the synthesis.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the calibration routine to be finished and have obtained calibration result.
/// HIPAR result/Returns result of the adjustment.
ViStatus _VI_FUNC rssgs_CALibration (ViSession instrSession,
                                     ViPChar command,
                                     ViInt32 timeout,
                                     ViInt32 *result)
{
  ViStatus     rssgs_status = VI_SUCCESS;
  ViChar       buffer[BUFFER_SIZE];
  unsigned int oldTimeout = 0;

  if (rssgs_invalidViInt32Range (instrSession, timeout, 0, 600000) == VI_TRUE)
    return VI_ERROR_PARAMETER2;
  if (result == NULL)
    return VI_ERROR_PARAMETER3;

  CHECKERR( rsGetAttribute (instrSession, VI_ATTR_TMO_VALUE, &oldTimeout));
  CHECKERR( rsSetAttribute (instrSession, VI_ATTR_TMO_VALUE, timeout));

  CHECKERR (rsPrintf (instrSession, "%s\n", command));
  CHECKERR (rsScanf (instrSession, "%[^\r\n]", buffer));

  CHECKERR( rsSetAttribute (instrSession, VI_ATTR_TMO_VALUE, oldTimeout));

  *result = 0;
  if (sscanf (buffer, "" _PERCD "", result) != 1)
    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;

  if (*result == 1) /* Calibration failed. */
    rssgs_status = RSSGS_ERROR_CALIBRATION_FAILED;

  return rssgs_status;
}



//=====================================================================================
/// HIFN This function perform all internal adjustments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the calibration routine to be finished and have obtained calibration result.
/// HIPAR result/Returns result of the adjustment
ViStatus _VI_FUNC rssgs_CALibrationAll (ViSession instrSession,
                                        ViInt32 timeout, 
                                        ViInt32 *result)
{
  return rssgs_CALibration(instrSession, "CAL:ALL?;*OPC?", timeout, result); 
}

//=====================================================================================
/// HIFN This function perform internal level adjustment.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the calibration routine to be finished and have obtained calibration result.
/// HIPAR result/Returns result of the adjustment.
ViStatus _VI_FUNC rssgs_CALibrationLevel (ViSession instrSession,
                                          ViInt32 timeout, 
                                          ViInt32 *result)
{
    return rssgs_CALibration(instrSession, "CAL:LEV:MEAS?;*OPC?", timeout, result); 
}    

//=====================================================================================
/// HIFN This function performs all adjustments which affect the synthesis.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the calibration routine to be finished and have obtained calibration result.
/// HIPAR result/Returns result of the adjustment.
ViStatus _VI_FUNC rssgs_CALibrationSynthesis (ViSession instrSession,
                                                   ViInt32 timeout,
                                                   ViInt32 *result)
{
  return rssgs_CALibration(instrSession, "CAL:FREQ?;*OPC?", timeout, result); 
}

/// HIFN This function starts the adjustment for the I/Q modulator for the entire frequency range. The I/Q modulator is adjusted with respect to carrier leakage, I/Q imbalance and quadrature.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the calibration routine to be finished and have obtained calibration result. If the length of time required for calibration exceeds the timeout value, then the function will return with a timeout error and the instrument will continue with calibration.
/// HIPAR result/Returns result of the adjustment. The response "0" indicates error-free adjustment, and the response "1" means that an error occurred during adjustment.
ViStatus _VI_FUNC rssgs_CALibrationIQModulator (ViSession instrSession,
                                                ViInt32 timeout, ViInt32 *result)
{
  return rssgs_CALibration(instrSession, "CAL:IQM:FULL?;*OPC?", timeout, result); 

}

/// HIFN This function starts the adjustment for the I/Q modulator for the current frequency. The I/Q modulator is adjusted with respect to carrier leakage, I/Q imbalance and quadrature.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the calibration routine to be finished and have obtained calibration result. If the length of time required for calibration exceeds the timeout value, then the function will return with a timeout error and the instrument will continue with calibration.
/// HIPAR result/Returns result of the adjustment. The response "0" indicates error-free adjustment, and the response "1" means that an error occurred during adjustment.
ViStatus _VI_FUNC rssgs_CALibrationIQModulatorForCurrentFrequency
             (ViSession instrSession, ViInt32 timeout, ViInt32 *result)
{
  return rssgs_CALibration(instrSession, "CAL:IQM:LOC?;*OPC?", timeout, result); 
}




//=====================================================================================
/// HIFN This function queries the delta temperature since the last performed IQ adjustment.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR temperature/Displays the delta temperature since the last performed IQ adjustment
ViStatus _VI_FUNC rssgs_GetCalibrationIQTemperature (ViSession instrSession,
                                      ViReal64 *temperature)
{
  return rssgs_getViReal64 (instrSession,
    ":CAL:IQM:TEMP?",
    temperature,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the delta temperature since the last performed level adjustment.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR temperature/Displays the delta temperature since the last performed level adjustment
ViStatus _VI_FUNC rssgs_GetCalibrationLevelTemperature (ViSession instrSession,
                                                     ViReal64 *temperature)
{
  return rssgs_getViReal64 (instrSession,
    ":CAL:LEV:TEMP?",
    temperature,
    VI_ERROR_PARAMETER2);
}
//=====================================================================================
/// HIFN This function queries the delta temperature since the last performed synthesis adjustment.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR temperature/Displays the delta temperature since the last performed synthesis adjustment
ViStatus _VI_FUNC rssgs_GetCalibrationSynthesisTemperature (ViSession instrSession,
                                                     ViReal64 *temperature)
{
  return rssgs_getViReal64 (instrSession,
    ":CAL:FREQ:TEMP?",
    temperature,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN Enables you to run an internal adjustment only on the extension instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This function enables you to run an internal adjustment only on the extension instrument.
ViStatus _VI_FUNC rssgs_SetCalibrationExtensionOnlyState (ViSession instrSession, ViBoolean state)
{
  return rssgs_setViBoolean (instrSession,
    ":CAL:OEXT",
    state,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the state of the extension only calibration parameter
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Displays the state of the extension only calibration parameter
ViStatus _VI_FUNC rssgs_GetCalibrationExtensionOnlyState (ViSession instrSession, ViBoolean *state)
{
  return rssgs_getViBoolean ( instrSession,
    ":CAL:OEXT?",
    state,
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the device's operating system name
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR osname/Displays the device's operating system name
ViStatus _VI_FUNC rssgs_GetSYSTemOperatingSystem (ViSession instrSession,
                                            ViInt32 arraySize,
                                            ViChar osname[])
{
  return rssgs_getViStringQuoted (instrSession, 
    ":SYST:OSYS?", 
    "\"",
    arraySize,
    osname, 
    VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the path to the user accessible memory
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR path/Displays path to the user memory
ViStatus _VI_FUNC rssgs_GetSYSTemUserMemoryPath (ViSession instrSession,
                                           ViInt32 arraySize,
                                           ViChar path[])
{
  return rssgs_getViStringQuoted (instrSession, 
    ":SYST:MMEM:PATH:USER?", 
    "\"",
    arraySize,
    path, 
    VI_ERROR_PARAMETER3);
}



//=====================================================================================
/// HIFN This function gets network settings from system.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_GetNetworkSystemSettings (ViSession instrSession)
{
    // leave for compatibility
    instrSession = 0;
    return VI_SUCCESS;
}

//=====================================================================================
/// HIFN This function queries the MAC Address.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR MACAddress/Displays the MAC Address.
ViStatus _VI_FUNC rssgs_GetNetworkMACAddress (ViSession instrSession,
                                              ViInt32 arraySize,
                                              ViChar MACAddress[])
{
    return rssgs_getViStringQuoted (instrSession, 
                              ":SYST:COMM:NETW:MAC?", 
                              "\"",
                              arraySize, 
                              MACAddress, 
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function configures the IP Address.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR addressMode/This control sets the IP Address Mode.
/// HIPAR IPAddress/This control sets the IP Address.
/// HIPAR subnetMask/This control sets the Subnet Mask.
/// HIPAR defaultGateway/This control sets the Default Gateway.
ViStatus _VI_FUNC rssgs_ConfigureNetworkIP (ViSession instrSession,
                                            ViInt32 addressMode,
                                            ViString IPAddress,
                                            ViString subnetMask,
                                            ViString defaultGateway)
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rssgs_SetNetworkIPAddressMode (instrSession, addressMode));

    rssgs_status = rssgs_SetNetworkIPAddress (instrSession, IPAddress);
    REMAP (2, 3);

    rssgs_status = rssgs_SetNetworkIPSubnetMask (instrSession, subnetMask);
    REMAP (2, 4);
    
    rssgs_status = rssgs_SetNetworkIPDefaultGateway (instrSession, defaultGateway);
    REMAP (2, 5);

    return rssgs_status;
}

/// HIFN This function restarts the network interface
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_NetworkRestart (ViSession instrSession)
{
  return rssgs_sendCmd (instrSession, "SYST:COMM:NETW:REST");
}

/// HIFN This function returns the resource string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/Array size of the parameter labeled 'Resource'.
/// HIPAR resource/This control displays the resource string.
ViStatus _VI_FUNC rssgs_GetNetworkResource(
    ViSession   instrSession,
    ViUInt32 arraySize,
    ViChar _VI_FAR  resource[]
)
{
    return rssgs_getViStringQuoted (instrSession,
                              "SYST:COMM:NETW:RES?",
                              "\"",
                              arraySize,
                              resource,
                              VI_ERROR_PARAMETER3);
}

/// HIFN This function returns the network configuration state.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR configuration/This control displays the network configuration state.
ViStatus _VI_FUNC rssgs_GetNetworkConfigurationState (ViSession instrSession,
                                                      ViBoolean *configuration)
{
    return rssgs_getViBoolean (instrSession,
                               "SYST:COMM:NETW:STAT?",
                               configuration,
                               VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the Hostname of the network.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR hostname/This control sets the Hostname of the network.
ViStatus _VI_FUNC rssgs_SetNetworkCommonHostname (ViSession instrSession,
                                                  ViString hostname)
{
    return rssgs_setViStringQuoted (instrSession, 
                              ":SYST:COMM:NETW:COMM:HOST", 
                              "'",
                              hostname, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the Hostname of the network.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR hostname/Displays the Hostname of the network.
ViStatus _VI_FUNC rssgs_GetNetworkCommonHostname (ViSession instrSession,
                                                  ViInt32 arraySize,
                                                  ViChar hostname[])
{
    return rssgs_getViStringQuoted (instrSession, 
                              ":SYST:COMM:NETW:COMM:HOST?", 
                              "\"",
                              arraySize,
                              hostname, 
                              VI_ERROR_PARAMETER3);
}



//=====================================================================================
/// HIFN This function sets the IP Address Mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR addressMode/This control sets the IP Address Mode.
ViStatus _VI_FUNC rssgs_SetNetworkIPAddressMode (ViSession instrSession,
                                                 ViInt32 addressMode)
{
    return rssgs_setEnum (instrSession, 
                          "SYST:COMM:NETW:IPAD:MODE", 
                          addressMode,
                          ConfigureNetworkIPArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the IP Address Mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR addressMode/Displays the IP Address Mode.
ViStatus _VI_FUNC rssgs_GetNetworkIPAddressMode (ViSession instrSession,
                                                 ViInt32 *addressMode)
{
    return rssgs_getEnum (instrSession, 
                          "SYST:COMM:NETW:IPAD:MODE?", 
                          addressMode,
                          ConfigureNetworkIPArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the IP Adrress.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR IPAddress/This control sets the IP Address.
ViStatus _VI_FUNC rssgs_SetNetworkIPAddress (ViSession instrSession,
                                             ViString IPAddress)
{
    return rssgs_setViStringQuoted (instrSession,
                              ":SYST:COMM:NETW:IPAD",
                              "'",
                              IPAddress,
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the IP Adrress.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR IPAddress/Displays the IP Address.
ViStatus _VI_FUNC rssgs_GetNetworkIPAddress (ViSession instrSession,
                                             ViInt32 arraySize,
                                             ViChar IPAddress[])
{
    return rssgs_getViString (instrSession,
                              ":SYST:COMM:NETW:IPAD?",
                              IPAddress,
                              arraySize,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function sets the Subnet Mask.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR subnetMask/This control sets the Subnet Mask.
ViStatus _VI_FUNC rssgs_SetNetworkIPSubnetMask (ViSession instrSession,
                                                ViString subnetMask)
{
    return rssgs_setViStringQuoted (instrSession, 
                              ":SYST:COMM:NETW:IPAD:SUBN:MASK", 
                              "'",
                              subnetMask, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the Subnet Mask.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR subnetMask/Displays the Subnet Mask.
ViStatus _VI_FUNC rssgs_GetNetworkIPSubnetMask (ViSession instrSession,
                                                ViInt32 arraySize,
                                                ViChar subnetMask[])
{
    return rssgs_getViString (instrSession, 
                              ":SYST:COMM:NETW:IPAD:SUBN:MASK?", 
                              subnetMask, 
                              arraySize, 
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function sets the Default Gateway.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR defaultGateway/This control sets the Default Gateway.
ViStatus _VI_FUNC rssgs_SetNetworkIPDefaultGateway (ViSession instrSession,
                                                    ViString defaultGateway)
{
    return rssgs_setViStringQuoted (instrSession, 
                              ":SYST:COMM:NETW:IPAD:GAT", 
                              "'",
                              defaultGateway, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function queries the Default Gateway.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR defaultGateway/Displays the Default Gateway.
ViStatus _VI_FUNC rssgs_GetNetworkIPDefaultGateway (ViSession instrSession,
                                                    ViInt32 arraySize,
                                                    ViChar defaultGateway[])
{
    return rssgs_getViString (instrSession, 
                              ":SYST:COMM:NETW:IPAD:GAT?", 
                              defaultGateway, 
                              arraySize, 
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function accepts the IP Address Settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_AcceptNetworkIPSettings (ViSession instrSession)
{
	instrSession = 0;
  // leave for compatibility
  return VI_SUCCESS;
}


//=====================================================================================
/// HIFN This function sets the baud rate of the serial port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR baudrate/Sets the baud rate

ViStatus _VI_FUNC rssgs_SetSerialBaudRate (ViSession instrSession,
                                           ViInt32 baudrate)
{
  return rssgs_setEnum (instrSession, 
    "SYST:COMM:SER:BAUD", 
    baudrate,
    SerialBaudRateArr, 
    VI_ERROR_PARAMETER2);

}


//=====================================================================================
/// HIFN This function queries the baud rate of the serial port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR baudrate/Displays the baud rate

ViStatus _VI_FUNC rssgs_GetSerialBaudRate (ViSession instrSession,
                                           ViInt32* baudrate)
{
  return rssgs_getEnum (instrSession, 
    "SYST:COMM:SER:BAUD?", 
    baudrate,
    SerialBaudRateArr, 
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the parity setting of the serial port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR parity/Displays the parity setting

ViStatus _VI_FUNC rssgs_SetSerialParity (ViSession instrSession,
                                         ViInt32 parity)
{
  return rssgs_setEnum (instrSession, 
    "SYST:COMM:SER:PAR", 
    parity,
    SerialParityArr, 
    VI_ERROR_PARAMETER2);

}


//=====================================================================================
/// HIFN This function queries the parity setting of the serial port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR parity/Displays the parity setting

ViStatus _VI_FUNC rssgs_GetSerialParity (ViSession instrSession,
                                         ViInt32* parity)
{
  return rssgs_getEnum (instrSession, 
    "SYST:COMM:SER:PAR?", 
    parity,
    SerialParityArr, 
    VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the stop bits setting of the serial port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR stopbits/Sets the stop bits

ViStatus _VI_FUNC rssgs_SetSerialStopBits (ViSession instrSession,
                                           ViInt32 stopbits)
{
  return rssgs_setEnum (instrSession, 
    "SYST:COMM:SER:SBIT", 
    stopbits,
    SerialStopBitsArr, 
    VI_ERROR_PARAMETER2);

}


//=====================================================================================
/// HIFN This function queries the stop bits setting of the serial port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR stopbits/Displays the stop bits setting

ViStatus _VI_FUNC rssgs_GetSerialStopBits (ViSession instrSession,
                                           ViInt32* stopbits)
{
  return rssgs_getEnum (instrSession, 
    "SYST:COMM:SER:SBIT?", 
    stopbits,
    SerialStopBitsArr, 
    VI_ERROR_PARAMETER2);
}













//=====================================================================================
/// HIFN CLEAR STATUS sets the status byte (STB), the standard event register (ESR), and the EVENt part of the QUEStionable and OPERation register to zero.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_ClearStatus (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, 
                          "*CLS");    
}

//=====================================================================================
/// HIFN EVENT STATUS ENABLE sets the event status enable register to the specified value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/EVENT STATUS ENABLE sets the event status enable register to the specified value.
ViStatus _VI_FUNC rssgs_EventStatusEnable (ViSession instrSession,
                                           ViInt32 registerValue)
{
    return rssgs_setViInt32 (instrSession, 
                             "*ESE", 
                             registerValue, 0, 255, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN STANDARD EVENT STATUS QUERY returns the content of the event status register in decimal form (0...255) and then sets the register to zero.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Returns the content of the event status register in decimal form (0...255).
ViStatus _VI_FUNC rssgs_EventStatusQuery (ViSession instrSession,
                                          ViInt32 *registerValue)
{
    return rssgs_getViInt32 (instrSession, 
                             "*ESR?", 
                             registerValue, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN IDENTIFICATION QUERY queries the instrument identification. The instrument type, serial number and firmware version are returned.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR instrumentIdentification/Returns the instrument identification string.
ViStatus _VI_FUNC rssgs_IdentificationQuery (ViSession instrSession,
                                             ViChar instrumentIdentification[])
{
    return rssgs_getViString (instrSession, 
                              "*IDN?", 
                              instrumentIdentification, 
                              256, 
                              VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN INDIVIDUAL STATUS QUERY returns the content of the IST flag in decimal form (0 | 1).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR ISTFlag/Returns the content of the IST flag in decimal form (0 | 1).
ViStatus _VI_FUNC rssgs_IndividualStatusQuery (ViSession instrSession,
                                               ViInt32 *ISTFlag)
{
    return rssgs_getViInt32 (instrSession, 
                             "*IST?", 
                             ISTFlag, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN OPERATION COMPLETE sets bit 0 in the event status register when all previous commands have been processed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_OperationComplete (ViSession instrSession)
{
    ViStatus            rssgs_status;
    rssgs_instrRange    instrPtr = NULL;
    ViBoolean           err_check;
    unsigned int        ptrHandle = 0xffffffff;

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));
    
    if (instrPtr == NULL)
        return VI_ERROR_PARAMETER1;

    err_check = instrPtr -> errorChecking;
    
    CHECKERR (rssgs_errorCheckState (instrSession, VI_FALSE));

    rssgs_status = rssgs_sendCmd (instrSession, "*OPC");
    (void) rssgs_errorCheckState (instrSession, err_check);
    
    return rssgs_status;
}

//=====================================================================================
/// HIFN OPERATION COMPLETE QUERY returns a 1 when all previous commands have been processed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR OPCState/Returns a 1 when all previous commands have been processed.
ViStatus _VI_FUNC rssgs_OperationCompleteQuery (ViSession instrSession,
                                                ViInt32 *OPCState)
{
    return rssgs_getViInt32 (instrSession, 
                             "*OPC?", 
                             OPCState, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN OPTION IDENTIFICATION QUERY queries the options configured in the instrument and returns a list of the installed options.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR optionIdentification/Returns a list of the installed options.
ViStatus _VI_FUNC rssgs_OptionIdentificationQuery (ViSession instrSession,
                                                   ViChar optionIdentification[])
{
    return rssgs_getViString (instrSession, 
                              "*OPT?",
                              optionIdentification, 
                              1024, 
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN PARALLEL POLL REGISTER ENABLE sets the parallel poll enable register to the specified value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Sets the parallel poll enable register to the specified value.
ViStatus _VI_FUNC rssgs_ParallelPollRegisterEnable (ViSession instrSession,
                                                    ViInt32 registerValue)
{
    return rssgs_setViInt32 (instrSession, 
                             "*PRE", 
                             registerValue, 0, 655, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN PARALLEL POLL REGISTER QUERY returns the content of the parallel poll enable register in decimal form.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Returns the content of the parallel poll enable register in decimal form.
ViStatus _VI_FUNC rssgs_ParallelPollRegisterQuery (ViSession instrSession,
                                                   ViInt32 *registerValue)
{
    return rssgs_getViInt32 (instrSession, 
                             "*PRE?", 
                             registerValue, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN POWER ON STATUS CLEAR determines whether the content of the ENABle registers are retained or reset at power on.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Determines whether the content of the ENABle registers are retained or reset at power on.
ViStatus _VI_FUNC rssgs_PowerOnStatusClear (ViSession instrSession,
                                            ViInt32 registerValue)
{
    return rssgs_setViInt32 (instrSession, 
                             "*PSC", 
                             registerValue, 0, 1, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN POWER ON STATUS CLEAR QUERY retrieves the contents of the Power-on-Status-Clear flag.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Returns the contents of the Power-on-Status-Clear flag. The answer is either 0 or 1.
ViStatus _VI_FUNC rssgs_PowerOnStatusClearQuery (ViSession instrSession,
                                                 ViInt32 *registerValue)
{
    return rssgs_getViInt32 (instrSession, 
                             "*PSC?", 
                             registerValue,
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN RECALL calls up the instrument status which was stored under the specified number using the *SAV command or using the MMEMory:STORe:STATe command. "0" corresponds to the power-on state, unless a new instrument setting was stored under this number in the meantime using the MMEMory:STORe:STATe command.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR instrumentStatus/Number under which is stored instrument setting.
ViStatus _VI_FUNC rssgs_Recall (ViSession instrSession,
                                ViInt32 instrumentStatus)
{
    return rssgs_setViInt32 (instrSession, 
                             "*RCL", 
                             instrumentStatus, 0, 10, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN SAVE stores the current device state under the specified number (see also *RCL).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR instrumentStatus/Number under which will be stored instrument setting.
ViStatus _VI_FUNC rssgs_Save (ViSession instrSession,
                              ViInt32 instrumentStatus)
{
    return rssgs_setViInt32 (instrSession, 
                             "*SAV", 
                             instrumentStatus, 0, 10, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN SERVICE REQUEST ENABLE sets the service request enable register to the specified value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Ssets the service request enable register to the specified value.
ViStatus _VI_FUNC rssgs_ServiceRequestEnable (ViSession instrSession,
                                              ViInt32 registerValue)
{
    return rssgs_setViInt32 (instrSession, 
                             "*SRE", 
                             registerValue, 0, 255, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN SERVICE REQUEST ENABLE reads out the content of the service request enable register in decimal form.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerValue/Returns the content of the service request enable register in decimal form.
ViStatus _VI_FUNC rssgs_ServiceRequestEnableQuery (ViSession instrSession,
                                                   ViInt32 *registerValue)
{
    return rssgs_getViInt32 (instrSession, 
                             "*SRE?", 
                             registerValue, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN READ STATUS BYTE QUERY reads out the content of the status byte in decimal form.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR statusByte/Returns the content of the status byte in decimal form.
ViStatus _VI_FUNC rssgs_StatusByteQuery (ViSession instrSession,
                                         ViInt32 *statusByte)
{
    return rssgs_getViInt32 (instrSession, 
                             "*STB?", 
                             statusByte, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN TRIGGER triggers all actions which are waiting for a trigger event.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_Trigger (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, 
                          "*TRG");
}

//=====================================================================================
/// HIFN This function triggers all actions waiting for a trigger event in the currently active window and waits for operation completed (OPC) before returning the status code.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR timeout/Sets the timeout for the triggering routine to be finished.
ViStatus _VI_FUNC rssgs_TriggerWaitOPC (ViSession instrSession,
                                        ViInt32 timeout)
{
    ViStatus rssgs_status = VI_SUCCESS;
    ViChar buffer[BUFFER_SIZE] = "";

    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, timeout, 0, 600000), 2);

    CHECKERR (rssgs_waitOPC (instrSession, "*TRG;*OPC?", buffer, timeout));

    return rssgs_status;
}

//=====================================================================================
/// HIFN SELF TEST QUERY triggers all self tests of the instrument and outputs an error code in decimal form (see Service Manual supplied with the instrument).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR errorCode/Returns an error code of the self tests.
ViStatus _VI_FUNC rssgs_SelfTestQuery (ViSession instrSession,
                                       ViInt32 *errorCode)
{
    return rssgs_getViInt32 (instrSession, 
                             "*TST?", 
                             errorCode, 
                             VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN WAIT-to-CONTINUE does not allow subsequent commands to be processed until all previous commands have been executed and all signals are in their transient condition.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_WaittoContinue (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, 
                          "*WAI");
}



/// HIFN This function queries the test points available in the instrument. A detailed description of the test points can be found in Chapter 3 of the Service Manual (supplied with the instrument).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR testPoints/Returns the test points available in the instrument (the various entries are separated by commas).
ViStatus _VI_FUNC rssgs_DIAGnosticTestPoints (ViSession instrSession, 
                                              ViInt32 arraySize, 
                                              ViChar _VI_FAR testPoints[])
{
    ViChar  buffer [BUFFER_SIZE];

    sprintf (buffer, "DIAG:POIN:CAT?");

    return rssgs_getCatalog (instrSession,
                             buffer,
                             (ViUInt32) arraySize,
                             testPoints,
                             VI_ERROR_PARAMETER3);
}

/// HIFN This function triggers voltage measurement at the specified test point and returns the measured voltage. A detailed description of the test points can be found in Chapter 3 of the Service Manual (supplied with the instrument). A list of the available test points can be queried using the DIAG:POIN:CAT? command .
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR testPointName/Test point name.
/// HIPAR measuredVoltage/Returns the voltage measured at the specified test point.
ViStatus _VI_FUNC rssgs_DIAGnosticTestPointVoltage (ViSession instrSession,
                                                     ViString testPointName,
                                                     ViReal64 *measuredVoltage)
{
    ViChar      buffer [BUFFER_SIZE];

    CHECKNULL (testPointName, 2);

    sprintf (buffer, "DIAG:POIN? '%s'", testPointName);

    return rssgs_getViReal64 (instrSession,
                              buffer,
                              measuredVoltage,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function returns the contents of a register on a specific device. For service purposes only.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR bar/identifies the device
/// HIPAR address/the register's address   
/// HIPAR data/the returned register contents

ViStatus _VI_FUNC rssgs_DIAGnosticGetMemoryContent (ViSession instrSession,
                                                    ViInt32 bar,
                                                    ViInt32 address,
                                                    ViPInt32 data)
{
  ViStatus rssgs_status = VI_SUCCESS;
  *data = 0;
  
  CHECKERR (rsPrintf (instrSession,":TEST:DIR? \"%s\"," _PERCD "\n", mmacBarArr[bar],address));
  CHECKERR (rsScanf (instrSession, "#H%x", data));
  
  CHECKERR( rssgs_checkStatus (instrSession));

  
  return rssgs_status;

}
//=====================================================================================
/// HIFN This function sets the contents of a register on a specific device. For service purposes only.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR bar/identifies the device
/// HIPAR address/the register's address   
/// HIPAR data/the register contents
ViStatus _VI_FUNC rssgs_DIAGnosticSetMemoryContent (ViSession instrSession,
                                                    ViInt32 bar,
                                                    ViInt32 address,
                                                    ViInt32 data)
{
  ViStatus rssgs_status = VI_SUCCESS;

  CHECKERR (rsPrintf (instrSession,":TEST:DIR \"%s\"," _PERCD ",#H%08x\n", mmacBarArr[bar],address, data));
  CHECKERR( rssgs_checkStatus (instrSession));

  return rssgs_status;
}                                                    

//=====================================================================================
/// HIFN This function retrieves information about the device's memory usage
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR TotalPhysMemKb/Total system memory 
/// HIPAR ApplicMemKb/Application memory
/// HIPAR HeapUsedKb/Used heap memory
/// HIPAR HeapAvailableKb/Available heap memory

ViStatus _VI_FUNC rssgs_MemoryHeapFree (ViSession instrSession,
                                        ViInt32* TotalPhysMemKb,
                                        ViInt32* ApplicMemKb,
                                        ViInt32* HeapUsedKb,
                                        ViInt32* HeapAvailableKb)
{
  ViStatus rssgs_status = VI_SUCCESS;

  CHECKERR (rsPrintf (instrSession,":MEM:HFR?\n"));
  CHECKERR (rsScanf (instrSession, "" _PERCD "," _PERCD "," _PERCD "," _PERCD "",TotalPhysMemKb,ApplicMemKb,HeapUsedKb,HeapAvailableKb));
  
  CHECKERR( rssgs_checkStatus (instrSession));   
                                                 

  return rssgs_status;

}


/// HIFN This function changes the default drive. This drive is used for all subsequent MMEM commands if no path is specified with them.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR drive/Drive name parameter.
ViStatus _VI_FUNC rssgs_MMEMoryChangeDefaultDrive (ViSession instrSession,
                                                   ViString drive)
{
    return rssgs_setViStringQuoted (instrSession,
                              "MMEM:MSIS",
                              "'",
                              drive,
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the default drive. This drive is used for all subsequent MMEM commands if no path is specified with them.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR drive/Default Drive name.
/// HIPAR arraySize/Size of drive array.
ViStatus _VI_FUNC rssgs_MMEMoryGetDefaultDrive (ViSession instrSession,
                                                ViInt32 arraySize,
                                                ViChar drive[])
{
    return rssgs_getViStringQuoted (instrSession,
                              "MMEM:MSIS?",
                              "\"",
                              arraySize,
                              drive,
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function changes the default directory. This directory is used for all subsequent MMEM commands if no path is specified with them. It is also possible to change to a higher directory using two dots '..'.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryName/Directory name parameter.
ViStatus _VI_FUNC rssgs_MMEMoryChangeDirectory (ViSession instrSession,
                                       ViString directoryName)
{
    return rssgs_setViStringQuoted (instrSession,
                            "MMEM:CDIR",
                            "'",
                            directoryName,
                            VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the current working directory.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/Defines size of array in argument Directory..
/// HIPAR directory/Directory name parameter.
ViStatus _VI_FUNC rssgs_MMEMoryGetDirectory (ViSession instrSession,
                                             ViInt32 arraySize,
                                             ViChar directory[])
{
    return rssgs_getViStringQuoted (instrSession,
                            "MMEM:CDIR?",
                            "\"",
                            arraySize,
                            directory,
                            VI_ERROR_PARAMETER2);
}

/// HIFN This function copies the first specified file to the second specified file. Instead of just a file, this function can also be used to copy a complete directory together with all its files.
/// HIFN If <destination> is not specified, <source> is copied to the MMEM:MSIS drive and the MMEM:CDIR directory. Files which already exist with the same name in the destination directory are overwritten without an error message.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR sourcePath/Source path of the file or a complete directory together with all its files.
/// HIPAR destinationPath/Destination path of the file or a directory. If 'Destination Path' is not specified, 'Source Path' is copied to the MMEM:MSIS drive and the MMEM:CDIR directory.
ViStatus _VI_FUNC rssgs_MMEMoryCopy (ViSession instrSession, ViString sourcePath,
                            ViString destinationPath)
{
    ViStatus    rssgs_status;
    ViChar      buffer[BUFFER_SIZE];

    CHECKNULL (sourcePath, 2);

    if (destinationPath)
        sprintf (buffer, "MMEM:COPY '%s','%s'", sourcePath, destinationPath);
    else
        sprintf (buffer, "MMEM:COPY '%s'", sourcePath);

    rssgs_status = rssgs_sendCmd (instrSession, buffer);

    return rssgs_status;
}

/// HIFN This function writes the block data <binary block data> to the file identified by <file_name>. The IEC/IEEE-bus terminator should be set to EOI in order to ensure correct data transfer.
/// HIFN This function can be used to transfer stored instrument settings or waveforms directly to the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR fileName/File name of the binary block data.
/// HIPAR binaryBlockLength/Defines the number of bytes in the 'Binary Block Data' array.
/// HIPAR binaryBlockData/Binary block data array to be writen to the file defined by  'File Name' parameter with the size limited with 'Binary Block Length' parameter.
ViStatus _VI_FUNC rssgs_MMEMoryBlockDataWrite (ViSession instrSession,
                                      ViString fileName,
                                      ViInt32 binaryBlockLength,
                                      ViChar _VI_FAR binaryBlockData[])
{
    ViStatus    rssgs_status    = VI_SUCCESS;
    ViUInt16    intfType        = VI_INTF_GPIB;
    ViChar*     pbuffer;
    ViChar*     p2buf;

    CHECKNULL (fileName, 2);
    CHECKNULL (binaryBlockData, 4);

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (rsWrite (instrSession, (ViBuf) "SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    p2buf = pbuffer = (ViChar *)(malloc (strlen (fileName) + binaryBlockLength + 1000));

    p2buf += sprintf (pbuffer, "MMEM:DATA '%s',#" _PERCD "" _PERCD "",
                               fileName,
                               (ViInt32)(floor(log10((double)binaryBlockLength)) + 1),
                               binaryBlockLength);

    memcpy (p2buf, binaryBlockData, (size_t) binaryBlockLength);
    p2buf[binaryBlockLength] = '\n';

    rssgs_status = rsWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32) ((p2buf - pbuffer) + binaryBlockLength+1), VI_NULL);

    if (pbuffer)
        free (pbuffer);

    if (intfType == VI_INTF_GPIB)
        CHECKERR (rsWrite (instrSession, (ViBuf) "SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL));

    return rssgs_status;
}

/// HIFN This function transfers the specified file from the R&S SGS to the IEC/IEEE bus and then on to the control computer. It is important to ensure that the intermediate memory on the control computer is large enough to take the file. In this case, the setting for the IEC/IEEE-bus terminator is irrelevant.
/// HIFN This function can be used to read stored instrument settings or waveforms directly from the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR fileName/File name of the binary block data.
/// HIPAR binaryBlockLength/Length of the binary block data array.
/// HIPAR binaryBlockData/Binary block data to be retrieved from the file defined by the 'File Name' parameter. Array size of this parameter must be large enough to store all the retrieved data.
ViStatus _VI_FUNC rssgs_MMEMoryBlockDataRead (ViSession instrSession,
                                              ViString fileName,
                                              ViInt32 binaryBlockLength,
                                              ViChar _VI_FAR binaryBlockData[])
{
    ViStatus    rssgs_status    = VI_SUCCESS;
    ViUInt16    intfType        = VI_INTF_GPIB;
    ViUInt32    count           = 0,
                lengthOfLength  = 0;
    ViChar      buffer[BUFFER_SIZE],
                format[32],
                *pbuffer = NULL, *p2buffer = NULL;

    CHECKNULL (fileName, 2);

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (rsWrite (instrSession, (ViBuf) "SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    sprintf (buffer, "MMEM:DATA? '%s'\n", fileName);

    CHECKERR (rsWrite (instrSession, (ViBuf) buffer, (ViUInt32) strlen(buffer), VI_NULL));
    CHECKERR (rssgs_readDataUnknownLength (instrSession, &pbuffer, &count));

    p2buffer = pbuffer;

    if (sscanf (p2buffer, "#" _PERC1U "", &lengthOfLength) != 1)
        {
        free (pbuffer);
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    sprintf (format, "%%" _PERCUD "", lengthOfLength);
    p2buffer += 2;

    if (sscanf (p2buffer, format, &count) != 1)
        {
        free (pbuffer);
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    p2buffer += lengthOfLength;

    if (intfType == VI_INTF_GPIB)
        if ((rssgs_status = rsWrite (instrSession, (ViBuf) "SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL)) < VI_SUCCESS)
            {
            free (pbuffer);
            return rssgs_status;
            }

    if ((rssgs_status = rssgs_checkStatus (instrSession)) < VI_SUCCESS)
        {
        free (pbuffer);
        return rssgs_status;
        }

    if (count > (ViUInt32) binaryBlockLength)
        {
        rssgs_status = RSSGS_WARNING_SMALL_BUFFER;
        count = (ViUInt32) binaryBlockLength;
        memcpy (binaryBlockData, p2buffer, (size_t) count);
        }
    else
        {
        rssgs_status = VI_SUCCESS;
        memcpy (binaryBlockData, p2buffer, (size_t) count);
        }

    free (pbuffer);

    return rssgs_status;
}


/// HIFN This function reads out the subdirectories and files of the specified directory. If no directory is specified, the default directory selected with the MMEM:CDIR command is read out. The directories are output in a list (the list entries are separated by commas).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR arraySize/Array size of the parameter labeled 'Catalog List'.
/// HIPAR catalogList/Reads back catalog list. Size of the list is limited by the 'Array Size' parameter.
ViStatus _VI_FUNC rssgs_MMEMoryCatalog (ViSession instrSession,
                                        ViString directoryPath,
                                        ViInt32 arraySize, ViChar catalogList[])
{
    ViStatus    rssgs_status;
    ViChar      buffer[BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, "MMEM:CAT?");
    else
        sprintf (buffer, "MMEM:CAT? '%s'", directoryPath);

    rssgs_status = rssgs_getCatalog (instrSession,
                                     buffer,
                                     (ViUInt32) arraySize,
                                     catalogList,
                                     VI_ERROR_PARAMETER4);

    return rssgs_status;
}

/// HIFN This function reads out the number of entries in specified directory. If no directory is specified, the default directory selected with the MMEM:CDIR command is read out. The directories are output in a list (the list entries are separated by commas).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR length/Reads back number of entries in the catalog.
ViStatus _VI_FUNC rssgs_MMEMoryCatalogLength (ViSession instrSession,
                                        ViString directoryPath,
                                        ViInt32* length)
{
  ViStatus    rssgs_status;
  ViChar      buffer[BUFFER_SIZE];

  if (!directoryPath)
    sprintf (buffer, "MMEM:CAT:LENG?");
  else
    sprintf (buffer, "MMEM:CAT:LENG? '%s'", directoryPath);

  rssgs_status = rssgs_getViInt32(instrSession,
    buffer,
    length,
    VI_ERROR_PARAMETER3);

  return rssgs_status;
}

/// HIFN This function reads out the subdirectories of the specified directory. If no directory is specified, the default directory selected with the MMEM:CDIR command is read out. The directories are output in a list (the list entries are separated by commas).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR arraySize/Array size of the parameter labeled 'Catalog List'.
/// HIPAR catalogList/Reads back catalog list. Size of the list is limited by the 'Array Size' parameter.
ViStatus _VI_FUNC rssgs_MMEMoryDirectoryCatalog (ViSession instrSession,
                                        ViString directoryPath,
                                        ViInt32 arraySize, ViChar _VI_FAR catalogList[])
{
    ViStatus    rssgs_status;
    ViChar      buffer[BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, "MMEM:DCAT?");
    else
        sprintf (buffer, "MMEM:DCAT? '%s'", directoryPath);

    rssgs_status = rssgs_getCatalog (instrSession,
                                     buffer,
                                     (ViUInt32) arraySize,
                                     catalogList,
                                     VI_ERROR_PARAMETER4);

    return rssgs_status;
}

/// HIFN This function reads out the number of directories in specified directory. If no directory is specified, the default directory selected with the MMEM:CDIR command is read out. The directories are output in a list (the list entries are separated by commas).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryPath/Directory path parameter.
/// HIPAR length/Reads back number of directories in the catalog.
ViStatus _VI_FUNC rssgs_MMEMoryDirectoryCatalogLength (ViSession instrSession,
                                              ViString directoryPath,
                                              ViInt32* length)
{
  ViStatus    rssgs_status;
  ViChar      buffer[BUFFER_SIZE];

  if (!directoryPath)
    sprintf (buffer, "MMEM:DCAT:LENG?");
  else
    sprintf (buffer, "MMEM:DCAT:LENG? '%s'", directoryPath);

  rssgs_status = rssgs_getViInt32(instrSession,
    buffer,
    length,
    VI_ERROR_PARAMETER3);

  return rssgs_status;
}

/// HIFN This function deletes the specified file.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR fileName/File name parameter.
ViStatus _VI_FUNC rssgs_MMEMoryDelete (ViSession instrSession, ViString fileName)
{
    return rssgs_setViStringQuoted (instrSession,
                              "MMEM:DEL",
                              "'",
                              fileName,
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function creates a new subdirectory in the specified  directory. If no directory is specified, a subdirectory is created in the default directory. This function can also be used to create a directory tree.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryName/Directory name parameter.
ViStatus _VI_FUNC rssgs_MMEMoryMakeDirectory (ViSession instrSession,
                                     ViString directoryName)
{
    return rssgs_setViStringQuoted (instrSession,
                              "MMEM:MDIR",
                              "'",
                              directoryName,
                              VI_ERROR_PARAMETER2);
}


/// HIFN This function renames an existing file ('File Source') if no path is specified for 'File Destination'. Otherwise the file is moved to the specified path and stored under the original file name or, if specified, a new file name. 
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR fileSource/Define file source.
/// HIPAR fileDestination/Define file destination.
ViStatus _VI_FUNC rssgs_MMEMoryMove (ViSession instrSession, ViString fileSource,
                            ViString fileDestination)
{
    ViStatus    rssgs_status;
    ViChar*     pbuffer;

    CHECKNULL (fileSource, 2);
    CHECKNULL (fileDestination, 3);

    pbuffer = (ViChar *)(malloc (strlen (fileSource) + strlen (fileDestination) + 20));

    sprintf (pbuffer, "MMEM:MOVE '%s','%s'", fileSource, fileDestination);

    rssgs_status = rssgs_sendCmd (instrSession, pbuffer);

    free (pbuffer);

    RETURN;
}

/// HIFN This function deletes the specified subdirectory in the specified directory. If no directory is specified, the subdirectory with the specified name is deleted in the default directory.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR directoryName/Directory name parameter.
ViStatus _VI_FUNC rssgs_MMEMoryRemoveDirectory (ViSession instrSession,
                                       ViString directoryName)
{
    return rssgs_setViStringQuoted (instrSession,
                              "MMEM:RDIR",
                              "'",
                              directoryName,
                              VI_ERROR_PARAMETER2);
}

/// HIFN This function loads the specified file stored under the specified name into internal memory
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR storage/number of the internal memory slot
/// HIPAR fileName/file name from which to load the settings
ViStatus _VI_FUNC rssgs_MMEMoryLoadState (ViSession instrSession,
                                          ViInt32 storage,  
                                          ViString fileName)
{
    ViChar buffer[BUFFER_SIZE];
    sprintf (buffer, "MMEM:LOAD:STATE " _PERCD ", \"%s\"\n", storage, fileName);
    return rsWrite (instrSession, (ViBuf) buffer, (ViUInt32) strlen(buffer), VI_NULL);
}

/// HIFN This function stores the current instrument setting from internal memory in the specified file.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR storage/number of the internal memory slot
/// HIPAR fileName/file name where to store the settings
ViStatus _VI_FUNC rssgs_MMEMoryStoreState (ViSession instrSession,
                                          ViInt32 storage,  
                                          ViString fileName)
{
    ViChar buffer[BUFFER_SIZE];
    sprintf (buffer, "MMEM:STORE:STATE " _PERCD ", \"%s\"\n", storage, fileName);
    return rsWrite (instrSession, (ViBuf) buffer, (ViUInt32) strlen(buffer), VI_NULL);
}


//=====================================================================================
/// HIFN The respective hardware assembly responds directly to this function; any safety mechanisms are bypassed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR component/Selects the hardware component
/// HIPAR address/Selects the address to read from
/// HIPAR data/Response of the respective hardware assembly.
ViStatus _VI_FUNC rssgs_TESTDirect (ViSession instrSession, 
                                    ViInt32 component,
                                    ViInt32 address,
                                    ViPInt32 data)

{
    return  rssgs_DIAGnosticGetMemoryContent (instrSession,
                                              component,
                                              address,
                                              data);

}

/// HIFN This function perform a test of the internal voltages. Returns 0 on success.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR result/Returns result of the test. The response "0" indicates error-free testing, and the response "1" means that an error occurred during test.
ViStatus _VI_FUNC rssgs_TESTVoltage (ViSession instrSession, ViInt32 *result)
{
    return rssgs_getViInt32 (instrSession,
                             "TEST:VOLT?",
                             result,
                             VI_ERROR_PARAMETER2);
}

/// HIFN This function gets the result or current state of a selftest. Returns 0 on success.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR result/Returns result of the test. The response "0" indicates error-free testing, and the response "1" means that an error occurred during test.
ViStatus _VI_FUNC rssgs_TESTSelftestResult (ViSession instrSession, ViInt32 *result)
{
  return rssgs_getEnum (instrSession, 
                        "TEST:ALL:RES?",
                        result, 
                        testResult, 
                        VI_ERROR_PARAMETER2);
}

/// HIFN This function perform selftest. Returns 0 on success.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_TESTSelftestStart (ViSession instrSession)
{
  return rssgs_sendCmd (instrSession, "TEST:ALL:STARt");
}

/// HIFN Enable/disable keyboard and LED test state..
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Enable/disable keyboard and LED test state.
ViStatus _VI_FUNC rssgs_SetTestKeyboardState (ViSession instrSession,
                                             ViBoolean state)
{
	return rssgs_setViBoolean (instrSession, 
                          "TEST:KEYB", 
                          state,
                          VI_ERROR_PARAMETER2);    
}

/// HIFN Returns keyboard and LED test state.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/Returns keyboard and LED test state.
ViStatus _VI_FUNC rssgs_GetTestKeyboardState (ViSession instrSession,
                                             ViBoolean *state)
{
	return rssgs_getViBoolean (instrSession, 
                          "TEST:KEYB?", 
                          state,
                          VI_ERROR_PARAMETER2);    
}

/// HIFN  Test all signal connections. 
/// HIFN  Equivalent to button "Test Signal Connections" in dialog "Extension".
/// HIFN  Reports errors to the SCPI error queue. (Use function rssgs_GetTestRFSignalConnection
/// HIFN  for a quick and silent test.)
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR testSignalConnections/Displays test all signal connections.
ViStatus _VI_FUNC rssgs_GetTestAllSignalConnections(
	ViSession	instrSession,
	ViInt32	*testSignalConnections)
{
    
    return rssgs_getViInt32 (instrSession,
                             ":TEST:EXT:CONN?",
                             testSignalConnections,
                             VI_ERROR_PARAMETER2);   
}


/// HIFN  Returns test RF signal connection only.
/// HIFN  Does not report errors to the SCPI error queue.
/// HIFN  Use this function to detect which extension is physically connected to a device.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR testRFSignalConnection/Displays sest RF signal connection only.
ViStatus _VI_FUNC rssgs_GetTestRFSignalConnection(
	ViSession	instrSession,
	ViInt32	*testRFSignalConnection)
{
    return rssgs_getViInt32 (instrSession,
                             ":TEST:EXT:CONN:RF?",
                             testRFSignalConnection,
                             VI_ERROR_PARAMETER2);   
}

/*============================================================================
 * Function: rssgs_confPulm
 * Purpose:  This function sets the parameters of pulse modulation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_confPulm (ViSession instrSession, ViBoolean state,
                                  ViInt32 source, ViInt32 polarity,
                                  ViBoolean pulseGeneratorState,
                                  ViReal64 pulsePeriod, ViReal64 pulseWidth,
                                  ViReal64 pulseDelay,
                                  ViInt32 pulseMode,
                                  ViReal64 doublePulseWidth,
                                  ViReal64 doublePulseDelay, ViInt32 triggerMode)
{

    INIT;
	  pulseGeneratorState = 0;
    if (instrPtr->instrOptions.K22 != 0)
    {
        rssgs_status = rssgs_SetPULMSource (instrSession, source);
        REMAP (2, 3);

        rssgs_status = rssgs_SetPULMPolarity (instrSession, polarity);
        REMAP (2, 4);

        rssgs_status = rssgs_SetPULMPeriod (instrSession, pulsePeriod);
        REMAP (2, 6);

        rssgs_status = rssgs_SetPULMWidth (instrSession, pulseWidth);
        REMAP (2, 7);


        rssgs_status = rssgs_SetPULMDelay (instrSession, pulseDelay);
        REMAP (2, 8);

        rssgs_status = rssgs_SetPULMTriggerMode (instrSession, triggerMode);
        REMAPINSTR (2,12);

        rssgs_status = rssgs_SetPULMMode   (instrSession, pulseMode);
        REMAPINSTR (2,9);
        
        rssgs_status = rssgs_SetPULMDoubleDelay (instrSession, doublePulseDelay);
        REMAPINSTR (2,11);

        rssgs_status = rssgs_SetPULMDoubleWidth (instrSession, doublePulseWidth);
        REMAPINSTR (2,10);
    }
            
    return rssgs_SetPULMState (instrSession, state);
}



/*============================================================================
 * Function: rssgs_SetPULMState
 * Purpose:  This function activates/deactivates the pulse modulation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMState (ViSession instrSession, 
                                      ViBoolean state)
{

    return rssgs_setViBoolean (instrSession,
                               "SOUR:PULM:STAT",
                               state,
                               VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMState
 * Purpose:  This function returns the pulse modulation state.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMState (ViSession instrSession, 
                                      ViBoolean *state)
{

    return rssgs_getViBoolean (instrSession,
                               "SOUR:PULM:STAT?",
                               state,
                               VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMPolarity
 * Purpose:  This function sets the polarity between modulating and modulated
 *           signal. This function is only effective for an external
 *           modulation signal.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMPolarity (ViSession instrSession,
                                         ViInt32   polarity)
{
    return rssgs_setEnum (instrSession,
                          "PULM:POL",
                          polarity,
                          polarityArr,
                          VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMPolarity
 * Purpose:  This function returns the polarity between modulating and
 *           modulated signal. This function is only effective for an
 *           external modulation signal.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMPolarity (ViSession instrSession,
                                         ViInt32   *polarity)
{
    return rssgs_getEnum (instrSession,
                          "PULM:POL?",
                          polarity,
                          polarityArr,
                          VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMSource
 * Purpose:  This function selects the source for pulse modulation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMSource (ViSession instrSession,
                                       ViInt32   pulmSource)
{
    return rssgs_setEnum (instrSession,
                          "PULM:SOUR",
                          pulmSource,
                          pulmSourceArr,
                          VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMSource
 * Purpose:  This function returns the source for pulse modulation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMSource (ViSession instrSession,
                                       ViInt32   *pulmSource)
{
    return rssgs_getEnum (instrSession,
                          "PULM:SOUR?",
                          pulmSource,
                          pulmSourceArr,
                          VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMPeriod
 * Purpose:  This function sets the period of the generated pulse. The period
 *           determines the repetition
 *           frequency of the internal signal.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMPeriod (ViSession instrSession,
                                       ViReal64  period)
{
    return rssgs_setViReal64 (instrSession,
                              "PULM:PER",
                              period,
                              0.0,0.0,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMPeriod
 * Purpose:  This function returns the period of the generated pulse.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMPeriod (ViSession instrSession,
                                       ViReal64 *period)
{
    return rssgs_getViReal64 (instrSession,
                              "PULM:PER?",
                              period,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMWidth
 * Purpose:  This function sets the width of the generated pulse. The width
 *           determines the pulse length. The pulse width must be at least 20
 *           ns less than the set pulse period.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMWidth (ViSession instrSession,
                                      ViReal64  width)
{
    return rssgs_setViReal64 (instrSession,
                              "PULM:WIDT",
                              width,
                              20.0e-9,1.0,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMWidth
 * Purpose:  This function returns the width of the generated pulse.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMWidth (ViSession instrSession,
                                      ViReal64  *period)
{
    INIT;
    CHECKOPT (K22);
    
    return rssgs_getViReal64 (instrSession,
                              "PULM:WIDT?",
                              period,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMDelay
 * Purpose:  This function sets the pulse delay. The pulse delay determines
 *           the time that elapses after a trigger event before pulse
 *           modulation starts. The pulse delay is not effective for double
 *           pulse generation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMDelay (ViSession instrSession,
                                      ViReal64  delay)
{
    INIT;
    CHECKOPT (K22);
    
    return rssgs_setViReal64 (instrSession,
                              "PULM:DEL",
                              delay,
                              10.0e-9,100.0,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMDelay
 * Purpose:  This function returns the pulse delay.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMDelay (ViSession instrSession,
                                      ViReal64  *delay)
{
    INIT;
    CHECKOPT (K22);
    
    return rssgs_getViReal64 (instrSession,
                              "PULM:DEL?",
                              delay,
                              VI_ERROR_PARAMETER2);
}


/*============================================================================
 * Function: rssgs_SetPULMDoubleWidth
 * Purpose:  This function sets the width of the second pulse in case of
 *           double pulse generation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMDoubleWidth (ViSession instrSession,
                                            ViReal64  width)
{
    INIT;
    CHECKOPT (K22);
    
    return rssgs_setViReal64 (instrSession,
                              "PULM:DOUBLE:WIDT",
                              width,
                              20.0e-9,1.0,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMDoubleWidth
 * Purpose:  This function returns the width of the second pulse in case of
 *           double pulse generation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMDoubleWidth (ViSession instrSession,
                                            ViReal64  *width)
{
    INIT;
    CHECKOPT (K22);
    
    return rssgs_getViReal64 (instrSession,
                              "PULM:DOUBLE:WIDT?",
                              width,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMDoubleDelay
 * Purpose:  This function sets the delay from the start of the first pulse
 *           to the start of the second pulse.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMDoubleDelay (ViSession instrSession,
                                            ViReal64  delay)
{
    INIT;
    CHECKOPT (K22);
    return rssgs_setViReal64 (instrSession,
                              "PULM:DOUBLE:DEL",
                              delay,
                              10.0e-9,100.0,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMDoubleDelay
 * Purpose:  This function returns the delay from the start of the first
 *           pulse to the start of the second pulse.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMDoubleDelay (ViSession instrSession,
                                            ViReal64  *delay)
{
    INIT;
    CHECKOPT (K22);
    return rssgs_getViReal64 (instrSession,
                              "PULM:DOUBLE:DEL?",
                              delay,
                              VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_SetPULMTriggerMode
 * Purpose:  This function selects the trigger mode for pulse modulation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMTriggerMode (ViSession instrSession,
                                            ViInt32   pulmTriggerMode)
{
    INIT;
    CHECKOPT (K22);
    return rssgs_setEnum (instrSession,
                          "PULM:TRIG:MODE",
                          pulmTriggerMode,
                          pulmTriggerModeArr,
                          VI_ERROR_PARAMETER2);
}

/*============================================================================
 * Function: rssgs_GetPULMTriggerMode
 * Purpose:  This function returns the trigger mode for pulse modulation.
 *===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMTriggerMode (ViSession instrSession,
                                            ViInt32   *pulmTriggerMode)
{
    INIT;
    CHECKOPT (K22);
    return rssgs_getEnum (instrSession,
                          "PULM:TRIG:MODE?",
                          pulmTriggerMode,
                          pulmTriggerModeArr,
                          VI_ERROR_PARAMETER2);
}



/*===========================================================================*/
// Function: Set PULM External Trigger Input Impedance
// Purpose:  This function selects the impedance for external pulse trigger.
//           The trigger is fed via the PULSE EXT input.
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMExternalTriggerInputImpedance (ViSession instrSession,
                                                              ViInt32 externalInputImpedance)
{
  INIT;
  CHECKOPT (K22);
  return rssgs_setEnum (instrSession,
    "PULM:TRIG:EXT:IMP",
    externalInputImpedance,
    dmemExternalInputImpedance,
    VI_ERROR_PARAMETER2);
}

/*===========================================================================*/
// Function: Get PULM External Trigger Input Impedance
// Purpose:  This function returns the impedance for external pulse trigger.
//           The trigger is fed via the PULSE EXT input.
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMExternalTriggerInputImpedance (ViSession instrSession,
                                                              ViInt32 *externalInputImpedance)
{
  INIT;
  CHECKOPT (K22);
  return rssgs_getEnum (instrSession,
    "PULM:TRIG:EXT:IMP?",
    externalInputImpedance,
    dmemExternalInputImpedance,
    VI_ERROR_PARAMETER2);
}

/*===========================================================================*/
// Function: Set PULM External Trigger Slope
// Purpose:  This function sets the polarity of the active slope of an
//           applied trigger at the PULSE EXT connector
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMExternalTriggerSlope (ViSession instrSession,
                                                     ViInt32 externalTriggerSlope)
{
  INIT;
  CHECKOPT (K22);
  return rssgs_setEnum (instrSession,
    "PULM:TRIG:EXT:SLOP",
    externalTriggerSlope,
    slopeArr,
    VI_ERROR_PARAMETER2);
}

/*===========================================================================*/
// Function: Get PULM External Trigger Slope
// Purpose:  This function returns the polarity of the active slope of an
//           applied trigger at the PULSE EXT connector
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMExternalTriggerSlope (ViSession instrSession,
                                                     ViInt32 *externalTriggerSlope)
{
  INIT;
  CHECKOPT (K22);
  return rssgs_getEnum (instrSession,
    "PULM:TRIG:EXT:SLOP?",
    externalTriggerSlope,
    slopeArr,
    VI_ERROR_PARAMETER2);
}

/*===========================================================================*/
// Function: Set PULM Gate Input Polarity
// Purpose:  This function selects the polarity of the external gate signal.
//           The signal is supplied via the PULSE EXT connector.
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMGateInputPolarity (ViSession instrSession,
                                                  ViInt32 gateInputPolarity)
{
  INIT;
  CHECKOPT (K22);
  return rssgs_setEnum (instrSession,
    "PULM:TRIG:EXT:GATE:POL",
    gateInputPolarity,
    polarityArr,
    VI_ERROR_PARAMETER2);
}

/*===========================================================================*/
// Function: Get PULM Gate Input Polarity
// Purpose:  This function returns the polarity of the external gate signal.
//           The signal is supplied via the PULSE EXT connector.
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMGateInputPolarity (ViSession instrSession,
                                                  ViInt32 *gateInputPolarity)
{
  INIT;
  CHECKOPT (K22);
  return rssgs_getEnum (instrSession,
    "PULM:TRIG:EXT:GATE:POL?",
    gateInputPolarity,
    polarityArr,
    VI_ERROR_PARAMETER2);
}

/*===========================================================================*/
// Function: Set PULM Mode
// Purpose:  This function selects the mode for pulse generation.
//           
//           Note(s):
//           
//           (1) This function is only available with option R&S SMF-K23, 
//           parameter PTRain is only available with option R&S SMF-K27.
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_SetPULMMode(
                                    ViSession   instrSession,
                                    ViInt32 mode
                                    )
{
  INIT;

  return rssgs_setEnum (instrSession,
    "SOUR:PULM:MODE",
    mode,
    PULMModeArr,
    VI_ERROR_PARAMETER2);    
}

/*===========================================================================*/
// Function: Get PULM Mode
// Purpose:  This function returns the mode for pulse generation.
//           
//           Note(s):
//           
//           (1) This function is only available with option R&S SMF-K23, 
//           parameter PTRain is only available with option R&S SMF-K27.
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_GetPULMMode(
                                    ViSession   instrSession,
                                    ViInt32*    mode
                                    )
{
  INIT;

  return rssgs_getEnum (instrSession,
    "SOUR:PULM:MODE?",
    mode,
    PULMModeArr,
    VI_ERROR_PARAMETER2);    
}



/// HIFN This function sets the value of selected status register.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerGroup/Selects the status register group.
/// HIPAR registerType/Selects the type of status register.
/// HIPAR value/Sets the value to be set to selected register.
ViStatus _VI_FUNC rssgs_SetRegister (ViSession instrSession, ViInt32 registerGroup,
                            ViInt32 registerType, ViInt32 value)
{
    ViChar  buffer[BUFFER_SIZE];

    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, registerGroup, 0, 1), 2);
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, registerType, 2, 4), 3);

    sprintf (buffer, "STAT:%s:%s", regGroupArr[registerGroup], regTypeArr[registerType]);

    return rssgs_setViInt32 (instrSession,
                             buffer,
                             value,
                             0, 0,
                             VI_ERROR_PARAMETER4);
}


/// HIFN This function sets the value of selected status register.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR registerGroup/Selects the status register group.
/// HIPAR registerType/Selects the type of status register.
/// HIPAR value/Returns the value of selected register.
ViStatus _VI_FUNC rssgs_GetRegister (ViSession instrSession, ViInt32 registerGroup,
                            ViInt32 registerType, ViInt32* value)
{
    ViChar  buffer[BUFFER_SIZE];

    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, registerGroup, 0, 1), 2);
    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, registerType, 0, 4), 3);

    sprintf (buffer, "STAT:%s:%s?", regGroupArr[registerGroup], regTypeArr[registerType]);

    return rssgs_getViInt32 (instrSession,
                             buffer,
                             value,
                             VI_ERROR_PARAMETER4);
}

//=====================================================================================
/// HIFN This function resets the status registers.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_STATusPRESet (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, 
                          "STAT:PRES");
}

/// HIFN This function queries all entries in the error queue and then deletes them. Positive error numbers denote instrument-specific errors, and negative error numbers denote error messages defined by SCPI (see Chapter 9). If the error queue is empty, 0 ("No error") is returned.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/Array size of the parameter labeled 'Error Queue'.
/// HIPAR errorQueue/Returns all entries in the error queue. Buffer length is limited by the 'Array Size' parameter.
ViStatus _VI_FUNC rssgs_SYSTemErrorQueueAll (ViSession instrSession,
                                    ViInt32 arraySize, ViChar _VI_FAR errorQueue[])
{
    return rssgs_getViString (instrSession,
                              "SYST:ERR:ALL?",
                              errorQueue,
                              (ViUInt32) arraySize,
                              VI_ERROR_PARAMETER3);
}

/// HIFN This function queries all entries in the error queue and then deletes them. Only the error numbers are returned and not the entire error text. Positive error numbers denote instrument-specific errors, and negative error numbers denote error messages defined by SCPI (see Chapter 9). If the error queue is empty, 0 ("No error") is returned.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/Array size of the parameter labeled 'Error Queue Code'.
/// HIPAR errorQueueCode/Returns all entries in the error queue. Only the error numbers are returned and not the entire error text. Buffer length is limited by the 'Array Size' parameter.
ViStatus _VI_FUNC rssgs_SYSTemErrorQueueCodeAll (ViSession instrSession,
                                        ViInt32 arraySize,
                                        ViChar _VI_FAR errorQueueCode[])
{
    return rssgs_getViString (instrSession,
                              "SYST:ERR:CODE:ALL?",
                              errorQueueCode,
                              (ViUInt32) arraySize,
                              VI_ERROR_PARAMETER3);
}

/// HIFN This function queries the oldest entry in the error queue and then deletes it. Only the error number is returned and not the entire error text. Positive error numbers denote instrument-specific errors, and negative error numbers denote error messages defined by SCPI (see Chapter 9). If the error queue is empty, 0 is returned.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR errorQueueCodeNext/Returns the oldest entry in the error queue. Only the error number is returned and not the entire error text.
ViStatus _VI_FUNC rssgs_SYSTemErrorQueueCodeNext (ViSession instrSession,
                                         ViInt32 *errorQueueCodeNext)
{
    return rssgs_getViInt32 (instrSession,
                             "SYST:ERR:CODE?",
                             errorQueueCodeNext,
                             VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the number of entries in the error queue. If the error queue is empty, '0' is returned.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR errorQueueCount/Returns the number of entries in the error queue.
ViStatus _VI_FUNC rssgs_SYSTemErrorQueueCount (ViSession instrSession,
                                      ViInt32 *errorQueueCount)
{
    return rssgs_getViInt32 (instrSession,
                             "SYST:ERR:COUN?",
                             errorQueueCount,
                             VI_ERROR_PARAMETER2);
}

/// HIFN This function returns a list of all errors existing at the time when the query is started. This list corresponds to the display on the info page under manual control.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/Array size of the parameter labeled 'Error Queue'.
/// HIPAR errorQueue/Returns all entries in the error queue. Buffer length is limited by the 'Array Size' parameter.
ViStatus _VI_FUNC rssgs_SYSTemStaticErrorQueue (ViSession instrSession,
                                                ViInt32 arraySize,
                                                ViChar errorQueue[])
{
    return rssgs_getViString (instrSession,
                              "SYST:SERR?",
                              errorQueue,
                              (ViUInt32) arraySize,
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the USB resource string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR USBResource/This control displays the USB resource string.
ViStatus _VI_FUNC rssgs_SYSTemUSBResource (ViSession instrSession,
                                           ViUInt32 arraySize,
                                           ViChar USBResource[])
{
    return rssgs_getViString (instrSession, 
                              "SYST:COMM:USB:RES?", 
                              USBResource, 
                              arraySize, 
                              VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the Socket resource string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR SocketResource/This control displays the Socket resource string.
ViStatus _VI_FUNC rssgs_SYSTemSocketResource (ViSession instrSession,
                                           ViUInt32 arraySize,
                                           ViChar SocketResource[])
{
  return rssgs_getViString (instrSession, 
    "SYST:COMM:SOCK:RES?", 
    SocketResource, 
    arraySize, 
    VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the Socket resource string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR PCIeResource/This control displays the PCIe resource string.
ViStatus _VI_FUNC rssgs_SYSTemPCIeResource (ViSession instrSession,
                                              ViUInt32 arraySize,
                                              ViChar PCIeResource[])
{
  return rssgs_getViString (instrSession, 
    "SYST:COMM:PCI:RES?", 
    PCIeResource, 
    arraySize, 
    VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the Hislip resource string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR HislipResource/This control displays the Hislip resource string.
ViStatus _VI_FUNC rssgs_SYSTemHislipResource (ViSession instrSession,
                                            ViUInt32 arraySize,
                                            ViChar HislipResource[])
{
  return rssgs_getViString (instrSession, 
    "SYST:COMM:HISL:RES?", 
    HislipResource, 
    arraySize, 
    VI_ERROR_PARAMETER3);
}

//=====================================================================================
/// HIFN This function queries the serial resource string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR arraySize/This control sets the array size.
/// HIPAR SerialResource/This control displays the serial resource string.
ViStatus _VI_FUNC rssgs_SYSTemSerialResource (ViSession instrSession,
                                              ViUInt32 arraySize,
                                              ViChar SerialResource[])
{
  return rssgs_getViString (instrSession, 
    "SYST:COMM:SER:RES?", 
    SerialResource, 
    arraySize, 
    VI_ERROR_PARAMETER3);
}



/// HIFN This function restore the device state to the factory settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_SYSTemRestoreFactorySettings (ViSession instrSession)
{
    return rssgs_sendCmd (instrSession, "SYST:FPR");
}

/// HIFN  This function enables and selects the Eco Mode. With enabled eco mode 
/// HIFN  EM1 the doubler stage in a 12 GHz instrument is permanently switched 
/// HIFN  off to reduce power consumption and the maximum frequency is limited to 
/// HIFN  6 GHz.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR mode/This control selects the Eco Mode..
ViStatus _VI_FUNC rssgs_SetSYSTemEcoMode (ViSession instrSession,
                                          ViInt32 mode)
{
	return rssgs_setEnum (instrSession,
                             "SYST:EMOD",
                             mode,
                             ecoModeArr,
                             VI_ERROR_PARAMETER2);    
}

/// HIFN  This function returns and selects the Eco Mode. With enabled eco mode 
/// HIFN  EM1 the doubler stage in a 12 GHz instrument is permanently switched 
/// HIFN  off to reduce power consumption and the maximum frequency is limited to 
/// HIFN  6 GHz.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR mode/This control returns the Eco Mode.
ViStatus _VI_FUNC rssgs_GetSYSTemEcoMode (ViSession instrSession,
                                          ViInt32 *mode)
{
	return rssgs_getEnum (instrSession,
                             "SYST:EMOD?",
                             mode,
                             ecoModeArr,
                             VI_ERROR_PARAMETER2);    
}


//=====================================================================================
/// HIFN This function activates and deactivates the specified protection level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR protectionLevel/Specify protection level.
/// HIPAR protection/Activates and deactivates the specified protection level.
/// HIPAR password/The password for the specified protection level.
ViStatus _VI_FUNC rssgs_SetSYSTemProtect (ViSession instrSession,
                                          ViInt32 protectionLevel,
                                          ViBoolean protection,
                                          ViString password)
{
    ViChar  buffer [BUFFER_SIZE];

    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, protectionLevel, 1, 5), 2);
    CHECKPARAM (rssgs_invalidViBooleanRange (instrSession, protection), 3);
    CHECKNULL (password, 4);

    if (protection == VI_FALSE)
        sprintf (buffer, "SYST:PROT" _PERCD ":STAT OFF,%s", protectionLevel, password);
    else
        sprintf (buffer, "SYST:PROT" _PERCD ":STAT ON", protectionLevel);

    return rssgs_sendCmd (instrSession, buffer);
}

//=====================================================================================
/// HIFN This function queries state of the specified protection level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR protectionLevel/Specify protection level.
/// HIPAR protection/Returns state of the specified protection level.
ViStatus _VI_FUNC rssgs_GetSYSTemProtect (ViSession instrSession,
                                          ViInt32 protectionLevel,
                                          ViBoolean *protection)
{
    ViChar  buffer [BUFFER_SIZE];

    CHECKPARAM (rssgs_invalidViInt32Range (instrSession, protectionLevel, 1, 5), 2);
    sprintf (buffer, "SYST:PROT" _PERCD ":STAT?", protectionLevel);

    return rssgs_getViBoolean (instrSession, 
                               buffer,
                               protection,
                               VI_ERROR_PARAMETER3);
}

/// HIFN This function queries the SCPI version with which the instrument complies.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR SCPIVersion/Returns the SCPI version with which the instrument complies.
ViStatus _VI_FUNC rssgs_GetSYSTemVersion (ViSession instrSession,
                                 ViChar _VI_FAR SCPIVersion[])
{
    ViStatus rssgs_status = VI_SUCCESS;

    rssgs_status = rssgs_getViString (instrSession,
                              "SYST:VERS?",
                              SCPIVersion,
                              32,
                              VI_ERROR_PARAMETER2);

    return (rssgs_status < VI_SUCCESS) ? rssgs_status : VI_SUCCESS;
}

//=====================================================================================
/// HIFN This function returns if the startup of the instrument is completed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR startupComplete/This control displays the startup of the instrument is completed.
ViStatus _VI_FUNC rssgs_GetSYSTemStartupStatus (ViSession instrSession,
                                                ViInt32 *startupComplete)
{
    return rssgs_getViInt32 (instrSession, 
                             "SYST:STAR:COMP?", 
                             startupComplete, 
                             VI_ERROR_PARAMETER2);
}

/// HIFN This function sets the Terminator mode for the IEC bus.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/This control sets the Terminator mode for the IEC bus.
ViStatus _VI_FUNC rssgs_SetSYSTemTerminatorMode (ViSession instrSession,
                                                 ViInt32 mode)
{
    return rssgs_setEnum (instrSession,
                          "SYST:COMM:GPIB:LTER",
                          mode,
                          terminatorModeArr,
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the Terminator mode for the IEC bus.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR mode/Displays the Terminator mode for the IEC bus.
ViStatus _VI_FUNC rssgs_GetSYSTemTerminatorMode (ViSession instrSession,
                                                 ViInt32 *mode)
{
    return rssgs_getEnum (instrSession,
                          "SYST:COMM:GPIB:LTER?",
                          mode,
                          terminatorModeArr,
                          VI_ERROR_PARAMETER2);
}





/// HIFN This function queries the list of system hardware assembly names.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR assemblyIndex/This control selects the assembly group (0 = common, 1 = RF)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of hardware assembly names
ViStatus _VI_FUNC rssgs_GetSYSTemHwAssemblyNames  (ViSession instrSession,
                                                 ViInt32 assemblyIndex,
                                                 ViInt32 arraySize,
                                                 ViChar _VI_FAR list[])
{
  char cmd[100];
  if (assemblyIndex<0 || assemblyIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,"SYSTem:HARDware:ASSembly" _PERCD ":NAMe?", assemblyIndex+1);
  return rssgs_getCatalog (instrSession,
      cmd,
      (ViUInt32) arraySize,
      list,
      VI_ERROR_PARAMETER3);
}  
                                                  
/// HIFN This function queries the list of system hardware part numbers.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR assemblyIndex/This control selects the assembly group (0 = common, 1 = RF)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of hardware part numbers
ViStatus _VI_FUNC rssgs_GetSYSTemHwPartNumbers  (ViSession instrSession,
                                                   ViInt32 assemblyIndex,
                                                   ViInt32 arraySize,
                                                   ViChar _VI_FAR list[])
{
  char cmd[100];
  if (assemblyIndex<0 || assemblyIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:HARDware:ASSembly" _PERCD ":PNUMber?", assemblyIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}  

/// HIFN This function queries the list of system hardware revisions
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR assemblyIndex/This control selects the assembly group (0 = common, 1 = RF)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of hardware revisions.
ViStatus _VI_FUNC rssgs_GetSYSTemHwRevisions  (ViSession instrSession,
                                                 ViInt32 assemblyIndex,
                                                 ViInt32 arraySize,
                                                 ViChar _VI_FAR list[])
{
  char cmd[100];
  if (assemblyIndex<0 || assemblyIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:HARDware:ASSembly" _PERCD ":REVision?", assemblyIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}  

/// HIFN This function queries the list of hardware module serial numbers.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR assemblyIndex/This control selects the assembly group (0 = common, 1 = RF)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of hardware serial numbers
ViStatus _VI_FUNC rssgs_GetSYSTemHwSerialNumbers  (ViSession instrSession,
                                               ViInt32 assemblyIndex,
                                               ViInt32 arraySize,
                                               ViChar _VI_FAR list[])
{
  char cmd[100];
  if (assemblyIndex<0 || assemblyIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:HARDware:ASSembly" _PERCD ":SNUMber?", assemblyIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}  




/// HIFN This function queries the list of system hardware revisions
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR optionIndex/This control selects the option group (0 = Hardware, 1 = Software)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of system firmware option designations
ViStatus _VI_FUNC rssgs_GetSYSTemSwOptionDesignations  (ViSession instrSession,
                                                   ViInt32 optionIndex,
                                                   ViInt32 arraySize,
                                                   ViChar _VI_FAR list[])
{
  char cmd[100];
  if (optionIndex<0 || optionIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:SOFTware:OPTion" _PERCD ":DESignation?", optionIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}  

/// HIFN This function queries the list of system firmware option names
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR optionIndex/This control selects the option group (0 = Hardware, 1 = Software)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of system firmware option names
ViStatus _VI_FUNC rssgs_GetSYSTemSwOptionNames  (ViSession instrSession,
                                                        ViInt32 optionIndex,
                                                        ViInt32 arraySize,
                                                        ViChar _VI_FAR list[])
{
  char cmd[100];
  if (optionIndex<0 || optionIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:SOFTware:OPTion" _PERCD ":NAME?", optionIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}  

/// HIFN This function queries the list of system firmware option expiration entries
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR optionIndex/This control selects the option group (0 = hardware, 1 = software)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of system firmware option expiration entries.
ViStatus _VI_FUNC rssgs_GetSYSTemSwOptionExpirations  (ViSession instrSession,
                                                 ViInt32 optionIndex,
                                                 ViInt32 arraySize,
                                                 ViChar _VI_FAR list[])
{
  char cmd[100];
  if (optionIndex<0 || optionIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:SOFTware:OPTion" _PERCD ":EXP?", optionIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}  

/// HIFN This function queries the list of system firmware option license count entries
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR optionIndex/This control selects the option group (0 = hardware, 1 = software)
/// HIPAR arraySize/This control sets the array size.
/// HIPAR list/List of system firmware option license count entries.
ViStatus _VI_FUNC rssgs_GetSYSTemSwOptionLicenses (ViSession instrSession,
                                                       ViInt32 optionIndex,
                                                       ViInt32 arraySize,
                                                       ViChar _VI_FAR list[])
{
  char cmd[100];
  if (optionIndex<0 || optionIndex >1)
  {
    return VI_ERROR_PARAMETER2;
  }
  sprintf (cmd,":SYSTem:SOFTware:OPTion" _PERCD ":LIC?", optionIndex+1);
  return rssgs_getCatalog (instrSession,
    cmd,
    (ViUInt32) arraySize,
    list,
    VI_ERROR_PARAMETER3);
}

/// HIFN This function restarts the device.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_Restart (ViSession instrSession)
{
  return rssgs_sendCmd (instrSession, "REST");
}

/// HIFN This function puts the device in the standby state.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_Standby (ViSession instrSession)
{
  return rssgs_sendCmd (instrSession, "STAN");
}

//=====================================================================================
/// HIFN This function sets the default unit for angles.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR angleUnits/Sets the default unit for angles.
ViStatus _VI_FUNC rssgs_SetAngleUnits (ViSession instrSession,
                                       ViInt32 angleUnits)
{
    return rssgs_setEnum (instrSession, 
                          "UNIT:ANGL", 
                          angleUnits,
                          AngleUnitsArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function returns the unit for angles.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR angleUnits/Gets the unit for angles.
ViStatus _VI_FUNC rssgs_GetAngleUnits (ViSession instrSession,
                                       ViInt32 *angleUnits)
{
    return rssgs_getEnum (instrSession, 
                          "UNIT:ANGL?", 
                          angleUnits,
                          AngleUnitsArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function sets the default unit for power.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR powerUnits/Sets the default unit for power.
ViStatus _VI_FUNC rssgs_SetPowerUnits (ViSession instrSession,
                                       ViInt32 powerUnits)
{
    return rssgs_setEnum (instrSession, 
                          "UNIT:POW", 
                          powerUnits,
                          PowerUnitsArr, 
                          VI_ERROR_PARAMETER2);
}

//=====================================================================================
/// HIFN This function returns the unit for power.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR powerUnits/Gets the unit for power.
ViStatus _VI_FUNC rssgs_GetPowerUnits (ViSession instrSession,
                                       ViInt32 *powerUnits)
{
    return rssgs_getEnum (instrSession, 
                          "UNIT:POW?", 
                          powerUnits,
                          PowerUnitsArr, 
                          VI_ERROR_PARAMETER2);
}



/// HIFN This function sets the data format for IEC bus.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR format/Sets the data format for IEC bus.
ViStatus _VI_FUNC rssgs_SetFormatData (ViSession instrSession, ViInt32 format)
{
    return rssgs_setEnum (instrSession,
                          "FORM:DATA",
                          format,
                          formatDataArr,
                          VI_ERROR_PARAMETER2);
}

/// HIFN This function queries the data format for IEC bus.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR format/Displays the data format for IEC bus.
ViStatus _VI_FUNC rssgs_GetFormatData (ViSession instrSession, ViInt32 *format)
{
    return rssgs_getEnum (instrSession,
                          "FORM:DATA?",
                          format,
                          formatDataArr,
                          VI_ERROR_PARAMETER2);
}

/// HIFN  This function determines the numerical format which is returned when 
/// HIFN  the status registers are queried.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR format/Sets the numerical format which is returned when the status registers 
/// HIPAR format/are queried.
ViStatus _VI_FUNC rssgs_SetFormatStatusRegisterData(
	ViSession	instrSession,
	ViInt32	format
)
{
	return rssgs_setEnum (instrSession,
                          "FORM:SREG",
                          format,
                          statusRegisterDataArr,
                          VI_ERROR_PARAMETER2);    
}


/// HIFN  This function queries the numerical format which is returned when the 
/// HIFN  status registers are queried.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR format/Returns the numerical format which is returned when the status 
/// HIPAR format/registers are queried.
ViStatus _VI_FUNC rssgs_GetFormatStatusRegisterData(
	ViSession	instrSession,
	ViInt32*	format
)
{
return rssgs_getEnum (instrSession,
                          "FORM:SREG?",
                          format,
                          statusRegisterDataArr,
                          VI_ERROR_PARAMETER2);    
}


/// HIFN  This function determines the sequence of bytes within a binary block. 
/// HIFN  This only affects blocks which use the IEEE754 format internally.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR byteOrder/Sets the sequence of bytes within a binary block. This only affects 
/// HIPAR byteOrder/blocks which use the IEEE754 format internally.
ViStatus _VI_FUNC rssgs_SetFormatByteOrder(
	ViSession	instrSession,
	ViInt32	byteOrder
)
{
	return rssgs_setEnum (instrSession,
                          "FORM:BORD",
                          byteOrder,
                          byteOrderArr,
                          VI_ERROR_PARAMETER2);    
}


/// HIFN  This function queries the sequence of bytes within a binary block. 
/// HIFN  This only affects blocks which use the IEEE754 format internally.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize 
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR byteOrder/Returns the sequence of bytes within a binary block. This only affects 
/// HIPAR byteOrder/blocks which use the IEEE754 format internally.
ViStatus _VI_FUNC rssgs_GetFormatByteOrder(
	ViSession	instrSession,
	ViInt32*	byteOrder
)
{
	return rssgs_getEnum (instrSession,
                          "FORM:BORD?",
                          byteOrder,
                          byteOrderArr,
                          VI_ERROR_PARAMETER2);     
}

/// HIFN This function switches state checking of the instrument (reading of the Standard Event Register and checking it for error) status subsystem. Driver functions are using state checking which is by default enabled.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control switches instrument state checking On or Off.
ViStatus _VI_FUNC rssgs_errorCheckState (ViSession instrSession,
                                          ViBoolean state)
{
    ViStatus    rssgs_status = VI_SUCCESS;
    rssgs_instrRange instrPtr = NULL;
    unsigned int     ptrHandle = 0xffffffff;

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));

    CHECKNULL (instrPtr, 1);
    
    instrPtr -> errorChecking = state;

    return rssgs_status;
}

/// HIFN This function switches option checking of the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR state/This control switches instrument and option checking On or Off.
ViStatus _VI_FUNC rssgs_optionCheckState (ViSession instrSession,
                                          ViBoolean state)
{
    ViStatus    rssgs_status = VI_SUCCESS;
    rssgs_instrRange instrPtr = NULL;
    unsigned int        ptrHandle = 0xffffffff;

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));

    if (instrPtr != NULL)
        instrPtr -> optionChecking = state;

    return rssgs_status;
    
}

/// HIFN This function switches range checking of the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR rangeChecking/This control switches instrument and range checking On or Off.
ViStatus _VI_FUNC rssgs_setCheckRange(
    ViSession   instrSession,
    ViBoolean   rangeChecking
)
{
    ViStatus    rssgs_status = VI_SUCCESS;
    rssgs_instrRange instrPtr;
    unsigned int     ptrHandle = 0xffffffff;

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));

    instrPtr -> rangeChecking = rangeChecking;

    return rssgs_status;
}

/// HIFN This function writes commands and queries to the instrument to modify parameters and query device settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR writeBuffer/The user can use this control to send common commands and queries to the instrument. This control can also be used to write any valid command to the instrument.
ViStatus _VI_FUNC rssgs_writeInstrData (ViSession instrSession, ViString writeBuffer)
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rsPrintf (instrSession, "%s\n", writeBuffer));

    return rssgs_status;
}

/// HIFN This function reads data from the instrument's output buffer and returns it to the specified variable in memory.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR numberBytesToRead/The number of bytes specified by this control should be greater than or equal to the number of bytes which are to be read from the instrument. If the actual number of bytes to be read is greater than the number this control specifies then multiple reads will be required to empty the instrument's output buffer.
/// HIPAR readBuffer/The incoming data from the instrument is placed into this variable.
/// HIPAR numBytesRead/This variable contains the actual number of bytes read from the instrument. This is the value which is returned by the read function.
ViStatus _VI_FUNC rssgs_readInstrData (ViSession instrSession,
                                        ViInt32 numberBytesToRead,
                                        ViChar _VI_FAR readBuffer[],
                                        ViPInt32 numBytesRead)
{
    ViStatus    rssgs_status = VI_SUCCESS;
    ViUInt32    tmpCnt = 0;

    if (numBytesRead != NULL)
        *numBytesRead = 0L;

    CHECKERR (rsRead (instrSession, (ViBuf)readBuffer, (ViUInt32) numberBytesToRead, &tmpCnt));

    if (numBytesRead)
        *numBytesRead = (ViInt32)tmpCnt;

    return rssgs_status;
}


/// HIFN RESET resets the instrument to a defined default state. The function has the same effect as pressing the PRESET key.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_reset (ViSession instrSession)
{
    ViStatus rssgs_status = VI_SUCCESS;

    /*  Initialize the instrument to a known state.  */
    CHECKERR (rsPrintf (instrSession, "*RST;*CLS\n"));
    CHECKERR (rssgs_defaultInstrSetup (instrSession));

    return rssgs_status;
}

//=====================================================================================
/// HIFN clears the device's input and output buffers
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
ViStatus _VI_FUNC rssgs_Clear (ViSession instrSession)
{
    return rsClear(instrSession);
}

/// HIFN This function runs the instrument's self test routine and returns the test result(s).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR testResult/This control contains the value returned from the instrument self test. Zero means success. For any other code, see the device's operator's manual.
/// HIPAR testMessage/This control contains the string returned from the self test. See the device's operation manual for an explanation of the string's contents.
ViStatus _VI_FUNC rssgs_self_test (ViSession instrSession,
                                    ViPInt16 ptestResult,
                                    ViChar _VI_FAR testMessage[])
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR( rsQueryf (instrSession, "*TST?\n", "%hd", ptestResult));

    sprintf (testMessage, "%s", (*ptestResult == 0) ? "Self-Test Passed" : "Self-Test Failed");

    return rssgs_status;
}

/// HIFN This function reads an error code from the instrument's error queue.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR errorCode/This control returns the error code read from the instrument's error queue.
/// HIPAR errorMessage/This control returns the error message string read from the instrument's error message queue.
ViStatus _VI_FUNC rssgs_error_query (ViSession instrSession,
                                      ViPInt32 errorCode,
                                      ViChar _VI_FAR errorMessage[])
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rsPrintf (instrSession, ":SYST:ERR?\n"));
    CHECKERR (rsScanf (instrSession, "" _PERCD ",\"%[^\"]", errorCode, errorMessage));

    return rssgs_status;
}

/// HIFN This function takes the Status Code returned by the instrument driver functions, interprets it and returns it as a user readable string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR statusCode/This control accepts the Status Code returned from the instrument driver functions.
/// HIPAR errMessage/This control returns the interpreted Status Code as a user readable message string.
ViStatus _VI_FUNC rssgs_error_message (ViSession instrSession,
                                        ViStatus statusCode,
                                        ViChar _VI_FAR errMessage[])
{
    ViStatus rssgs_status = VI_WARN_UNKNOWN_STATUS;
    int i;

    static rssgs_tStringValPair statusDescArray[] = {
      // VI_xx error code messages added here to provide all-English error messages as far as possible
      // (rsStatusDesc may return local-language output and )

        { VI_SUCCESS,                                       "Operation completed successfully." },
        { VI_SUCCESS_EVENT_EN,                              "Specified event is already enabled for at least one of the specified mechanisms." },
        { VI_SUCCESS_EVENT_DIS,                             "Specified event is already disabled for at least one of the specified mechanisms." },
        { VI_SUCCESS_QUEUE_EMPTY,                           "Operation completed successfully, but queue was already empty." },
        { VI_SUCCESS_TERM_CHAR,                             "The specified termination character was read." },
        { VI_SUCCESS_MAX_CNT,                               "The number of bytes read is equal to the input count." },
        { VI_WARN_CONFIG_NLOADED,                           "The specified configuration either does not exist or could not be loaded; using VISA-specified defaults." },
        { VI_SUCCESS_DEV_NPRESENT,                          "Session opened successfully, but the device at the specified address is not responding." },
        { VI_SUCCESS_TRIG_MAPPED,                           "The path from trigSrc to trigDest is already mapped." },
        { VI_SUCCESS_QUEUE_NEMPTY,                          "Wait terminated successfully on receipt of an event notification. There is still at least one more event occurrence of the requested type(s) available for this session." },
        { VI_WARN_NULL_OBJECT,                              "The specified object reference is uninitialized." },
        { VI_WARN_NSUP_ATTR_STATE,                          "Although the specified state of the attribute is valid, it is not supported by this resource implementation." },
        { VI_WARN_UNKNOWN_STATUS,                           "The status code passed to the operation could not be interpreted." },
        { VI_WARN_NSUP_BUF,                                 "The specified buffer is not supported." },
        { VI_WARN_EXT_FUNC_NIMPL,                           "The operation succeeded, but a lower level driver did not implement the extended functionality." },
        { VI_SUCCESS_NCHAIN,                                "Event handled successfully. Do not invoke any other handlers on this session for this event." },
        { VI_SUCCESS_NESTED_SHARED,                         "Operation completed successfully, and this session has nested shared locks." },
        { VI_SUCCESS_NESTED_EXCLUSIVE,                      "Operation completed successfully, and this session has nested exclusive locks." },
        { VI_SUCCESS_SYNC,                                  "Asynchronous operation request was actually performed synchronously." },
        { VI_ERROR_SYSTEM_ERROR,                            "Unknown system error (miscellaneous error)." },
        { VI_ERROR_INV_OBJECT,                              "The given session or object reference is invalid." },
        { VI_ERROR_RSRC_LOCKED,                             "Specified type of lock cannot be obtained or specified operation cannot be performed, because the resource is locked." },
        { VI_ERROR_INV_EXPR,                                "Invalid expression specified for search." },
        { VI_ERROR_RSRC_NFOUND,                             "Insufficient location information or the device or resource is not present in the system." },
        { VI_ERROR_INV_RSRC_NAME,                           "Invalid resource reference specified. Parsing error." },
        { VI_ERROR_INV_ACC_MODE,                            "Invalid access mode." },
        { VI_ERROR_TMO,                                     "Timeout expired before operation completed." },
        { VI_ERROR_CLOSING_FAILED,                          "Unable to deallocate the previously allocated data structures corresponding to this session or object reference." },
        { VI_ERROR_INV_DEGREE,                              "Specified degree is invalid." },
        { VI_ERROR_INV_JOB_ID,                              "Specified job identifier is invalid." },
        { VI_ERROR_NSUP_ATTR,                               "The specified attribute is not defined or supported by the referenced session, event, or find list." },
        { VI_ERROR_NSUP_ATTR_STATE,                         "The specified state of the attribute is not valid, or is not supported as defined by the session, event, or find list." },
        { VI_ERROR_ATTR_READONLY,                           "The specified attribute is read-only." },
        { VI_ERROR_INV_LOCK_TYPE,                           "The specified type of lock is not supported by this resource." },
        { VI_ERROR_INV_ACCESS_KEY,                          "The access key to the resource associated with this session is invalid." },
        { VI_ERROR_INV_EVENT,                               "Specified event type is not supported by the resource." },
        { VI_ERROR_INV_MECH,                                "Invalid mechanism specified." },
        { VI_ERROR_HNDLR_NINSTALLED,                        "A handler is not currently installed for the specified event." },
        { VI_ERROR_INV_HNDLR_REF,                           "The given handler reference is invalid." },
        { VI_ERROR_INV_CONTEXT,                             "Specified event context is invalid." },
        { VI_ERROR_QUEUE_OVERFLOW,                          "The event queue for the specified type has overflowed (usually due to previous events not having been closed)." },
        { VI_ERROR_NENABLED,                                "The session must be enabled for events of the specified type in order to receive them." },
        { VI_ERROR_ABORT,                                   "The operation was aborted." },
        { VI_ERROR_RAW_WR_PROT_VIOL,                        "Violation of raw write protocol occurred during transfer." },
        { VI_ERROR_RAW_RD_PROT_VIOL,                        "Violation of raw read protocol occurred during transfer." },
        { VI_ERROR_OUTP_PROT_VIOL,                          "Device reported an output protocol error during transfer." },
        { VI_ERROR_INP_PROT_VIOL,                           "Device reported an input protocol error during transfer." },
        { VI_ERROR_BERR,                                    "Bus error occurred during transfer." },
        { VI_ERROR_IN_PROGRESS,                             "Unable to queue the asynchronous operation because there is already an operation in progress." },
        { VI_ERROR_INV_SETUP,                               "Unable to start operation because setup is invalid (due to attributes being set to an inconsistent state)." },
        { VI_ERROR_QUEUE_ERROR,                             "Unable to queue asynchronous operation." },
        { VI_ERROR_ALLOC,                                   "Insufficient system resources to perform necessary memory allocation." },
        { VI_ERROR_INV_MASK,                                "Invalid buffer mask specified." },
        { VI_ERROR_IO,                                      "Could not perform operation because of I/O error." },
        { VI_ERROR_INV_FMT,                                 "A format specifier in the format string is invalid." },
        { VI_ERROR_NSUP_FMT,                                "A format specifier in the format string is not supported." },
        { VI_ERROR_LINE_IN_USE,                             "The specified trigger line is currently in use." },
        { VI_ERROR_NSUP_MODE,                               "The specified mode is not supported by this VISA implementation." },
        { VI_ERROR_SRQ_NOCCURRED,                           "Service request has not been received for the session." },
        { VI_ERROR_INV_SPACE,                               "Invalid address space specified." },
        { VI_ERROR_INV_OFFSET,                              "Invalid offset specified." },
        { VI_ERROR_INV_WIDTH,                               "Invalid source or destination width specified." },
        { VI_ERROR_NSUP_OFFSET,                             "Specified offset is not accessible from this hardware." },
        { VI_ERROR_NSUP_VAR_WIDTH,                          "Cannot support source and destination widths that are different." },
        { VI_ERROR_WINDOW_NMAPPED,                          "The specified session is not currently mapped." },
        { VI_ERROR_RESP_PENDING,                            "A previous response is still pending, causing a multiple query error." },
        { VI_ERROR_NLISTENERS,                              "No Listeners condition is detected (both NRFD and NDAC are deasserted)." },
        { VI_ERROR_NCIC,                                    "The interface associated with this session is not currently the controller in charge." },
        { VI_ERROR_NSYS_CNTLR,                              "The interface associated with this session is not the system controller." },
        { VI_ERROR_NSUP_OPER,                               "The given session or object reference does not support this operation." },
        { VI_ERROR_INTR_PENDING,                            "An interrupt is still pending from a previous call." },
        { VI_ERROR_ASRL_PARITY,                             "A parity error occurred during transfer." },
        { VI_ERROR_ASRL_FRAMING,                            "A framing error occurred during transfer." },
        { VI_ERROR_ASRL_OVERRUN,                            "An overrun error occurred during transfer. A character was not read from the hardware before the next character arrived." },
        { VI_ERROR_TRIG_NMAPPED,                            "The path from trigSrc to trigDest is not currently mapped." },
        { VI_ERROR_NSUP_ALIGN_OFFSET,                       "The specified offset is not properly aligned for the access width of the operation." },
        { VI_ERROR_USER_BUF,                                "A specified user buffer is not valid or cannot be accessed for the required size." },
        { VI_ERROR_RSRC_BUSY,                               "The resource is valid, but VISA cannot currently access it." },
        { VI_ERROR_NSUP_WIDTH,                              "Specified width is not supported by this hardware." },
        { VI_ERROR_INV_PARAMETER,                           "The value of some parameter,which parameter is not known,is invalid." },
        { VI_ERROR_INV_PROT,                                "The protocol specified is invalid." },
        { VI_ERROR_INV_SIZE,                                "Invalid size of window specified." },
        { VI_ERROR_WINDOW_MAPPED,                           "The specified session currently contains a mapped window." },
        { VI_ERROR_NIMPL_OPER,                              "The given operation is not implemented." },
        { VI_ERROR_INV_LENGTH,                              "Invalid length specified." },
        { VI_ERROR_INV_MODE,                                "The specified mode is invalid." },
        { VI_ERROR_SESN_NLOCKED,                            "The current session did not have any lock on the resource." },
        { VI_ERROR_MEM_NSHARED,                             "The device does not export any memory." },
        { VI_ERROR_LIBRARY_NFOUND,                          "A code library required by VISA could not be located or loaded." },
        { VI_ERROR_NSUP_INTR,                               "The interface cannot generate an interrupt on the requested level or with the requested statusID value." },
        { VI_ERROR_INV_LINE,                                "The value specified by the line parameter is invalid." },
        { VI_ERROR_FILE_ACCESS,                             "An error occurred while trying to open the specified file. Possible reasons include an invalid path or lack of access rights." },
        { VI_ERROR_FILE_IO,                                 "An error occurred while performing I/O on the specified file." },
        { VI_ERROR_NSUP_LINE,                               "One of the specified lines (trigSrc or trigDest) is not supported by this VISA implementation, or the combination of lines is not a valid mapping." },
        { VI_ERROR_NSUP_MECH,                               "The specified mechanism is not supported by the given event type." },
        { VI_ERROR_INTF_NUM_NCONFIG,                        "The interface type is valid but the specified interface number is not configured." },
        { VI_ERROR_CONN_LOST,                               "The connection for the given session has been lost." },
        { VI_ERROR_NPERMISSION,                             "Access to the resource or remote machine is denied." },


        { VI_ERROR_PARAMETER1,                              "Parameter 1 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER2,                              "Parameter 2 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER3,                              "Parameter 3 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER4,                              "Parameter 4 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER5,                              "Parameter 5 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER6,                              "Parameter 6 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER7,                              "Parameter 7 out of range, or error trying to set it." },
        { VI_ERROR_PARAMETER8,                              "Parameter 8 out of range, or error trying to set it." },
        { VI_ERROR_FAIL_ID_QUERY,                           "Instrument failed the ID Query." },
        { VI_ERROR_INV_RESPONSE,                            "Invalid response from instrument." },


        { VI_WARN_NSUP_ID_QUERY,                            "Identification query not supported." },
        { VI_WARN_NSUP_RESET,                               "Reset operation not supported." },
        { VI_WARN_NSUP_SELF_TEST,                           "Self test operation not supported." },
        { VI_WARN_NSUP_ERROR_QUERY,                         "Error query operation not supported." },
        { VI_WARN_NSUP_REV_QUERY,                           "Revision query not supported." },


      /* --- Instrument specific error codes --- */

      {RSSGS_ERROR_INSTRUMENT_ERROR,          "ERROR: Instrument status error"},
      {RSSGS_ERROR_INVALID_CONFIGURATION,     "ERROR: Instrument configuration error"},
      {RSSGS_ERROR_INSTRUMENT_OPTION,         "ERROR: Option not installed"},
      {RSSGS_ERROR_INSTRUMENT_NSUP_MODEL,     "ERROR: Required Instrument model is not available."},
      {RSSGS_ERROR_SETTINGS_CONFLICT,         "ERROR: Instrument settings conflict."},
      {RSSGS_ERROR_CALIBRATION_FAILED,        "ERROR: Calibration failed."},
      {RSSGS_ERROR_INSTRUMENT_DATA_NOT_AVAILABLE, "ERROR: Data not available."},

      /* --- Instrument specific warnings --- */

      {RSSGS_WARNING_SMALL_BUFFER,            "WARNING: Buffer too small"},

      {VI_NULL, VI_NULL}
    };

    for (i = 0; statusDescArray[i].stringName; i++)
    {
      if (statusDescArray[i].stringVal == statusCode)
      {
        sprintf(errMessage, "%s", statusDescArray[i].stringName);
        rssgs_status = VI_SUCCESS;
        break;
      }
    }

    if (rssgs_status != VI_SUCCESS)
    {
      rssgs_status = rsStatusDesc(instrSession, statusCode, errMessage);
      if (rssgs_status != VI_SUCCESS)
        sprintf(errMessage, "The status code passed to the operation could not be interpreted.");
    }

    return rssgs_status;
}

/// HIFN This function returns the revision numbers of the instrument driver and instrument firmware, and tells the user with which instrument firmware this revision of the driver is compatible.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize function to select the desired instrument driver session.
/// HIPAR driverRev/This control returns the Instrument Driver Software Revision.
/// HIPAR instrRev/This control returns the Instrument Firmware Revision.
ViStatus _VI_FUNC rssgs_revision_query (ViSession instrSession,
                    ViChar _VI_FAR driverRev[], ViChar _VI_FAR instrRev[])
{
    ViStatus rssgs_status = VI_SUCCESS;

    CHECKERR (rsPrintf (instrSession, "*IDN?\n"));
    //CHECKERR (rsScanf (instrSession, "%*[^,],%*[^,],%*[^,],%[^\r\n]", instrRev));
    CHECKERR (rsScanf (instrSession, "%*[^-]-%[^ (\r\n]", instrRev));

    sprintf (driverRev, "%s", RSSGS_REVISION);

    return rssgs_status;
}

/// HIFN Find devices on the PCIe bus and return their serial numbers
/// HIRET Error code
/// HIPAR vendorId/PCI id of the device's manufacturer
/// HIPAR deviceId/PCI id of the device
/// HIPAR serialNumbers/array with returned serial numbers. Allocated by caller
/// HIPAR serialNumberCount/allocated size of array. On return: number of found devices
ViStatus _VI_FUNC rssgs_enumerate_pcie_devices (ViInt32 vendorId, ViInt32 deviceId, ViInt32 serialNumbers[], ViPInt32 serialNumberCount)
{
  ViStatus status = viEnumerate_pcie_devices (vendorId, deviceId, serialNumbers, serialNumberCount);
  return status;
}

/// HIFN Find fast socket devices and return their types, serial numbers
/// HIFN and IP addresses. Call function with infoCount=0/info = NULL,
/// HIFN allocate buffer and call again with valid buffers.  
/// HIRET Returns the status code of this operation.
/// HIPAR info/Array of pointers to string buffers. Each buffer must be at least 64 bytes long.
/// HIPAR info/Each string returned by the function is a comma-separated list of instrument type, 
/// HIPAR info/serial number and IP address e.g. SGT100A,100027,169.254.45.172. 
/// HIPAR infoCount/Number of buffers in array info. On return: the number of found devices.
ViStatus _VI_FUNC rssgs_enumerate_socket_devices (ViPString *info, ViPInt32 infoCount)                                               
{
  ViStatus status = viEnumerate_socket_devices (info, infoCount);
  return status;
}

/*===========================================================================*/
/* Function: Delay                                                           */
/* Purpose : With a call to rssgs_delay, the current program is suspended    */
/*           from execution for the number of seconds specified by the       */
/*           argument seconds. rssgs_delay is accurate to a millisecond.     */
/*           This function is platform independent.                          */
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_delay (/*@unused@*/ViSession instrSession, ViReal64 seconds)
{
  clock_t StartTime = clock();
  instrSession = 0;
  
  while ((((double)(clock() - StartTime))/CLOCKS_PER_SEC) < (seconds)){
  #ifdef _WIN32
          Sleep(0);
  #endif
  }
  return VI_SUCCESS;
}

/*===========================================================================*/
/* Function: Close                                                           */
/* Purpose:  This function closes the instrument.                            */
/*===========================================================================*/
ViStatus _VI_FUNC rssgs_close (ViSession instrSession)
{
    rssgs_instrRange instrPtr = NULL;
    ViSession        rmSession = 0;
    ViStatus         rssgs_status = VI_SUCCESS;
    unsigned int     ptrHandle = 0xffffffff;

    CHECKERR( rsGetAttribute (instrSession, VI_ATTR_RM_SESSION, &rmSession));

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));
    rssgs_RemovePointer(ptrHandle);
    
    if (instrPtr != NULL)
        free (instrPtr);

    rssgs_status = rsClose (instrSession);

    return rssgs_status;
}

/*===========================================================================*/
/*----------- INSERT INSTRUMENT-DEPENDENT UTILITY ROUTINES HERE -------------*/
/*===========================================================================*/

/*===========================================================================*/
/* Function: Default Instrument Setup                                        */
/* Purpose:  This function sends a default setup to the instrument.  This    */
/*           function is called by the rssgs_reset operation and by the     */
/*           rssgs_init function if the reset option has not been           */
/*           selected.  This function is useful for configuring any          */
/*           instrument settings that are required by the rest of the        */
/*           instrument driver functions such as turning headers ON or OFF   */
/*           or using the long or short form for commands, queries, and data.*/
/*===========================================================================*/
ViStatus rssgs_defaultInstrSetup (ViSession instrSession)
{
    ViStatus            rssgs_status = VI_SUCCESS;
    rssgs_instrRange    instrPtr = NULL;
    unsigned int        ptrHandle = 0xffffffff;

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));
    
    CHECKNULL (instrPtr, 1);
    
    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */

    CHECKERR (rsGetAttribute (instrSession, VI_ATTR_USER_DATA, &ptrHandle));
    instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));

    instrPtr -> errorChecking = VI_TRUE;
    instrPtr -> optionChecking = VI_TRUE;
    strcpy (instrPtr -> instrDriverRevision, RSSGS_REVISION);

    return rssgs_status;
}

/*===========================================================================*/
/* Function: Initialize Clean Up                                             */
/* Purpose:  This function is used only by the rssgs_init function.  When   */
/*           an error is detected this function is called to close the       */
/*           open resource manager and instrument object sessions and to     */
/*           set the instrSession that is returned from rssgs_init to       */
/*           VI_NULL.                                                        */
/*===========================================================================*/
ViStatus rssgs_initCleanUp (ViSession openRMSession,
                             ViPSession openInstrSession,
                             ViStatus currentStatus)
{
    unsigned int        ptrHandle = 0xffffffff;

    if (rsGetAttribute (*openInstrSession, VI_ATTR_USER_DATA, &ptrHandle) >= 0)
    {
      rssgs_instrRange instrPtr = NULL;
      instrPtr = (rssgs_instrRange)(rssgs_GetPointer (ptrHandle));
      if (instrPtr != NULL)
        free (instrPtr);
    
    }

    (void) rsClose (*openInstrSession);
    (void) rsClose (openRMSession);
    *openInstrSession = VI_NULL;

    return currentStatus;
}

/*===========================================================================*/
/* Function: get/set Attribute                                               */
/* Purpose:                                                                  */
/*===========================================================================*/

ViStatus _VI_FUNC rssgs_getAttribute(ViSession instrSession, ViAttr attribute, void* value)
{
  ViStatus            rssgs_status = VI_SUCCESS;
  CHECKERR(rsGetAttribute(instrSession, attribute, value));
  return rssgs_status;
}

ViStatus _VI_FUNC rssgs_setAttribute(ViSession instrSession, ViAttr attribute, ViAttrState value)
{
  ViStatus            rssgs_status = VI_SUCCESS;
  CHECKERR(rsSetAttribute(instrSession, attribute, value));
  return rssgs_status;
}
